
/* png.c - location for general purpose libpng functions
*
* libpng version 1.2.8 - December 3, 2004
* For conditions of distribution and use, see copyright notice in png.h
* Copyright (c) 1998-2004 Glenn Randers-Pehrson
* (Version 0.96 Copyright (c) 1996, 1997 Andreas Dilger)
* (Version 0.88 Copyright (c) 1995, 1996 Guy Eric Schalnat, Group 42, Inc.)
*/
#define PNG_INTERNAL
#define PNG_NO_EXTERN
#include "png.h"
typedef version_1_2_8 Your_png_h_is_not_version_1_2_8;
#ifdef PNG_USE_GLOBAL_ARRAYS
const char png_libpng_ver[18] = PNG_LIBPNG_VER_STRING;
const png_byte FARDATA png_sig[8] = {137, 80, 78, 71, 13, 10, 26, 10};
PNG_IHDR;
PNG_IDAT;
PNG_IEND;
PNG_PLTE;
PNG_bKGD;
PNG_cHRM;
PNG_gAMA;
PNG_hIST;
PNG_iCCP;
PNG_iTXt;
PNG_oFFs;
PNG_pCAL;
PNG_sCAL;
PNG_pHYs;
PNG_sBIT;
PNG_sPLT;
PNG_sRGB;
PNG_tEXt;
PNG_tIME;
PNG_tRNS;
PNG_zTXt;
const int FARDATA png_pass_start[] = {0, 4, 0, 2, 0, 1, 0};
const int FARDATA png_pass_inc[] = {8, 8, 4, 4, 2, 2, 1};
const int FARDATA png_pass_ystart[] = {0, 0, 4, 0, 2, 0, 1};
const int FARDATA png_pass_yinc[] = {8, 8, 8, 4, 4, 2, 2};
#ifdef PNG_HAVE_ASSEMBLER_COMBINE_ROW
const int FARDATA png_pass_width[] = {8, 4, 4, 2, 2, 1, 1};
#endif
const int FARDATA png_pass_mask[] = {0x80, 0x08, 0x88, 0x22, 0xaa, 0x55, 0xff};
const int FARDATA png_pass_dsp_mask[]
= {0xff, 0x0f, 0xff, 0x33, 0xff, 0x55, 0xff};
#endif
void PNGAPI
png_set_sig_bytes(png_structp png_ptr, int num_bytes)
{
	png_debug(1, "in png_set_sig_bytes\n");
	if(num_bytes > 8)
		png_error(png_ptr, "Too many bytes for PNG signature.");
	png_ptr->sig_bytes = (png_byte)(num_bytes < 0 ? 0 : num_bytes);
}
int PNGAPI
png_sig_cmp(png_bytep sig, png_size_t start, png_size_t num_to_check)
{
	png_byte png_signature[8] = {137, 80, 78, 71, 13, 10, 26, 10};
	if(num_to_check > 8)
		num_to_check = 8;
	else if(num_to_check < 1)
		return (0);
	if(start > 7)
		return (0);
	if(start + num_to_check > 8)
		num_to_check = 8 - start;
	return ((int)(png_memcmp(&sig[start], &png_signature[start], num_to_check)));
}
int PNGAPI
png_check_sig(png_bytep sig, int num)
{
	return ((int)!png_sig_cmp(sig, (png_size_t)0, (png_size_t)num));
}
#ifdef PNG_1_0_X
voidpf PNGAPI
#else
voidpf
#endif
png_zalloc(voidpf png_ptr, uInt items, uInt size)
{
	png_voidp ptr;
	png_structp p = (png_structp)png_ptr;
	png_uint_32 save_flags = p->flags;
	png_uint_32 num_bytes;
	if(items > PNG_UINT_32_MAX/size)
	{
		png_warning ((png_structp)png_ptr, "Potential overflow in png_zalloc()");
		return (NULL);
	}
	num_bytes = (png_uint_32)items * size;
	p->flags |= PNG_FLAG_MALLOC_NULL_MEM_OK;
	ptr = (png_voidp)png_malloc((png_structp)png_ptr, num_bytes);
	p->flags = save_flags;
#if defined(PNG_1_0_X) && !defined(PNG_NO_ZALLOC_ZERO)
	if(ptr == NULL)
		return ((voidpf)ptr);
	if(num_bytes > (png_uint_32)0x8000L)
	{
		png_memset(ptr, 0, (png_size_t)0x8000L);
		png_memset((png_bytep)ptr + (png_size_t)0x8000L, 0,
			(png_size_t)(num_bytes - (png_uint_32)0x8000L));
	}
	else
	{
		png_memset(ptr, 0, (png_size_t)num_bytes);
	}
#endif
	return ((voidpf)ptr);
}
#ifdef PNG_1_0_X
void PNGAPI
#else
void
#endif
png_zfree(voidpf png_ptr, voidpf ptr)
{
	png_free((png_structp)png_ptr, (png_voidp)ptr);
}
void
png_reset_crc(png_structp png_ptr)
{
	png_ptr->crc = crc32(0, Z_NULL, 0);
}
void
png_calculate_crc(png_structp png_ptr, png_bytep ptr, png_size_t length)
{
	int need_crc = 1;
	if(png_ptr->chunk_name[0] & 0x20)
	{
		if((png_ptr->flags & PNG_FLAG_CRC_ANCILLARY_MASK) ==
			(PNG_FLAG_CRC_ANCILLARY_USE | PNG_FLAG_CRC_ANCILLARY_NOWARN))
			need_crc = 0;
	}
	else
	{
		if(png_ptr->flags & PNG_FLAG_CRC_CRITICAL_IGNORE)
			need_crc = 0;
	}
	if(need_crc)
		png_ptr->crc = crc32(png_ptr->crc, ptr, (uInt)length);
}
png_infop PNGAPI
png_create_info_struct(png_structp png_ptr)
{
	png_infop info_ptr;
	png_debug(1, "in png_create_info_struct\n");
	if(png_ptr == NULL) return (NULL);
#ifdef PNG_USER_MEM_SUPPORTED
	info_ptr = (png_infop)png_create_struct_2(PNG_STRUCT_INFO,
		png_ptr->malloc_fn, png_ptr->mem_ptr);
#else
	info_ptr = (png_infop)png_create_struct(PNG_STRUCT_INFO);
#endif
	if(info_ptr != NULL)
		png_info_init_3(&info_ptr, png_sizeof(png_info));
	return (info_ptr);
}
void PNGAPI
png_destroy_info_struct(png_structp png_ptr, png_infopp info_ptr_ptr)
{
	png_infop info_ptr = NULL;
	png_debug(1, "in png_destroy_info_struct\n");
	if(info_ptr_ptr != NULL)
		info_ptr = *info_ptr_ptr;
	if(info_ptr != NULL)
	{
		png_info_destroy(png_ptr, info_ptr);
#ifdef PNG_USER_MEM_SUPPORTED
		png_destroy_struct_2((png_voidp)info_ptr, png_ptr->free_fn,
			png_ptr->mem_ptr);
#else
		png_destroy_struct((png_voidp)info_ptr);
#endif
		*info_ptr_ptr = NULL;
	}
}
#if defined(PNG_1_0_X) || defined (PNG_1_2_X)
#undef png_info_init
void PNGAPI
png_info_init(png_infop info_ptr)
{
	png_info_init_3(&info_ptr, 0);
}
#endif
void PNGAPI
png_info_init_3(png_infopp ptr_ptr, png_size_t png_info_struct_size)
{
	png_infop info_ptr = *ptr_ptr;
	png_debug(1, "in png_info_init_3\n");
	if(png_sizeof(png_info) > png_info_struct_size)
	{
		png_destroy_struct(info_ptr);
		info_ptr = (png_infop)png_create_struct(PNG_STRUCT_INFO);
		*ptr_ptr = info_ptr;
	}
	png_memset(info_ptr, 0, png_sizeof (png_info));
}
#ifdef PNG_FREE_ME_SUPPORTED
void PNGAPI
png_data_freer(png_structp png_ptr, png_infop info_ptr,
			   int freer, png_uint_32 mask)
{
	png_debug(1, "in png_data_freer\n");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(freer == PNG_DESTROY_WILL_FREE_DATA)
		info_ptr->free_me |= mask;
	else if(freer == PNG_USER_WILL_FREE_DATA)
		info_ptr->free_me &= ~mask;
	else
		png_warning(png_ptr,
		"Unknown freer parameter in png_data_freer.");
}
#endif
void PNGAPI
png_free_data(png_structp png_ptr, png_infop info_ptr, png_uint_32 mask,
			  int num)
{
	png_debug(1, "in png_free_data\n");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
#if defined(PNG_TEXT_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_TEXT) & info_ptr->free_me)
#else
	if(mask & PNG_FREE_TEXT)
#endif
	{
		if(num != -1)
		{
			if(info_ptr->text && info_ptr->text[num].key)
			{
				png_free(png_ptr, info_ptr->text[num].key);
				info_ptr->text[num].key = NULL;
			}
		}
		else
		{
			int i;
			for(i = 0; i < info_ptr->num_text; i++)
				png_free_data(png_ptr, info_ptr, PNG_FREE_TEXT, i);
			png_free(png_ptr, info_ptr->text);
			info_ptr->text = NULL;
			info_ptr->num_text = 0;
		}
	}
#endif
#if defined(PNG_tRNS_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_TRNS) & info_ptr->free_me)
#else
	if((mask & PNG_FREE_TRNS) && (png_ptr->flags & PNG_FLAG_FREE_TRNS))
#endif
	{
		png_free(png_ptr, info_ptr->trans);
		info_ptr->valid &= ~PNG_INFO_tRNS;
#ifndef PNG_FREE_ME_SUPPORTED
		png_ptr->flags &= ~PNG_FLAG_FREE_TRNS;
#endif
		info_ptr->trans = NULL;
	}
#endif
#if defined(PNG_sCAL_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_SCAL) & info_ptr->free_me)
#else
	if(mask & PNG_FREE_SCAL)
#endif
	{
#if defined(PNG_FIXED_POINT_SUPPORTED) && !defined(PNG_FLOATING_POINT_SUPPORTED)
		png_free(png_ptr, info_ptr->scal_s_width);
		png_free(png_ptr, info_ptr->scal_s_height);
		info_ptr->scal_s_width = NULL;
		info_ptr->scal_s_height = NULL;
#endif
		info_ptr->valid &= ~PNG_INFO_sCAL;
	}
#endif
#if defined(PNG_pCAL_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_PCAL) & info_ptr->free_me)
#else
	if(mask & PNG_FREE_PCAL)
#endif
	{
		png_free(png_ptr, info_ptr->pcal_purpose);
		png_free(png_ptr, info_ptr->pcal_units);
		info_ptr->pcal_purpose = NULL;
		info_ptr->pcal_units = NULL;
		if(info_ptr->pcal_params != NULL)
		{
			int i;
			for(i = 0; i < (int)info_ptr->pcal_nparams; i++)
			{
				png_free(png_ptr, info_ptr->pcal_params[i]);
				info_ptr->pcal_params[i] = NULL;
			}
			png_free(png_ptr, info_ptr->pcal_params);
			info_ptr->pcal_params = NULL;
		}
		info_ptr->valid &= ~PNG_INFO_pCAL;
	}
#endif
#if defined(PNG_iCCP_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_ICCP) & info_ptr->free_me)
#else
	if(mask & PNG_FREE_ICCP)
#endif
	{
		png_free(png_ptr, info_ptr->iccp_name);
		png_free(png_ptr, info_ptr->iccp_profile);
		info_ptr->iccp_name = NULL;
		info_ptr->iccp_profile = NULL;
		info_ptr->valid &= ~PNG_INFO_iCCP;
	}
#endif
#if defined(PNG_sPLT_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_SPLT) & info_ptr->free_me)
#else
	if(mask & PNG_FREE_SPLT)
#endif
	{
		if(num != -1)
		{
			if(info_ptr->splt_palettes)
			{
				png_free(png_ptr, info_ptr->splt_palettes[num].name);
				png_free(png_ptr, info_ptr->splt_palettes[num].entries);
				info_ptr->splt_palettes[num].name = NULL;
				info_ptr->splt_palettes[num].entries = NULL;
			}
		}
		else
		{
			if(info_ptr->splt_palettes_num)
			{
				int i;
				for(i = 0; i < (int)info_ptr->splt_palettes_num; i++)
					png_free_data(png_ptr, info_ptr, PNG_FREE_SPLT, i);
				png_free(png_ptr, info_ptr->splt_palettes);
				info_ptr->splt_palettes = NULL;
				info_ptr->splt_palettes_num = 0;
			}
			info_ptr->valid &= ~PNG_INFO_sPLT;
		}
	}
#endif
#if defined(PNG_UNKNOWN_CHUNKS_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_UNKN) & info_ptr->free_me)
#else
	if(mask & PNG_FREE_UNKN)
#endif
	{
		if(num != -1)
		{
			if(info_ptr->unknown_chunks)
			{
				png_free(png_ptr, info_ptr->unknown_chunks[num].data);
				info_ptr->unknown_chunks[num].data = NULL;
			}
		}
		else
		{
			int i;
			if(info_ptr->unknown_chunks_num)
			{
				for(i = 0; i < (int)info_ptr->unknown_chunks_num; i++)
					png_free_data(png_ptr, info_ptr, PNG_FREE_UNKN, i);
				png_free(png_ptr, info_ptr->unknown_chunks);
				info_ptr->unknown_chunks = NULL;
				info_ptr->unknown_chunks_num = 0;
			}
		}
	}
#endif
#if defined(PNG_hIST_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_HIST)  & info_ptr->free_me)
#else
	if((mask & PNG_FREE_HIST) && (png_ptr->flags & PNG_FLAG_FREE_HIST))
#endif
	{
		png_free(png_ptr, info_ptr->hist);
		info_ptr->hist = NULL;
		info_ptr->valid &= ~PNG_INFO_hIST;
#ifndef PNG_FREE_ME_SUPPORTED
		png_ptr->flags &= ~PNG_FLAG_FREE_HIST;
#endif
	}
#endif
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_PLTE) & info_ptr->free_me)
#else
	if((mask & PNG_FREE_PLTE) && (png_ptr->flags & PNG_FLAG_FREE_PLTE))
#endif
	{
		png_zfree(png_ptr, info_ptr->palette);
		info_ptr->palette = NULL;
		info_ptr->valid &= ~PNG_INFO_PLTE;
#ifndef PNG_FREE_ME_SUPPORTED
		png_ptr->flags &= ~PNG_FLAG_FREE_PLTE;
#endif
		info_ptr->num_palette = 0;
	}
#if defined(PNG_INFO_IMAGE_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if((mask & PNG_FREE_ROWS) & info_ptr->free_me)
#else
	if(mask & PNG_FREE_ROWS)
#endif
	{
		if(info_ptr->row_pointers)
		{
			int row;
			for(row = 0; row < (int)info_ptr->height; row++)
			{
				png_free(png_ptr, info_ptr->row_pointers[row]);
				info_ptr->row_pointers[row] = NULL;
			}
			png_free(png_ptr, info_ptr->row_pointers);
			info_ptr->row_pointers = NULL;
		}
		info_ptr->valid &= ~PNG_INFO_IDAT;
	}
#endif
#ifdef PNG_FREE_ME_SUPPORTED
	if(num == -1)
		info_ptr->free_me &= ~mask;
	else
		info_ptr->free_me &= ~(mask & ~PNG_FREE_MUL);
#endif
}
void
png_info_destroy(png_structp png_ptr, png_infop info_ptr)
{
	png_debug(1, "in png_info_destroy\n");
	png_free_data(png_ptr, info_ptr, PNG_FREE_ALL, -1);
#if defined(PNG_UNKNOWN_CHUNKS_SUPPORTED)
	if(png_ptr->num_chunk_list)
	{
		png_free(png_ptr, png_ptr->chunk_list);
		png_ptr->chunk_list = NULL;
		png_ptr->num_chunk_list = 0;
	}
#endif
	png_info_init_3(&info_ptr, png_sizeof(png_info));
}
png_voidp PNGAPI
png_get_io_ptr(png_structp png_ptr)
{
	return (png_ptr->io_ptr);
}
#if !defined(PNG_NO_STDIO)
void PNGAPI
png_init_io(png_structp png_ptr, png_FILE_p fp)
{
	png_debug(1, "in png_init_io\n");
	png_ptr->io_ptr = (png_voidp)fp;
}
#endif
#if defined(PNG_TIME_RFC1123_SUPPORTED)
png_charp PNGAPI
png_convert_to_rfc1123(png_structp png_ptr, png_timep ptime)
{
	static PNG_CONST char short_months[12][4] =
	{"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	if(png_ptr->time_buffer == NULL)
	{
		png_ptr->time_buffer = (png_charp)png_malloc(png_ptr, (png_uint_32)(29*
			png_sizeof(char)));
	}
#if defined(_WIN32_WCE)
	{
		wchar_t time_buf[29];
		wsprintf(time_buf, TEXT("%d %S %d %02d:%02d:%02d +0000"),
			ptime->day % 32, short_months[(ptime->month - 1) % 12],
			ptime->year, ptime->hour % 24, ptime->minute % 60,
			ptime->second % 61);
		WideCharToMultiByte(CP_ACP, 0, time_buf, -1, png_ptr->time_buffer, 29,
			NULL, NULL);
	}
#else
#ifdef USE_FAR_KEYWORD
	{
		char near_time_buf[29];
		sprintf(near_time_buf, "%d %s %d %02d:%02d:%02d +0000",
			ptime->day % 32, short_months[(ptime->month - 1) % 12],
			ptime->year, ptime->hour % 24, ptime->minute % 60,
			ptime->second % 61);
		png_memcpy(png_ptr->time_buffer, near_time_buf,
			29*png_sizeof(char));
	}
#else
	sprintf(png_ptr->time_buffer, "%d %s %d %02d:%02d:%02d +0000",
		ptime->day % 32, short_months[(ptime->month - 1) % 12],
		ptime->year, ptime->hour % 24, ptime->minute % 60,
		ptime->second % 61);
#endif
#endif
	return ((png_charp)png_ptr->time_buffer);
}
#endif
#if 0
png_bytep PNGAPI
png_sig_bytes()
{
	return ((png_bytep)"\211\120\116\107\015\012\032\012");
}
#endif
png_charp PNGAPI
png_get_copyright(png_structp png_ptr)
{
	if(&png_ptr != NULL)
		return ((png_charp) "\n libpng version 1.2.8 - December 3, 2004\n\
							Copyright (c) 1998-2004 Glenn Randers-Pehrson\n\
							Copyright (c) 1996-1997 Andreas Dilger\n\
							Copyright (c) 1995-1996 Guy Eric Schalnat, Group 42, Inc.\n");
	return ((png_charp) "");
}
png_charp PNGAPI
png_get_libpng_ver(png_structp png_ptr)
{
	if(&png_ptr != NULL)
		return ((png_charp) PNG_LIBPNG_VER_STRING);
	return ((png_charp) "");
}
png_charp PNGAPI
png_get_header_ver(png_structp png_ptr)
{
	if(&png_ptr != NULL)
		return ((png_charp) PNG_LIBPNG_VER_STRING);
	return ((png_charp) "");
}
png_charp PNGAPI
png_get_header_version(png_structp png_ptr)
{
	if(&png_ptr != NULL)
		return ((png_charp) PNG_HEADER_VERSION_STRING);
	return ((png_charp) "");
}
#ifdef PNG_HANDLE_AS_UNKNOWN_SUPPORTED
int PNGAPI
png_handle_as_unknown(png_structp png_ptr, png_bytep chunk_name)
{
	int i;
	png_bytep p;
	if((png_ptr == NULL && chunk_name == NULL) || png_ptr->num_chunk_list <= 0)
		return 0;
	p = png_ptr->chunk_list+png_ptr->num_chunk_list*5-5;
	for(i = png_ptr->num_chunk_list; i; i--, p -= 5)
		if(!png_memcmp(chunk_name, p, 4))
			return ((int)*(p+4));
	return 0;
}
#endif
int PNGAPI
png_reset_zstream(png_structp png_ptr)
{
	return (inflateReset(&png_ptr->zstream));
}
png_uint_32 PNGAPI
png_access_version_number()
{
	return((png_uint_32) PNG_LIBPNG_VER);
}
#if !defined(PNG_1_0_X)
#if defined(PNG_ASSEMBLER_CODE_SUPPORTED)
void
png_init_mmx_flags (png_structp png_ptr)
{
	png_ptr->mmx_rowbytes_threshold = 0;
	png_ptr->mmx_bitdepth_threshold = 0;
#  if(defined(PNG_USE_PNGVCRD) || defined(PNG_USE_PNGGCCRD))
	png_ptr->asm_flags |= PNG_ASM_FLAG_MMX_SUPPORT_COMPILED;
	if(png_mmx_support() > 0)
	{
		png_ptr->asm_flags |= PNG_ASM_FLAG_MMX_SUPPORT_IN_CPU
#    ifdef PNG_HAVE_ASSEMBLER_COMBINE_ROW
			| PNG_ASM_FLAG_MMX_READ_COMBINE_ROW
#    endif
#    ifdef PNG_HAVE_ASSEMBLER_READ_INTERLACE
			| PNG_ASM_FLAG_MMX_READ_INTERLACE
#    endif
#    ifndef PNG_HAVE_ASSEMBLER_READ_FILTER_ROW
			;
#    else
			| PNG_ASM_FLAG_MMX_READ_FILTER_SUB
			| PNG_ASM_FLAG_MMX_READ_FILTER_UP
			| PNG_ASM_FLAG_MMX_READ_FILTER_AVG
			| PNG_ASM_FLAG_MMX_READ_FILTER_PAETH ;
		png_ptr->mmx_rowbytes_threshold = PNG_MMX_ROWBYTES_THRESHOLD_DEFAULT;
		png_ptr->mmx_bitdepth_threshold = PNG_MMX_BITDEPTH_THRESHOLD_DEFAULT;
#    endif
	} else
	{
		png_ptr->asm_flags &= ~( PNG_ASM_FLAG_MMX_SUPPORT_IN_CPU
			| PNG_MMX_READ_FLAGS
			| PNG_MMX_WRITE_FLAGS );
	}
#  else
	png_ptr->asm_flags &= ~( PNG_MMX_FLAGS );
#  endif
}
#endif
#if !defined(PNG_USE_PNGGCCRD) && \
	!(defined(PNG_ASSEMBLER_CODE_SUPPORTED) && defined(PNG_USE_PNGVCRD))
int PNGAPI
png_mmx_support()
{
	return -1;
}
#endif
#endif
#ifdef PNG_SIZE_T
PNG_EXTERN png_size_t PNGAPI png_convert_size PNGARG((size_t size));
png_size_t PNGAPI
png_convert_size(size_t size)
{
	if(size > (png_size_t)-1)
		PNG_ABORT();
	return ((png_size_t)size);
}
#endif
#define PNG_INTERNAL
#include "png.h"
static void
png_default_error PNGARG((png_structp png_ptr,
						 png_const_charp error_message));
static void
png_default_warning PNGARG((png_structp png_ptr,
						   png_const_charp warning_message));
void PNGAPI
png_error(png_structp png_ptr, png_const_charp error_message)
{
#ifdef PNG_ERROR_NUMBERS_SUPPORTED
	char msg[16];
	if(png_ptr->flags&(PNG_FLAG_STRIP_ERROR_NUMBERS|PNG_FLAG_STRIP_ERROR_TEXT))
	{
		if(*error_message == '#')
		{
			int offset;
			for(offset = 1; offset<15; offset++)
				if(*(error_message+offset) == ' ')
					break;
			if(png_ptr->flags&PNG_FLAG_STRIP_ERROR_TEXT)
			{
				int i;
				for(i = 0; i<offset-1; i++)
					msg[i] = error_message[i+1];
				msg[i]='\0';
				error_message = msg;
			}
			else
				error_message += offset;
		}
		else
		{
			if(png_ptr->flags&PNG_FLAG_STRIP_ERROR_TEXT)
			{
				msg[0]='0';
				msg[1]='\0';
				error_message = msg;
			}
		}
	}
#endif
	if(png_ptr != NULL && png_ptr->error_fn != NULL)
		(*(png_ptr->error_fn))(png_ptr, error_message);
	png_default_error(png_ptr, error_message);
}
void PNGAPI
png_warning(png_structp png_ptr, png_const_charp warning_message)
{
	int offset = 0;
#ifdef PNG_ERROR_NUMBERS_SUPPORTED
	if(png_ptr->flags&(PNG_FLAG_STRIP_ERROR_NUMBERS|PNG_FLAG_STRIP_ERROR_TEXT))
#endif
	{
		if(*warning_message == '#')
		{
			for(offset = 1; offset<15; offset++)
				if(*(warning_message+offset) == ' ')
					break;
		}
	}
	if(png_ptr != NULL && png_ptr->warning_fn != NULL)
		(*(png_ptr->warning_fn))(png_ptr, warning_message+offset);
	else
		png_default_warning(png_ptr, warning_message+offset);
}
#define isnonalpha(c) ((c) < 65 || (c) > 122 || ((c) > 90 && (c) < 97))
static PNG_CONST char png_digit[16] =
{
	'0','1','2','3','4','5','6','7','8','9',
		'A','B','C','D','E','F'
};
static void
png_format_buffer(png_structp png_ptr, png_charp buffer, png_const_charp
				  error_message)
{
	int iout = 0, iin = 0;
	while(iin < 4)
	{
		int c = png_ptr->chunk_name[iin++];
		if(isnonalpha(c))
		{
			buffer[iout++] = '[';
			buffer[iout++] = png_digit[(c & 0xf0) >> 4];
			buffer[iout++] = png_digit[c & 0x0f];
			buffer[iout++] = ']';
		}
		else
		{
			buffer[iout++] = (png_byte)c;
		}
	}
	if(error_message == NULL)
		buffer[iout] = 0;
	else
	{
		buffer[iout++] = ':';
		buffer[iout++] = ' ';
		png_strncpy(buffer+iout, error_message, 63);
		buffer[iout+63] = 0;
	}
}
void PNGAPI
png_chunk_error(png_structp png_ptr, png_const_charp error_message)
{
	char msg[18+64];
	png_format_buffer(png_ptr, msg, error_message);
	png_error(png_ptr, msg);
}
void PNGAPI
png_chunk_warning(png_structp png_ptr, png_const_charp warning_message)
{
	char msg[18+64];
	png_format_buffer(png_ptr, msg, warning_message);
	png_warning(png_ptr, msg);
}
static void
png_default_error(png_structp png_ptr, png_const_charp error_message)
{
#ifndef PNG_NO_CONSOLE_IO
#ifdef PNG_ERROR_NUMBERS_SUPPORTED
	if(*error_message == '#')
	{
		int offset;
		char error_number[16];
		for(offset = 0; offset<15; offset++)
		{
			error_number[offset] = *(error_message+offset+1);
			if(*(error_message+offset) == ' ')
				break;
		}
		if((offset > 1) && (offset < 15))
		{
			error_number[offset-1]='\0';
			fprintf(stderr, "libpng error no. %s: %s\n", error_number,
				error_message+offset);
		}
		else
			fprintf(stderr, "libpng error: %s, offset = %d\n", error_message,offset);
	}
	else
#endif
		fprintf(stderr, "libpng error: %s\n", error_message);
#endif
#ifdef PNG_SETJMP_SUPPORTED
#  ifdef USE_FAR_KEYWORD
	{
		jmp_buf jmpbuf;
		png_memcpy(jmpbuf,png_ptr->jmpbuf,png_sizeof(jmp_buf));
		longjmp(jmpbuf, 1);
	}
#  else
	longjmp(png_ptr->jmpbuf, 1);
# endif
#else
	;
	if(png_ptr)
		PNG_ABORT();
#endif
#ifdef PNG_NO_CONSOLE_IO
	;
	if(&error_message != NULL)
		return;
#endif
}
static void
png_default_warning(png_structp png_ptr, png_const_charp warning_message)
{
#ifndef PNG_NO_CONSOLE_IO
#  ifdef PNG_ERROR_NUMBERS_SUPPORTED
	if(*warning_message == '#')
	{
		int offset;
		char warning_number[16];
		for(offset = 0; offset<15; offset++)
		{
			warning_number[offset] = *(warning_message+offset+1);
			if(*(warning_message+offset) == ' ')
				break;
		}
		if((offset > 1) && (offset < 15))
		{
			warning_number[offset-1]='\0';
			fprintf(stderr, "libpng warning no. %s: %s\n", warning_number,
				warning_message+offset);
		}
		else
			fprintf(stderr, "libpng warning: %s\n", warning_message);
	}
	else
#  endif
		fprintf(stderr, "libpng warning: %s\n", warning_message);
#else
	;
	if(warning_message)
		return;
#endif
	;
	if(png_ptr)
		return;
}
void PNGAPI
png_set_error_fn(png_structp png_ptr, png_voidp error_ptr,
				 png_error_ptr error_fn, png_error_ptr warning_fn)
{
	png_ptr->error_ptr = error_ptr;
	png_ptr->error_fn = error_fn;
	png_ptr->warning_fn = warning_fn;
}
png_voidp PNGAPI
png_get_error_ptr(png_structp png_ptr)
{
	return ((png_voidp)png_ptr->error_ptr);
}
#ifdef PNG_ERROR_NUMBERS_SUPPORTED
void PNGAPI
png_set_strip_error_numbers(png_structp png_ptr, png_uint_32 strip_mode)
{
	if(png_ptr != NULL)
	{
		png_ptr->flags &=
			((~(PNG_FLAG_STRIP_ERROR_NUMBERS|PNG_FLAG_STRIP_ERROR_TEXT))&strip_mode);
	}
}
#endif
#define PNG_INTERNAL
#include "png.h"
#ifdef PNG_WRITE_SUPPORTED
void
png_do_write_transformations(png_structp png_ptr)
{
	png_debug(1, "in png_do_write_transformations\n");
	if(png_ptr == NULL)
		return;
#if defined(PNG_WRITE_USER_TRANSFORM_SUPPORTED)
	if(png_ptr->transformations & PNG_USER_TRANSFORM)
		if(png_ptr->write_user_transform_fn != NULL)
			(*(png_ptr->write_user_transform_fn))
			(png_ptr,
			&(png_ptr->row_info),
			png_ptr->row_buf + 1);
#endif
#if defined(PNG_WRITE_FILLER_SUPPORTED)
	if(png_ptr->transformations & PNG_FILLER)
		png_do_strip_filler(&(png_ptr->row_info), png_ptr->row_buf + 1,
		png_ptr->flags);
#endif
#if defined(PNG_WRITE_PACKSWAP_SUPPORTED)
	if(png_ptr->transformations & PNG_PACKSWAP)
		png_do_packswap(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_WRITE_PACK_SUPPORTED)
	if(png_ptr->transformations & PNG_PACK)
		png_do_pack(&(png_ptr->row_info), png_ptr->row_buf + 1,
		(png_uint_32)png_ptr->bit_depth);
#endif
#if defined(PNG_WRITE_SWAP_SUPPORTED)
	if(png_ptr->transformations & PNG_SWAP_BYTES)
		png_do_swap(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_WRITE_SHIFT_SUPPORTED)
	if(png_ptr->transformations & PNG_SHIFT)
		png_do_shift(&(png_ptr->row_info), png_ptr->row_buf + 1,
		&(png_ptr->shift));
#endif
#if defined(PNG_WRITE_INVERT_ALPHA_SUPPORTED)
	if(png_ptr->transformations & PNG_INVERT_ALPHA)
		png_do_write_invert_alpha(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_WRITE_SWAP_ALPHA_SUPPORTED)
	if(png_ptr->transformations & PNG_SWAP_ALPHA)
		png_do_write_swap_alpha(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_WRITE_BGR_SUPPORTED)
	if(png_ptr->transformations & PNG_BGR)
		png_do_bgr(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_WRITE_INVERT_SUPPORTED)
	if(png_ptr->transformations & PNG_INVERT_MONO)
		png_do_invert(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
}
#if defined(PNG_WRITE_PACK_SUPPORTED)
void
png_do_pack(png_row_infop row_info, png_bytep row, png_uint_32 bit_depth)
{
	png_debug(1, "in png_do_pack\n");
	if(row_info->bit_depth == 8 &&
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		row_info->channels == 1)
	{
		switch ((int)bit_depth)
		{
		case 1:
			{
				png_bytep sp, dp;
				int mask, v;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				sp = row;
				dp = row;
				mask = 0x80;
				v = 0;
				for(i = 0; i < row_width; i++)
				{
					if(*sp != 0)
						v |= mask;
					sp++;
					if(mask > 1)
						mask >>= 1;
					else
					{
						mask = 0x80;
						*dp = (png_byte)v;
						dp++;
						v = 0;
					}
				}
				if(mask != 0x80)
					*dp = (png_byte)v;
				break;
			}
		case 2:
			{
				png_bytep sp, dp;
				int shift, v;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				sp = row;
				dp = row;
				shift = 6;
				v = 0;
				for(i = 0; i < row_width; i++)
				{
					png_byte value;
					value = (png_byte)(*sp & 0x03);
					v |= (value << shift);
					if(shift == 0)
					{
						shift = 6;
						*dp = (png_byte)v;
						dp++;
						v = 0;
					}
					else
						shift -= 2;
					sp++;
				}
				if(shift != 6)
					*dp = (png_byte)v;
				break;
			}
		case 4:
			{
				png_bytep sp, dp;
				int shift, v;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				sp = row;
				dp = row;
				shift = 4;
				v = 0;
				for(i = 0; i < row_width; i++)
				{
					png_byte value;
					value = (png_byte)(*sp & 0x0f);
					v |= (value << shift);
					if(shift == 0)
					{
						shift = 4;
						*dp = (png_byte)v;
						dp++;
						v = 0;
					}
					else
						shift -= 4;
					sp++;
				}
				if(shift != 4)
					*dp = (png_byte)v;
				break;
			}
		}
		row_info->bit_depth = (png_byte)bit_depth;
		row_info->pixel_depth = (png_byte)(bit_depth * row_info->channels);
		row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,
			row_info->width);
	}
}
#endif
#if defined(PNG_WRITE_SHIFT_SUPPORTED)
void
png_do_shift(png_row_infop row_info, png_bytep row, png_color_8p bit_depth)
{
	png_debug(1, "in png_do_shift\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL &&
#else
	if(
#endif
		row_info->color_type != PNG_COLOR_TYPE_PALETTE)
	{
		int shift_start[4], shift_dec[4];
		int channels = 0;
		if(row_info->color_type & PNG_COLOR_MASK_COLOR)
		{
			shift_start[channels] = row_info->bit_depth - bit_depth->red;
			shift_dec[channels] = bit_depth->red;
			channels++;
			shift_start[channels] = row_info->bit_depth - bit_depth->green;
			shift_dec[channels] = bit_depth->green;
			channels++;
			shift_start[channels] = row_info->bit_depth - bit_depth->blue;
			shift_dec[channels] = bit_depth->blue;
			channels++;
		}
		else
		{
			shift_start[channels] = row_info->bit_depth - bit_depth->gray;
			shift_dec[channels] = bit_depth->gray;
			channels++;
		}
		if(row_info->color_type & PNG_COLOR_MASK_ALPHA)
		{
			shift_start[channels] = row_info->bit_depth - bit_depth->alpha;
			shift_dec[channels] = bit_depth->alpha;
			channels++;
		}
		if(row_info->bit_depth < 8)
		{
			png_bytep bp = row;
			png_uint_32 i;
			png_byte mask;
			png_uint_32 row_bytes = row_info->rowbytes;
			if(bit_depth->gray == 1 && row_info->bit_depth == 2)
				mask = 0x55;
			else if(row_info->bit_depth == 4 && bit_depth->gray == 3)
				mask = 0x11;
			else
				mask = 0xff;
			for(i = 0; i < row_bytes; i++, bp++)
			{
				png_uint_16 v;
				int j;
				v = *bp;
				*bp = 0;
				for(j = shift_start[0]; j > -shift_dec[0]; j -= shift_dec[0])
				{
					if(j > 0)
						*bp |= (png_byte)((v << j) & 0xff);
					else
						*bp |= (png_byte)((v >> (-j)) & mask);
				}
			}
		}
		else if(row_info->bit_depth == 8)
		{
			png_bytep bp = row;
			png_uint_32 i;
			png_uint_32 istop = channels * row_info->width;
			for(i = 0; i < istop; i++, bp++)
			{
				png_uint_16 v;
				int j;
				int c = (int)(i%channels);
				v = *bp;
				*bp = 0;
				for(j = shift_start[c]; j > -shift_dec[c]; j -= shift_dec[c])
				{
					if(j > 0)
						*bp |= (png_byte)((v << j) & 0xff);
					else
						*bp |= (png_byte)((v >> (-j)) & 0xff);
				}
			}
		}
		else
		{
			png_bytep bp;
			png_uint_32 i;
			png_uint_32 istop = channels * row_info->width;
			for(bp = row, i = 0; i < istop; i++)
			{
				int c = (int)(i%channels);
				png_uint_16 value, v;
				int j;
				v = (png_uint_16)(((png_uint_16)(*bp) << 8) + *(bp + 1));
				value = 0;
				for(j = shift_start[c]; j > -shift_dec[c]; j -= shift_dec[c])
				{
					if(j > 0)
						value |= (png_uint_16)((v << j) & (png_uint_16)0xffff);
					else
						value |= (png_uint_16)((v >> (-j)) & (png_uint_16)0xffff);
				}
				*bp++ = (png_byte)(value >> 8);
				*bp++ = (png_byte)(value & 0xff);
			}
		}
	}
}
#endif
#if defined(PNG_WRITE_SWAP_ALPHA_SUPPORTED)
void
png_do_write_swap_alpha(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_write_swap_alpha\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL)
#endif
	{
		if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					png_byte save = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = save;
				}
			}
			else
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					png_byte save[2];
					save[0] = *(sp++);
					save[1] = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = save[0];
					*(dp++) = save[1];
				}
			}
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					png_byte save = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = save;
				}
			}
			else
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					png_byte save[2];
					save[0] = *(sp++);
					save[1] = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = save[0];
					*(dp++) = save[1];
				}
			}
		}
	}
}
#endif
#if defined(PNG_WRITE_INVERT_ALPHA_SUPPORTED)
void
png_do_write_invert_alpha(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_write_invert_alpha\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL)
#endif
	{
		if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = (png_byte)(255 - *(sp++));
				}
			}
			else
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = (png_byte)(255 - *(sp++));
					*(dp++) = (png_byte)(255 - *(sp++));
				}
			}
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					*(dp++) = *(sp++);
					*(dp++) = (png_byte)(255 - *(sp++));
				}
			}
			else
			{
				png_bytep sp, dp;
				png_uint_32 i;
				png_uint_32 row_width = row_info->width;
				for(i = 0, sp = dp = row; i < row_width; i++)
				{
					*(dp++) = *(sp++);
					*(dp++) = *(sp++);
					*(dp++) = (png_byte)(255 - *(sp++));
					*(dp++) = (png_byte)(255 - *(sp++));
				}
			}
		}
	}
}
#endif
#if defined(PNG_MNG_FEATURES_SUPPORTED)
void
png_do_write_intrapixel(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_write_intrapixel\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		(row_info->color_type & PNG_COLOR_MASK_COLOR))
	{
		int bytes_per_pixel;
		png_uint_32 row_width = row_info->width;
		if(row_info->bit_depth == 8)
		{
			png_bytep rp;
			png_uint_32 i;
			if(row_info->color_type == PNG_COLOR_TYPE_RGB)
				bytes_per_pixel = 3;
			else if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
				bytes_per_pixel = 4;
			else
				return;
			for(i = 0, rp = row; i < row_width; i++, rp += bytes_per_pixel)
			{
				*(rp)   = (png_byte)((*rp     - *(rp+1))&0xff);
				*(rp+2) = (png_byte)((*(rp+2) - *(rp+1))&0xff);
			}
		}
		else if(row_info->bit_depth == 16)
		{
			png_bytep rp;
			png_uint_32 i;
			if(row_info->color_type == PNG_COLOR_TYPE_RGB)
				bytes_per_pixel = 6;
			else if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
				bytes_per_pixel = 8;
			else
				return;
			for(i = 0, rp = row; i < row_width; i++, rp += bytes_per_pixel)
			{
				png_uint_32 s0   = (*(rp  ) << 8) | *(rp+1);
				png_uint_32 s1   = (*(rp+2) << 8) | *(rp+3);
				png_uint_32 s2   = (*(rp+4) << 8) | *(rp+5);
				png_uint_32 red  = (png_uint_32)((s0-s1) & 0xffffL);
				png_uint_32 blue = (png_uint_32)((s2-s1) & 0xffffL);
				*(rp  ) = (png_byte)((red >> 8) & 0xff);
				*(rp+1) = (png_byte)(red & 0xff);
				*(rp+4) = (png_byte)((blue >> 8) & 0xff);
				*(rp+5) = (png_byte)(blue & 0xff);
			}
		}
	}
}
#endif
#endif
#define PNG_INTERNAL
#include "png.h"
png_uint_32 PNGAPI
png_get_valid(png_structp png_ptr, png_infop info_ptr, png_uint_32 flag)
{
	if(png_ptr != NULL && info_ptr != NULL)
		return(info_ptr->valid & flag);
	else
		return(0);
}
png_uint_32 PNGAPI
png_get_rowbytes(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
		return(info_ptr->rowbytes);
	else
		return(0);
}
#if defined(PNG_INFO_IMAGE_SUPPORTED)
png_bytepp PNGAPI
png_get_rows(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
		return(info_ptr->row_pointers);
	else
		return(0);
}
#endif
#ifdef PNG_EASY_ACCESS_SUPPORTED
png_uint_32 PNGAPI
png_get_image_width(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
	{
		return info_ptr->width;
	}
	return (0);
}
png_uint_32 PNGAPI
png_get_image_height(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
	{
		return info_ptr->height;
	}
	return (0);
}
png_byte PNGAPI
png_get_bit_depth(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
	{
		return info_ptr->bit_depth;
	}
	return (0);
}
png_byte PNGAPI
png_get_color_type(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
	{
		return info_ptr->color_type;
	}
	return (0);
}
png_byte PNGAPI
png_get_filter_type(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
	{
		return info_ptr->filter_type;
	}
	return (0);
}
png_byte PNGAPI
png_get_interlace_type(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
	{
		return info_ptr->interlace_type;
	}
	return (0);
}
png_byte PNGAPI
png_get_compression_type(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
	{
		return info_ptr->compression_type;
	}
	return (0);
}
png_uint_32 PNGAPI
png_get_x_pixels_per_meter(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_pHYs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_pHYs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_x_pixels_per_meter");
			if(info_ptr->phys_unit_type != PNG_RESOLUTION_METER)
				return (0);
			else return (info_ptr->x_pixels_per_unit);
		}
#else
		return (0);
#endif
	return (0);
}
png_uint_32 PNGAPI
png_get_y_pixels_per_meter(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_pHYs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_pHYs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_y_pixels_per_meter");
			if(info_ptr->phys_unit_type != PNG_RESOLUTION_METER)
				return (0);
			else return (info_ptr->y_pixels_per_unit);
		}
#else
		return (0);
#endif
	return (0);
}
png_uint_32 PNGAPI
png_get_pixels_per_meter(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_pHYs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_pHYs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_pixels_per_meter");
			if(info_ptr->phys_unit_type != PNG_RESOLUTION_METER ||
				info_ptr->x_pixels_per_unit != info_ptr->y_pixels_per_unit)
				return (0);
			else return (info_ptr->x_pixels_per_unit);
		}
#else
		return (0);
#endif
	return (0);
}
#ifdef PNG_FLOATING_POINT_SUPPORTED
float PNGAPI
png_get_pixel_aspect_ratio(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_pHYs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_pHYs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_aspect_ratio");
			if(info_ptr->x_pixels_per_unit == 0)
				return ((float)0.0);
			else
				return ((float)((float)info_ptr->y_pixels_per_unit
				/(float)info_ptr->x_pixels_per_unit));
		}
#else
		return (0.0);
#endif
	return ((float)0.0);
}
#endif
png_int_32 PNGAPI
png_get_x_offset_microns(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_oFFs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_oFFs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_x_offset_microns");
			if(info_ptr->offset_unit_type != PNG_OFFSET_MICROMETER)
				return (0);
			else return (info_ptr->x_offset);
		}
#else
		return (0);
#endif
	return (0);
}
png_int_32 PNGAPI
png_get_y_offset_microns(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_oFFs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_oFFs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_y_offset_microns");
			if(info_ptr->offset_unit_type != PNG_OFFSET_MICROMETER)
				return (0);
			else return (info_ptr->y_offset);
		}
#else
		return (0);
#endif
	return (0);
}
png_int_32 PNGAPI
png_get_x_offset_pixels(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_oFFs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_oFFs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_x_offset_microns");
			if(info_ptr->offset_unit_type != PNG_OFFSET_PIXEL)
				return (0);
			else return (info_ptr->x_offset);
		}
#else
		return (0);
#endif
	return (0);
}
png_int_32 PNGAPI
png_get_y_offset_pixels(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
#if defined(PNG_oFFs_SUPPORTED)
		if(info_ptr->valid & PNG_INFO_oFFs)
		{
			png_debug1(1, "in %s retrieval function\n", "png_get_y_offset_microns");
			if(info_ptr->offset_unit_type != PNG_OFFSET_PIXEL)
				return (0);
			else return (info_ptr->y_offset);
		}
#else
		return (0);
#endif
	return (0);
}
#if defined(PNG_INCH_CONVERSIONS) && defined(PNG_FLOATING_POINT_SUPPORTED)
png_uint_32 PNGAPI
png_get_pixels_per_inch(png_structp png_ptr, png_infop info_ptr)
{
	return ((png_uint_32)((float)png_get_pixels_per_meter(png_ptr, info_ptr)
		*.0254 +.5));
}
png_uint_32 PNGAPI
png_get_x_pixels_per_inch(png_structp png_ptr, png_infop info_ptr)
{
	return ((png_uint_32)((float)png_get_x_pixels_per_meter(png_ptr, info_ptr)
		*.0254 +.5));
}
png_uint_32 PNGAPI
png_get_y_pixels_per_inch(png_structp png_ptr, png_infop info_ptr)
{
	return ((png_uint_32)((float)png_get_y_pixels_per_meter(png_ptr, info_ptr)
		*.0254 +.5));
}
float PNGAPI
png_get_x_offset_inches(png_structp png_ptr, png_infop info_ptr)
{
	return ((float)png_get_x_offset_microns(png_ptr, info_ptr)
		*.00003937);
}
float PNGAPI
png_get_y_offset_inches(png_structp png_ptr, png_infop info_ptr)
{
	return ((float)png_get_y_offset_microns(png_ptr, info_ptr)
		*.00003937);
}
#if defined(PNG_pHYs_SUPPORTED)
png_uint_32 PNGAPI
png_get_pHYs_dpi(png_structp png_ptr, png_infop info_ptr,
				 png_uint_32 *res_x, png_uint_32 *res_y, int *unit_type)
{
	png_uint_32 retval = 0;
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_pHYs))
	{
		png_debug1(1, "in %s retrieval function\n", "pHYs");
		if(res_x != NULL)
		{
			*res_x = info_ptr->x_pixels_per_unit;
			retval |= PNG_INFO_pHYs;
		}
		if(res_y != NULL)
		{
			*res_y = info_ptr->y_pixels_per_unit;
			retval |= PNG_INFO_pHYs;
		}
		if(unit_type != NULL)
		{
			*unit_type = (int)info_ptr->phys_unit_type;
			retval |= PNG_INFO_pHYs;
			if(*unit_type == 1)
			{
				if(res_x != NULL) *res_x = (png_uint_32)(*res_x * .0254 + .50);
				if(res_y != NULL) *res_y = (png_uint_32)(*res_y * .0254 + .50);
			}
		}
	}
	return (retval);
}
#endif
#endif
#endif
png_byte PNGAPI
png_get_channels(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
		return(info_ptr->channels);
	else
		return (0);
}
png_bytep PNGAPI
png_get_signature(png_structp png_ptr, png_infop info_ptr)
{
	if(png_ptr != NULL && info_ptr != NULL)
		return(info_ptr->signature);
	else
		return (NULL);
}
#if defined(PNG_bKGD_SUPPORTED)
png_uint_32 PNGAPI
png_get_bKGD(png_structp png_ptr, png_infop info_ptr,
			 png_color_16p *background)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_bKGD)
		&& background != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "bKGD");
		*background = &(info_ptr->background);
		return (PNG_INFO_bKGD);
	}
	return (0);
}
#endif
#if defined(PNG_cHRM_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
png_uint_32 PNGAPI
png_get_cHRM(png_structp png_ptr, png_infop info_ptr,
			 double *white_x, double *white_y, double *red_x, double *red_y,
			 double *green_x, double *green_y, double *blue_x, double *blue_y)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_cHRM))
	{
		png_debug1(1, "in %s retrieval function\n", "cHRM");
		if(white_x != NULL)
			*white_x = (double)info_ptr->x_white;
		if(white_y != NULL)
			*white_y = (double)info_ptr->y_white;
		if(red_x != NULL)
			*red_x = (double)info_ptr->x_red;
		if(red_y != NULL)
			*red_y = (double)info_ptr->y_red;
		if(green_x != NULL)
			*green_x = (double)info_ptr->x_green;
		if(green_y != NULL)
			*green_y = (double)info_ptr->y_green;
		if(blue_x != NULL)
			*blue_x = (double)info_ptr->x_blue;
		if(blue_y != NULL)
			*blue_y = (double)info_ptr->y_blue;
		return (PNG_INFO_cHRM);
	}
	return (0);
}
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
png_uint_32 PNGAPI
png_get_cHRM_fixed(png_structp png_ptr, png_infop info_ptr,
				   png_fixed_point *white_x, png_fixed_point *white_y, png_fixed_point *red_x,
				   png_fixed_point *red_y, png_fixed_point *green_x, png_fixed_point *green_y,
				   png_fixed_point *blue_x, png_fixed_point *blue_y)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_cHRM))
	{
		png_debug1(1, "in %s retrieval function\n", "cHRM");
		if(white_x != NULL)
			*white_x = info_ptr->int_x_white;
		if(white_y != NULL)
			*white_y = info_ptr->int_y_white;
		if(red_x != NULL)
			*red_x = info_ptr->int_x_red;
		if(red_y != NULL)
			*red_y = info_ptr->int_y_red;
		if(green_x != NULL)
			*green_x = info_ptr->int_x_green;
		if(green_y != NULL)
			*green_y = info_ptr->int_y_green;
		if(blue_x != NULL)
			*blue_x = info_ptr->int_x_blue;
		if(blue_y != NULL)
			*blue_y = info_ptr->int_y_blue;
		return (PNG_INFO_cHRM);
	}
	return (0);
}
#endif
#endif
#if defined(PNG_gAMA_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
png_uint_32 PNGAPI
png_get_gAMA(png_structp png_ptr, png_infop info_ptr, double *file_gamma)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_gAMA)
		&& file_gamma != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "gAMA");
		*file_gamma = (double)info_ptr->gamma;
		return (PNG_INFO_gAMA);
	}
	return (0);
}
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
png_uint_32 PNGAPI
png_get_gAMA_fixed(png_structp png_ptr, png_infop info_ptr,
				   png_fixed_point *int_file_gamma)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_gAMA)
		&& int_file_gamma != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "gAMA");
		*int_file_gamma = info_ptr->int_gamma;
		return (PNG_INFO_gAMA);
	}
	return (0);
}
#endif
#endif
#if defined(PNG_sRGB_SUPPORTED)
png_uint_32 PNGAPI
png_get_sRGB(png_structp png_ptr, png_infop info_ptr, int *file_srgb_intent)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_sRGB)
		&& file_srgb_intent != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "sRGB");
		*file_srgb_intent = (int)info_ptr->srgb_intent;
		return (PNG_INFO_sRGB);
	}
	return (0);
}
#endif
#if defined(PNG_iCCP_SUPPORTED)
png_uint_32 PNGAPI
png_get_iCCP(png_structp png_ptr, png_infop info_ptr,
			 png_charpp name, int *compression_type,
			 png_charpp profile, png_uint_32 *proflen)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_iCCP)
		&& name != NULL && profile != NULL && proflen != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "iCCP");
		*name = info_ptr->iccp_name;
		*profile = info_ptr->iccp_profile;
		*proflen = (int)info_ptr->iccp_proflen;
		*compression_type = (int)info_ptr->iccp_compression;
		return (PNG_INFO_iCCP);
	}
	return (0);
}
#endif
#if defined(PNG_sPLT_SUPPORTED)
png_uint_32 PNGAPI
png_get_sPLT(png_structp png_ptr, png_infop info_ptr,
			 png_sPLT_tpp spalettes)
{
	if(png_ptr != NULL && info_ptr != NULL && spalettes != NULL)
		*spalettes = info_ptr->splt_palettes;
	return ((png_uint_32)info_ptr->splt_palettes_num);
}
#endif
#if defined(PNG_hIST_SUPPORTED)
png_uint_32 PNGAPI
png_get_hIST(png_structp png_ptr, png_infop info_ptr, png_uint_16p *hist)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_hIST)
		&& hist != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "hIST");
		*hist = info_ptr->hist;
		return (PNG_INFO_hIST);
	}
	return (0);
}
#endif
png_uint_32 PNGAPI
png_get_IHDR(png_structp png_ptr, png_infop info_ptr,
			 png_uint_32 *width, png_uint_32 *height, int *bit_depth,
			 int *color_type, int *interlace_type, int *compression_type,
			 int *filter_type)
{
	if(png_ptr != NULL && info_ptr != NULL && width != NULL && height != NULL &&
		bit_depth != NULL && color_type != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "IHDR");
		*width = info_ptr->width;
		*height = info_ptr->height;
		*bit_depth = info_ptr->bit_depth;
		if(info_ptr->bit_depth < 1 || info_ptr->bit_depth > 16)
			png_error(png_ptr, "Invalid bit depth");
		*color_type = info_ptr->color_type;
		if(info_ptr->color_type > 6)
			png_error(png_ptr, "Invalid color type");
		if(compression_type != NULL)
			*compression_type = info_ptr->compression_type;
		if(filter_type != NULL)
			*filter_type = info_ptr->filter_type;
		if(interlace_type != NULL)
			*interlace_type = info_ptr->interlace_type;
		if(*width == 0 || *width > PNG_UINT_31_MAX)
			png_error(png_ptr, "Invalid image width");
		if(*height == 0 || *height > PNG_UINT_31_MAX)
			png_error(png_ptr, "Invalid image height");
		if(info_ptr->width > (PNG_UINT_32_MAX
			>> 3)
			- 64
			- 1
			- 7*8
			- 8)
		{
			png_warning(png_ptr,
				"Width too large for libpng to process image data.");
		}
		return (1);
	}
	return (0);
}
#if defined(PNG_oFFs_SUPPORTED)
png_uint_32 PNGAPI
png_get_oFFs(png_structp png_ptr, png_infop info_ptr,
			 png_int_32 *offset_x, png_int_32 *offset_y, int *unit_type)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_oFFs)
		&& offset_x != NULL && offset_y != NULL && unit_type != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "oFFs");
		*offset_x = info_ptr->x_offset;
		*offset_y = info_ptr->y_offset;
		*unit_type = (int)info_ptr->offset_unit_type;
		return (PNG_INFO_oFFs);
	}
	return (0);
}
#endif
#if defined(PNG_pCAL_SUPPORTED)
png_uint_32 PNGAPI
png_get_pCAL(png_structp png_ptr, png_infop info_ptr,
			 png_charp *purpose, png_int_32 *X0, png_int_32 *X1, int *type, int *nparams,
			 png_charp *units, png_charpp *params)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_pCAL)
		&& purpose != NULL && X0 != NULL && X1 != NULL && type != NULL &&
		nparams != NULL && units != NULL && params != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "pCAL");
		*purpose = info_ptr->pcal_purpose;
		*X0 = info_ptr->pcal_X0;
		*X1 = info_ptr->pcal_X1;
		*type = (int)info_ptr->pcal_type;
		*nparams = (int)info_ptr->pcal_nparams;
		*units = info_ptr->pcal_units;
		*params = info_ptr->pcal_params;
		return (PNG_INFO_pCAL);
	}
	return (0);
}
#endif
#if defined(PNG_sCAL_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
png_uint_32 PNGAPI
png_get_sCAL(png_structp png_ptr, png_infop info_ptr,
			 int *unit, double *width, double *height)
{
	if(png_ptr != NULL && info_ptr != NULL &&
		(info_ptr->valid & PNG_INFO_sCAL))
	{
		*unit = info_ptr->scal_unit;
		*width = info_ptr->scal_pixel_width;
		*height = info_ptr->scal_pixel_height;
		return (PNG_INFO_sCAL);
	}
	return(0);
}
#else
#ifdef PNG_FIXED_POINT_SUPPORTED
png_uint_32 PNGAPI
png_get_sCAL_s(png_structp png_ptr, png_infop info_ptr,
			   int *unit, png_charpp width, png_charpp height)
{
	if(png_ptr != NULL && info_ptr != NULL &&
		(info_ptr->valid & PNG_INFO_sCAL))
	{
		*unit = info_ptr->scal_unit;
		*width = info_ptr->scal_s_width;
		*height = info_ptr->scal_s_height;
		return (PNG_INFO_sCAL);
	}
	return(0);
}
#endif
#endif
#endif
#if defined(PNG_pHYs_SUPPORTED)
png_uint_32 PNGAPI
png_get_pHYs(png_structp png_ptr, png_infop info_ptr,
			 png_uint_32 *res_x, png_uint_32 *res_y, int *unit_type)
{
	png_uint_32 retval = 0;
	if(png_ptr != NULL && info_ptr != NULL &&
		(info_ptr->valid & PNG_INFO_pHYs))
	{
		png_debug1(1, "in %s retrieval function\n", "pHYs");
		if(res_x != NULL)
		{
			*res_x = info_ptr->x_pixels_per_unit;
			retval |= PNG_INFO_pHYs;
		}
		if(res_y != NULL)
		{
			*res_y = info_ptr->y_pixels_per_unit;
			retval |= PNG_INFO_pHYs;
		}
		if(unit_type != NULL)
		{
			*unit_type = (int)info_ptr->phys_unit_type;
			retval |= PNG_INFO_pHYs;
		}
	}
	return (retval);
}
#endif
png_uint_32 PNGAPI
png_get_PLTE(png_structp png_ptr, png_infop info_ptr, png_colorp *palette,
			 int *num_palette)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_PLTE)
		&& palette != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "PLTE");
		*palette = info_ptr->palette;
		*num_palette = info_ptr->num_palette;
		png_debug1(3, "num_palette = %d\n", *num_palette);
		return (PNG_INFO_PLTE);
	}
	return (0);
}
#if defined(PNG_sBIT_SUPPORTED)
png_uint_32 PNGAPI
png_get_sBIT(png_structp png_ptr, png_infop info_ptr, png_color_8p *sig_bit)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_sBIT)
		&& sig_bit != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "sBIT");
		*sig_bit = &(info_ptr->sig_bit);
		return (PNG_INFO_sBIT);
	}
	return (0);
}
#endif
#if defined(PNG_TEXT_SUPPORTED)
png_uint_32 PNGAPI
png_get_text(png_structp png_ptr, png_infop info_ptr, png_textp *text_ptr,
			 int *num_text)
{
	if(png_ptr != NULL && info_ptr != NULL && info_ptr->num_text > 0)
	{
		png_debug1(1, "in %s retrieval function\n",
			(png_ptr->chunk_name[0] == '\0' ? "text"
			: (png_const_charp)png_ptr->chunk_name));
		if(text_ptr != NULL)
			*text_ptr = info_ptr->text;
		if(num_text != NULL)
			*num_text = info_ptr->num_text;
		return ((png_uint_32)info_ptr->num_text);
	}
	if(num_text != NULL)
		*num_text = 0;
	return(0);
}
#endif
#if defined(PNG_tIME_SUPPORTED)
png_uint_32 PNGAPI
png_get_tIME(png_structp png_ptr, png_infop info_ptr, png_timep *mod_time)
{
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_tIME)
		&& mod_time != NULL)
	{
		png_debug1(1, "in %s retrieval function\n", "tIME");
		*mod_time = &(info_ptr->mod_time);
		return (PNG_INFO_tIME);
	}
	return (0);
}
#endif
#if defined(PNG_tRNS_SUPPORTED)
png_uint_32 PNGAPI
png_get_tRNS(png_structp png_ptr, png_infop info_ptr,
			 png_bytep *trans, int *num_trans, png_color_16p *trans_values)
{
	png_uint_32 retval = 0;
	if(png_ptr != NULL && info_ptr != NULL && (info_ptr->valid & PNG_INFO_tRNS))
	{
		png_debug1(1, "in %s retrieval function\n", "tRNS");
		if(info_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		{
			if(trans != NULL)
			{
				*trans = info_ptr->trans;
				retval |= PNG_INFO_tRNS;
			}
			if(trans_values != NULL)
				*trans_values = &(info_ptr->trans_values);
		}
		else
		{
			if(trans_values != NULL)
			{
				*trans_values = &(info_ptr->trans_values);
				retval |= PNG_INFO_tRNS;
			}
			if(trans != NULL)
				*trans = NULL;
		}
		if(num_trans != NULL)
		{
			*num_trans = info_ptr->num_trans;
			retval |= PNG_INFO_tRNS;
		}
	}
	return (retval);
}
#endif
#if defined(PNG_UNKNOWN_CHUNKS_SUPPORTED)
png_uint_32 PNGAPI
png_get_unknown_chunks(png_structp png_ptr, png_infop info_ptr,
					   png_unknown_chunkpp unknowns)
{
	if(png_ptr != NULL && info_ptr != NULL && unknowns != NULL)
		*unknowns = info_ptr->unknown_chunks;
	return ((png_uint_32)info_ptr->unknown_chunks_num);
}
#endif
#if defined(PNG_READ_RGB_TO_GRAY_SUPPORTED)
png_byte PNGAPI
png_get_rgb_to_gray_status (png_structp png_ptr)
{
	return (png_byte)(png_ptr? png_ptr->rgb_to_gray_status : 0);
}
#endif
#if defined(PNG_USER_CHUNKS_SUPPORTED)
png_voidp PNGAPI
png_get_user_chunk_ptr(png_structp png_ptr)
{
	return (png_ptr? png_ptr->user_chunk_ptr : NULL);
}
#endif
#ifdef PNG_WRITE_SUPPORTED
png_uint_32 PNGAPI
png_get_compression_buffer_size(png_structp png_ptr)
{
	return (png_uint_32)(png_ptr? png_ptr->zbuf_size : 0L);
}
#endif
#ifndef PNG_1_0_X
#ifdef PNG_ASSEMBLER_CODE_SUPPORTED
png_uint_32 PNGAPI
png_get_asm_flags (png_structp png_ptr)
{
	return (png_uint_32)(png_ptr? png_ptr->asm_flags : 0L);
}
png_uint_32 PNGAPI
png_get_asm_flagmask (int flag_select)
{
	png_uint_32 settable_asm_flags = 0;
	if(flag_select & PNG_SELECT_READ)
		settable_asm_flags |=
		PNG_ASM_FLAG_MMX_READ_COMBINE_ROW  |
		PNG_ASM_FLAG_MMX_READ_INTERLACE    |
		PNG_ASM_FLAG_MMX_READ_FILTER_SUB   |
		PNG_ASM_FLAG_MMX_READ_FILTER_UP    |
		PNG_ASM_FLAG_MMX_READ_FILTER_AVG   |
		PNG_ASM_FLAG_MMX_READ_FILTER_PAETH ;
#if 0
	if(flag_select & PNG_SELECT_WRITE)
		settable_asm_flags |=
		PNG_ASM_FLAG_MMX_WRITE_ [whatever] ;
#endif
	return settable_asm_flags;
}
#endif
#if defined(PNG_ASSEMBLER_CODE_SUPPORTED)
png_uint_32 PNGAPI
png_get_mmx_flagmask (int flag_select, int *compilerID)
{
	png_uint_32 settable_mmx_flags = 0;
	if(flag_select & PNG_SELECT_READ)
		settable_mmx_flags |=
		PNG_ASM_FLAG_MMX_READ_COMBINE_ROW  |
		PNG_ASM_FLAG_MMX_READ_INTERLACE    |
		PNG_ASM_FLAG_MMX_READ_FILTER_SUB   |
		PNG_ASM_FLAG_MMX_READ_FILTER_UP    |
		PNG_ASM_FLAG_MMX_READ_FILTER_AVG   |
		PNG_ASM_FLAG_MMX_READ_FILTER_PAETH ;
#if 0
	if(flag_select & PNG_SELECT_WRITE)
		settable_mmx_flags |=
		PNG_ASM_FLAG_MMX_WRITE_ [whatever] ;
#endif
	if(compilerID != NULL)
	{
#ifdef PNG_USE_PNGVCRD
		*compilerID = 1;
#else
#ifdef PNG_USE_PNGGCCRD
		*compilerID = 2;
#else
		*compilerID = -1;
#endif
#endif
	}
	return settable_mmx_flags;
}
png_byte PNGAPI
png_get_mmx_bitdepth_threshold (png_structp png_ptr)
{
	return (png_byte)(png_ptr? png_ptr->mmx_bitdepth_threshold : 0);
}
png_uint_32 PNGAPI
png_get_mmx_rowbytes_threshold (png_structp png_ptr)
{
	return (png_uint_32)(png_ptr? png_ptr->mmx_rowbytes_threshold : 0L);
}
#endif
#ifdef PNG_SET_USER_LIMITS_SUPPORTED
png_uint_32 PNGAPI
png_get_user_width_max (png_structp png_ptr)
{
	return (png_ptr? png_ptr->user_width_max : 0);
}
png_uint_32 PNGAPI
png_get_user_height_max (png_structp png_ptr)
{
	return (png_ptr? png_ptr->user_height_max : 0);
}
#endif
#endif
#define PNG_INTERNAL
#include "png.h"
#if defined(__TURBOC__) && !defined(_Windows) && !defined(__FLAT__)
png_voidp
png_create_struct(int type)
{
#ifdef PNG_USER_MEM_SUPPORTED
	return (png_create_struct_2(type, png_malloc_ptr_NULL, png_voidp_NULL));
}
png_voidp
png_create_struct_2(int type, png_malloc_ptr malloc_fn, png_voidp mem_ptr)
{
#endif
	png_size_t size;
	png_voidp struct_ptr;
	if(type == PNG_STRUCT_INFO)
		size = png_sizeof(png_info);
	else if(type == PNG_STRUCT_PNG)
		size = png_sizeof(png_struct);
	else
		return (png_get_copyright(NULL));
#ifdef PNG_USER_MEM_SUPPORTED
	if(malloc_fn != NULL)
	{
		png_struct dummy_struct;
		png_structp png_ptr = &dummy_struct;
		png_ptr->mem_ptr = mem_ptr;
		struct_ptr = (*(malloc_fn))(png_ptr, (png_uint_32)size);
	}
	else
#endif
		struct_ptr = (png_voidp)farmalloc(size);
	if(struct_ptr != NULL)
		png_memset(struct_ptr, 0, size);
	return (struct_ptr);
}
void
png_destroy_struct(png_voidp struct_ptr)
{
#ifdef PNG_USER_MEM_SUPPORTED
	png_destroy_struct_2(struct_ptr, png_free_ptr_NULL, png_voidp_NULL);
}
void
png_destroy_struct_2(png_voidp struct_ptr, png_free_ptr free_fn,
					 png_voidp mem_ptr)
{
#endif
	if(struct_ptr != NULL)
	{
#ifdef PNG_USER_MEM_SUPPORTED
		if(free_fn != NULL)
		{
			png_struct dummy_struct;
			png_structp png_ptr = &dummy_struct;
			png_ptr->mem_ptr = mem_ptr;
			(*(free_fn))(png_ptr, struct_ptr);
			return;
		}
#endif
		farfree (struct_ptr);
	}
}
png_voidp PNGAPI
png_malloc(png_structp png_ptr, png_uint_32 size)
{
	png_voidp ret;
	if(png_ptr == NULL || size == 0)
		return (NULL);
#ifdef PNG_USER_MEM_SUPPORTED
	if(png_ptr->malloc_fn != NULL)
		ret = ((png_voidp)(*(png_ptr->malloc_fn))(png_ptr, (png_size_t)size));
	else
		ret = (png_malloc_default(png_ptr, size));
	if(ret == NULL && (png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
		png_error(png_ptr, "Out of memory!");
	return (ret);
}
png_voidp PNGAPI
png_malloc_default(png_structp png_ptr, png_uint_32 size)
{
	png_voidp ret;
#endif
#ifdef PNG_MAX_MALLOC_64K
	if(size > (png_uint_32)65536L)
	{
		png_warning(png_ptr, "Cannot Allocate > 64K");
		ret = NULL;
	}
	else
#endif
		if(size != (size_t)size)
			ret = NULL;
		else if(size == (png_uint_32)65536L)
		{
			if(png_ptr->offset_table == NULL)
			{
				ret = farmalloc(size);
				if(ret == NULL || ((png_size_t)ret & 0xffff))
				{
					int num_blocks;
					png_uint_32 total_size;
					png_bytep table;
					int i;
					png_byte huge * hptr;
					if(ret != NULL)
					{
						farfree(ret);
						ret = NULL;
					}
					if(png_ptr->zlib_window_bits > 14)
						num_blocks = (int)(1 << (png_ptr->zlib_window_bits - 14));
					else
						num_blocks = 1;
					if(png_ptr->zlib_mem_level >= 7)
						num_blocks += (int)(1 << (png_ptr->zlib_mem_level - 7));
					else
						num_blocks++;
					total_size = ((png_uint_32)65536L) * (png_uint_32)num_blocks+16;
					table = farmalloc(total_size);
					if(table == NULL)
					{
#ifndef PNG_USER_MEM_SUPPORTED
						if((png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
							png_error(png_ptr, "Out Of Memory.");
						else
							png_warning(png_ptr, "Out Of Memory.");
#endif
						return (NULL);
					}
					if((png_size_t)table & 0xfff0)
					{
#ifndef PNG_USER_MEM_SUPPORTED
						if((png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
							png_error(png_ptr,
							"Farmalloc didn't return normalized pointer");
						else
							png_warning(png_ptr,
							"Farmalloc didn't return normalized pointer");
#endif
						return (NULL);
					}
					png_ptr->offset_table = table;
					png_ptr->offset_table_ptr = farmalloc(num_blocks *
						png_sizeof (png_bytep));
					if(png_ptr->offset_table_ptr == NULL)
					{
#ifndef PNG_USER_MEM_SUPPORTED
						if((png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
							png_error(png_ptr, "Out Of memory.");
						else
							png_warning(png_ptr, "Out Of memory.");
#endif
						return (NULL);
					}
					hptr = (png_byte huge *)table;
					if((png_size_t)hptr & 0xf)
					{
						hptr = (png_byte huge *)((long)(hptr) & 0xfffffff0L);
						hptr = hptr + 16L;
					}
					for(i = 0; i < num_blocks; i++)
					{
						png_ptr->offset_table_ptr[i] = (png_bytep)hptr;
						hptr = hptr + (png_uint_32)65536L;
					}
					png_ptr->offset_table_number = num_blocks;
					png_ptr->offset_table_count = 0;
					png_ptr->offset_table_count_free = 0;
				}
			}
			if(png_ptr->offset_table_count >= png_ptr->offset_table_number)
			{
#ifndef PNG_USER_MEM_SUPPORTED
				if((png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
					png_error(png_ptr, "Out of Memory.");
				else
					png_warning(png_ptr, "Out of Memory.");
#endif
				return (NULL);
			}
			ret = png_ptr->offset_table_ptr[png_ptr->offset_table_count++];
		}
		else
			ret = farmalloc(size);
#ifndef PNG_USER_MEM_SUPPORTED
		if(ret == NULL)
		{
			if((png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
				png_error(png_ptr, "Out of memory.");
			else
				png_warning(png_ptr, "Out of memory.");
		}
#endif
		return (ret);
}
void PNGAPI
png_free(png_structp png_ptr, png_voidp ptr)
{
	if(png_ptr == NULL || ptr == NULL)
		return;
#ifdef PNG_USER_MEM_SUPPORTED
	if(png_ptr->free_fn != NULL)
	{
		(*(png_ptr->free_fn))(png_ptr, ptr);
		return;
	}
	else png_free_default(png_ptr, ptr);
}
void PNGAPI
png_free_default(png_structp png_ptr, png_voidp ptr)
{
#endif
	if(png_ptr->offset_table != NULL)
	{
		int i;
		for(i = 0; i < png_ptr->offset_table_count; i++)
		{
			if(ptr == png_ptr->offset_table_ptr[i])
			{
				ptr = NULL;
				png_ptr->offset_table_count_free++;
				break;
			}
		}
		if(png_ptr->offset_table_count_free == png_ptr->offset_table_count)
		{
			farfree(png_ptr->offset_table);
			farfree(png_ptr->offset_table_ptr);
			png_ptr->offset_table = NULL;
			png_ptr->offset_table_ptr = NULL;
		}
	}
	if(ptr != NULL)
	{
		farfree(ptr);
	}
}
#else
png_voidp
png_create_struct(int type)
{
#ifdef PNG_USER_MEM_SUPPORTED
	return (png_create_struct_2(type, png_malloc_ptr_NULL, png_voidp_NULL));
}
png_voidp
png_create_struct_2(int type, png_malloc_ptr malloc_fn, png_voidp mem_ptr)
{
#endif
	png_size_t size;
	png_voidp struct_ptr;
	if(type == PNG_STRUCT_INFO)
		size = png_sizeof(png_info);
	else if(type == PNG_STRUCT_PNG)
		size = png_sizeof(png_struct);
	else
		return (NULL);
#ifdef PNG_USER_MEM_SUPPORTED
	if(malloc_fn != NULL)
	{
		png_struct dummy_struct;
		png_structp png_ptr = &dummy_struct;
		png_ptr->mem_ptr = mem_ptr;
		struct_ptr = (*(malloc_fn))(png_ptr, size);
		if(struct_ptr != NULL)
			png_memset(struct_ptr, 0, size);
		return (struct_ptr);
	}
#endif
#if defined(__TURBOC__) && !defined(__FLAT__)
	struct_ptr = (png_voidp)farmalloc(size);
#else
# if defined(_MSC_VER) && defined(MAXSEG_64K)
	struct_ptr = (png_voidp)halloc(size,1);
# else
	struct_ptr = (png_voidp)malloc(size);
# endif
#endif
	if(struct_ptr != NULL)
		png_memset(struct_ptr, 0, size);
	return (struct_ptr);
}
void
png_destroy_struct(png_voidp struct_ptr)
{
#ifdef PNG_USER_MEM_SUPPORTED
	png_destroy_struct_2(struct_ptr, png_free_ptr_NULL, png_voidp_NULL);
}
void
png_destroy_struct_2(png_voidp struct_ptr, png_free_ptr free_fn,
					 png_voidp mem_ptr)
{
#endif
	if(struct_ptr != NULL)
	{
#ifdef PNG_USER_MEM_SUPPORTED
		if(free_fn != NULL)
		{
			png_struct dummy_struct;
			png_structp png_ptr = &dummy_struct;
			png_ptr->mem_ptr = mem_ptr;
			(*(free_fn))(png_ptr, struct_ptr);
			return;
		}
#endif
#if defined(__TURBOC__) && !defined(__FLAT__)
		farfree(struct_ptr);
#else
# if defined(_MSC_VER) && defined(MAXSEG_64K)
		hfree(struct_ptr);
# else
		free(struct_ptr);
# endif
#endif
	}
}
png_voidp PNGAPI
png_malloc(png_structp png_ptr, png_uint_32 size)
{
	png_voidp ret;
#ifdef PNG_USER_MEM_SUPPORTED
	if(png_ptr == NULL || size == 0)
		return (NULL);
	if(png_ptr->malloc_fn != NULL)
		ret = ((png_voidp)(*(png_ptr->malloc_fn))(png_ptr, (png_size_t)size));
	else
		ret = (png_malloc_default(png_ptr, size));
	if(ret == NULL && (png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
		png_error(png_ptr, "Out of Memory!");
	return (ret);
}
png_voidp PNGAPI
png_malloc_default(png_structp png_ptr, png_uint_32 size)
{
	png_voidp ret;
#endif
	if(png_ptr == NULL || size == 0)
		return (NULL);
#ifdef PNG_MAX_MALLOC_64K
	if(size > (png_uint_32)65536L)
	{
#ifndef PNG_USER_MEM_SUPPORTED
		if(png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
			png_error(png_ptr, "Cannot Allocate > 64K");
		else
#endif
			return NULL;
	}
#endif
#if defined(__TURBOC__) && !defined(__FLAT__)
	if(size != (unsigned long)size)
		ret = NULL;
	else
		ret = farmalloc(size);
#else
# if defined(_MSC_VER) && defined(MAXSEG_64K)
	if(size != (unsigned long)size)
		ret = NULL;
	else
		ret = halloc(size, 1);
# else
	if(size != (size_t)size)
		ret = NULL;
	else
		ret = malloc((size_t)size);
# endif
#endif
#ifndef PNG_USER_MEM_SUPPORTED
	if(ret == NULL && (png_ptr->flags&PNG_FLAG_MALLOC_NULL_MEM_OK) == 0)
		png_error(png_ptr, "Out of Memory");
#endif
	return (ret);
}
void PNGAPI
png_free(png_structp png_ptr, png_voidp ptr)
{
	if(png_ptr == NULL || ptr == NULL)
		return;
#ifdef PNG_USER_MEM_SUPPORTED
	if(png_ptr->free_fn != NULL)
	{
		(*(png_ptr->free_fn))(png_ptr, ptr);
		return;
	}
	else png_free_default(png_ptr, ptr);
}
void PNGAPI
png_free_default(png_structp png_ptr, png_voidp ptr)
{
	if(png_ptr == NULL || ptr == NULL)
		return;
#endif
#if defined(__TURBOC__) && !defined(__FLAT__)
	farfree(ptr);
#else
# if defined(_MSC_VER) && defined(MAXSEG_64K)
	hfree(ptr);
# else
	free(ptr);
# endif
#endif
}
#endif
#if defined(PNG_1_0_X)
#  define png_malloc_warn png_malloc
#else
png_voidp PNGAPI
png_malloc_warn(png_structp png_ptr, png_uint_32 size)
{
	png_voidp ptr;
	png_uint_32 save_flags = png_ptr->flags;
	png_ptr->flags |= PNG_FLAG_MALLOC_NULL_MEM_OK;
	ptr = (png_voidp)png_malloc((png_structp)png_ptr, size);
	png_ptr->flags = save_flags;
	return(ptr);
}
#endif
png_voidp PNGAPI
png_memcpy_check (png_structp png_ptr, png_voidp s1, png_voidp s2,
				  png_uint_32 length)
{
	png_size_t size;
	size = (png_size_t)length;
	if((png_uint_32)size != length)
		png_error(png_ptr,"Overflow in png_memcpy_check.");
	return(png_memcpy (s1, s2, size));
}
png_voidp PNGAPI
png_memset_check (png_structp png_ptr, png_voidp s1, int value,
				  png_uint_32 length)
{
	png_size_t size;
	size = (png_size_t)length;
	if((png_uint_32)size != length)
		png_error(png_ptr,"Overflow in png_memset_check.");
	return (png_memset (s1, value, size));
}
#ifdef PNG_USER_MEM_SUPPORTED
void PNGAPI
png_set_mem_fn(png_structp png_ptr, png_voidp mem_ptr, png_malloc_ptr
			   malloc_fn, png_free_ptr free_fn)
{
	png_ptr->mem_ptr = mem_ptr;
	png_ptr->malloc_fn = malloc_fn;
	png_ptr->free_fn = free_fn;
}
png_voidp PNGAPI
png_get_mem_ptr(png_structp png_ptr)
{
	return ((png_voidp)png_ptr->mem_ptr);
}
#endif
#define PNG_INTERNAL
#include "png.h"
png_structp PNGAPI
png_create_read_struct(png_const_charp user_png_ver, png_voidp error_ptr,
					   png_error_ptr error_fn, png_error_ptr warn_fn)
{
#ifdef PNG_USER_MEM_SUPPORTED
	return (png_create_read_struct_2(user_png_ver, error_ptr, error_fn,
		warn_fn, png_voidp_NULL, png_malloc_ptr_NULL, png_free_ptr_NULL));
}
png_structp PNGAPI
png_create_read_struct_2(png_const_charp user_png_ver, png_voidp error_ptr,
						 png_error_ptr error_fn, png_error_ptr warn_fn, png_voidp mem_ptr,
						 png_malloc_ptr malloc_fn, png_free_ptr free_fn)
{
#endif
	png_structp png_ptr;
#ifdef PNG_SETJMP_SUPPORTED
#ifdef USE_FAR_KEYWORD
	jmp_buf jmpbuf;
#endif
#endif
	int i;
	png_debug(1, "in png_create_read_struct\n");
#ifdef PNG_USER_MEM_SUPPORTED
	png_ptr = (png_structp)png_create_struct_2(PNG_STRUCT_PNG,
		(png_malloc_ptr)malloc_fn, (png_voidp)mem_ptr);
#else
	png_ptr = (png_structp)png_create_struct(PNG_STRUCT_PNG);
#endif
	if(png_ptr == NULL)
		return (NULL);
#if !defined(PNG_1_0_X)
#ifdef PNG_ASSEMBLER_CODE_SUPPORTED
	png_init_mmx_flags(png_ptr);
#endif
#endif
#ifdef PNG_SET_USER_LIMITS_SUPPORTED
	png_ptr->user_width_max = PNG_USER_WIDTH_MAX;
	png_ptr->user_height_max = PNG_USER_HEIGHT_MAX;
#endif
#ifdef PNG_SETJMP_SUPPORTED
#ifdef USE_FAR_KEYWORD
	if(setjmp(jmpbuf))
#else
	if(setjmp(png_ptr->jmpbuf))
#endif
	{
		png_free(png_ptr, png_ptr->zbuf);
		png_ptr->zbuf = NULL;
#ifdef PNG_USER_MEM_SUPPORTED
		png_destroy_struct_2((png_voidp)png_ptr,
			(png_free_ptr)free_fn, (png_voidp)mem_ptr);
#else
		png_destroy_struct((png_voidp)png_ptr);
#endif
		return (NULL);
	}
#ifdef USE_FAR_KEYWORD
	png_memcpy(png_ptr->jmpbuf,jmpbuf,png_sizeof(jmp_buf));
#endif
#endif
#ifdef PNG_USER_MEM_SUPPORTED
	png_set_mem_fn(png_ptr, mem_ptr, malloc_fn, free_fn);
#endif
	png_set_error_fn(png_ptr, error_ptr, error_fn, warn_fn);
	i = 0;
	do
	{
		if(user_png_ver[i] != png_libpng_ver[i])
			png_ptr->flags |= PNG_FLAG_LIBRARY_MISMATCH;
	} while(png_libpng_ver[i++]);
	if(png_ptr->flags & PNG_FLAG_LIBRARY_MISMATCH)
	{
		if(user_png_ver == NULL || user_png_ver[0] != png_libpng_ver[0] ||
			(user_png_ver[0] == '1' && user_png_ver[2] != png_libpng_ver[2]) ||
			(user_png_ver[0] == '0' && user_png_ver[2] < '9'))
		{
#if !defined(PNG_NO_STDIO) && !defined(_WIN32_WCE)
			char msg[80];
			if(user_png_ver)
			{
				sprintf(msg, "Application was compiled with png.h from libpng-%.20s",
					user_png_ver);
				png_warning(png_ptr, msg);
			}
			sprintf(msg, "Application  is  running with png.c from libpng-%.20s",
				png_libpng_ver);
			png_warning(png_ptr, msg);
#endif
#ifdef PNG_ERROR_NUMBERS_SUPPORTED
			png_ptr->flags = 0;
#endif
			png_error(png_ptr,
				"Incompatible libpng version in application and library");
		}
	}
	png_ptr->zbuf_size = PNG_ZBUF_SIZE;
	png_ptr->zbuf = (png_bytep)png_malloc(png_ptr,
		(png_uint_32)png_ptr->zbuf_size);
	png_ptr->zstream.zalloc = png_zalloc;
	png_ptr->zstream.zfree = png_zfree;
	png_ptr->zstream.opaque = (voidpf)png_ptr;
	switch (inflateInit(&png_ptr->zstream))
	{
	case Z_OK:  break;
	case Z_MEM_ERROR:
	case Z_STREAM_ERROR: png_error(png_ptr, "zlib memory error"); break;
	case Z_VERSION_ERROR: png_error(png_ptr, "zlib version error"); break;
	default: png_error(png_ptr, "Unknown zlib error");
	}
	png_ptr->zstream.next_out = png_ptr->zbuf;
	png_ptr->zstream.avail_out = (uInt)png_ptr->zbuf_size;
	png_set_read_fn(png_ptr, png_voidp_NULL, png_rw_ptr_NULL);
#ifdef PNG_SETJMP_SUPPORTED
#ifdef USE_FAR_KEYWORD
	if(setjmp(jmpbuf))
		PNG_ABORT();
	png_memcpy(png_ptr->jmpbuf,jmpbuf,png_sizeof(jmp_buf));
#else
	if(setjmp(png_ptr->jmpbuf))
		PNG_ABORT();
#endif
#endif
	return (png_ptr);
}
#if defined(PNG_1_0_X) || defined (PNG_1_2_X)
#undef png_read_init
void PNGAPI
png_read_init(png_structp png_ptr)
{
	png_read_init_2(png_ptr, "1.0.6 or earlier", 0, 0);
}
#endif
void PNGAPI
png_read_init_2(png_structp png_ptr, png_const_charp user_png_ver,
				png_size_t png_struct_size, png_size_t png_info_size)
{
#if !defined(PNG_NO_STDIO) && !defined(_WIN32_WCE)
	if(png_sizeof(png_struct) > png_struct_size ||
		png_sizeof(png_info) > png_info_size)
	{
		char msg[80];
		png_ptr->warning_fn = NULL;
		if(user_png_ver)
		{
			sprintf(msg, "Application was compiled with png.h from libpng-%.20s",
				user_png_ver);
			png_warning(png_ptr, msg);
		}
		sprintf(msg, "Application  is  running with png.c from libpng-%.20s",
			png_libpng_ver);
		png_warning(png_ptr, msg);
	}
#endif
	if(png_sizeof(png_struct) > png_struct_size)
	{
		png_ptr->error_fn = NULL;
#ifdef PNG_ERROR_NUMBERS_SUPPORTED
		png_ptr->flags = 0;
#endif
		png_error(png_ptr,
			"The png struct allocated by the application for reading is too small.");
	}
	if(png_sizeof(png_info) > png_info_size)
	{
		png_ptr->error_fn = NULL;
#ifdef PNG_ERROR_NUMBERS_SUPPORTED
		png_ptr->flags = 0;
#endif
		png_error(png_ptr,
			"The info struct allocated by application for reading is too small.");
	}
	png_read_init_3(&png_ptr, user_png_ver, png_struct_size);
}
void PNGAPI
png_read_init_3(png_structpp ptr_ptr, png_const_charp user_png_ver,
				png_size_t png_struct_size)
{
#ifdef PNG_SETJMP_SUPPORTED
	jmp_buf tmp_jmp;
#endif
	int i = 0;
	png_structp png_ptr = *ptr_ptr;
	do
	{
		if(user_png_ver[i] != png_libpng_ver[i])
		{
#ifdef PNG_LEGACY_SUPPORTED
			png_ptr->flags |= PNG_FLAG_LIBRARY_MISMATCH;
#else
			png_ptr->warning_fn = NULL;
			png_warning(png_ptr,
				"Application uses deprecated png_read_init() and should be recompiled.");
			break;
#endif
		}
	} while(png_libpng_ver[i++]);
	png_debug(1, "in png_read_init_3\n");
#ifdef PNG_SETJMP_SUPPORTED
	png_memcpy(tmp_jmp, png_ptr->jmpbuf, png_sizeof (jmp_buf));
#endif
	if(png_sizeof(png_struct) > png_struct_size)
	{
		png_destroy_struct(png_ptr);
		*ptr_ptr = (png_structp)png_create_struct(PNG_STRUCT_PNG);
		png_ptr = *ptr_ptr;
	}
	png_memset(png_ptr, 0, png_sizeof (png_struct));
#ifdef PNG_SETJMP_SUPPORTED
	png_memcpy(png_ptr->jmpbuf, tmp_jmp, png_sizeof (jmp_buf));
#endif
#ifdef PNG_SET_USER_LIMITS_SUPPORTED
	png_ptr->user_width_max = PNG_USER_WIDTH_MAX;
	png_ptr->user_height_max = PNG_USER_HEIGHT_MAX;
#endif
	png_ptr->zbuf_size = PNG_ZBUF_SIZE;
	png_ptr->zbuf = (png_bytep)png_malloc(png_ptr,
		(png_uint_32)png_ptr->zbuf_size);
	png_ptr->zstream.zalloc = png_zalloc;
	png_ptr->zstream.zfree = png_zfree;
	png_ptr->zstream.opaque = (voidpf)png_ptr;
	switch (inflateInit(&png_ptr->zstream))
	{
	case Z_OK:  break;
	case Z_MEM_ERROR:
	case Z_STREAM_ERROR: png_error(png_ptr, "zlib memory"); break;
	case Z_VERSION_ERROR: png_error(png_ptr, "zlib version"); break;
	default: png_error(png_ptr, "Unknown zlib error");
	}
	png_ptr->zstream.next_out = png_ptr->zbuf;
	png_ptr->zstream.avail_out = (uInt)png_ptr->zbuf_size;
	png_set_read_fn(png_ptr, png_voidp_NULL, png_rw_ptr_NULL);
}
#ifndef PNG_NO_SEQUENTIAL_READ_SUPPORTED
void PNGAPI
png_read_info(png_structp png_ptr, png_infop info_ptr)
{
	png_debug(1, "in png_read_info\n");
	if(png_ptr->sig_bytes < 8)
	{
		png_size_t num_checked = png_ptr->sig_bytes,
			num_to_check = 8 - num_checked;
		png_read_data(png_ptr, &(info_ptr->signature[num_checked]), num_to_check);
		png_ptr->sig_bytes = 8;
		if(png_sig_cmp(info_ptr->signature, num_checked, num_to_check))
		{
			if(num_checked < 4 &&
				png_sig_cmp(info_ptr->signature, num_checked, num_to_check - 4))
				png_error(png_ptr, "Not a PNG file");
			else
				png_error(png_ptr, "PNG file corrupted by ASCII conversion");
		}
		if(num_checked < 3)
			png_ptr->mode |= PNG_HAVE_PNG_SIGNATURE;
	}
	for(;;)
	{
#ifdef PNG_USE_LOCAL_ARRAYS
		PNG_IHDR;
		PNG_IDAT;
		PNG_IEND;
		PNG_PLTE;
#if defined(PNG_READ_bKGD_SUPPORTED)
		PNG_bKGD;
#endif
#if defined(PNG_READ_cHRM_SUPPORTED)
		PNG_cHRM;
#endif
#if defined(PNG_READ_gAMA_SUPPORTED)
		PNG_gAMA;
#endif
#if defined(PNG_READ_hIST_SUPPORTED)
		PNG_hIST;
#endif
#if defined(PNG_READ_iCCP_SUPPORTED)
		PNG_iCCP;
#endif
#if defined(PNG_READ_iTXt_SUPPORTED)
		PNG_iTXt;
#endif
#if defined(PNG_READ_oFFs_SUPPORTED)
		PNG_oFFs;
#endif
#if defined(PNG_READ_pCAL_SUPPORTED)
		PNG_pCAL;
#endif
#if defined(PNG_READ_pHYs_SUPPORTED)
		PNG_pHYs;
#endif
#if defined(PNG_READ_sBIT_SUPPORTED)
		PNG_sBIT;
#endif
#if defined(PNG_READ_sCAL_SUPPORTED)
		PNG_sCAL;
#endif
#if defined(PNG_READ_sPLT_SUPPORTED)
		PNG_sPLT;
#endif
#if defined(PNG_READ_sRGB_SUPPORTED)
		PNG_sRGB;
#endif
#if defined(PNG_READ_tEXt_SUPPORTED)
		PNG_tEXt;
#endif
#if defined(PNG_READ_tIME_SUPPORTED)
		PNG_tIME;
#endif
#if defined(PNG_READ_tRNS_SUPPORTED)
		PNG_tRNS;
#endif
#if defined(PNG_READ_zTXt_SUPPORTED)
		PNG_zTXt;
#endif
#endif
		png_byte chunk_length[4];
		png_uint_32 length;
		png_read_data(png_ptr, chunk_length, 4);
		length = png_get_uint_31(png_ptr,chunk_length);
		png_reset_crc(png_ptr);
		png_crc_read(png_ptr, png_ptr->chunk_name, 4);
		png_debug2(0, "Reading %s chunk, length = %lu.\n", png_ptr->chunk_name,
			length);
		if(!png_memcmp(png_ptr->chunk_name, png_IHDR, 4))
			png_handle_IHDR(png_ptr, info_ptr, length);
		else if(!png_memcmp(png_ptr->chunk_name, png_IEND, 4))
			png_handle_IEND(png_ptr, info_ptr, length);
#ifdef PNG_HANDLE_AS_UNKNOWN_SUPPORTED
		else if(png_handle_as_unknown(png_ptr, png_ptr->chunk_name))
		{
			if(!png_memcmp(png_ptr->chunk_name, png_IDAT, 4))
				png_ptr->mode |= PNG_HAVE_IDAT;
			png_handle_unknown(png_ptr, info_ptr, length);
			if(!png_memcmp(png_ptr->chunk_name, png_PLTE, 4))
				png_ptr->mode |= PNG_HAVE_PLTE;
			else if(!png_memcmp(png_ptr->chunk_name, png_IDAT, 4))
			{
				if(!(png_ptr->mode & PNG_HAVE_IHDR))
					png_error(png_ptr, "Missing IHDR before IDAT");
				else if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE &&
					!(png_ptr->mode & PNG_HAVE_PLTE))
					png_error(png_ptr, "Missing PLTE before IDAT");
				break;
			}
		}
#endif
		else if(!png_memcmp(png_ptr->chunk_name, png_PLTE, 4))
			png_handle_PLTE(png_ptr, info_ptr, length);
		else if(!png_memcmp(png_ptr->chunk_name, png_IDAT, 4))
		{
			if(!(png_ptr->mode & PNG_HAVE_IHDR))
				png_error(png_ptr, "Missing IHDR before IDAT");
			else if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE &&
				!(png_ptr->mode & PNG_HAVE_PLTE))
				png_error(png_ptr, "Missing PLTE before IDAT");
			png_ptr->idat_size = length;
			png_ptr->mode |= PNG_HAVE_IDAT;
			break;
		}
#if defined(PNG_READ_bKGD_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_bKGD, 4))
			png_handle_bKGD(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_cHRM_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_cHRM, 4))
			png_handle_cHRM(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_gAMA_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_gAMA, 4))
			png_handle_gAMA(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_hIST_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_hIST, 4))
			png_handle_hIST(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_oFFs_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_oFFs, 4))
			png_handle_oFFs(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_pCAL_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_pCAL, 4))
			png_handle_pCAL(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sCAL_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sCAL, 4))
			png_handle_sCAL(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_pHYs_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_pHYs, 4))
			png_handle_pHYs(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sBIT_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sBIT, 4))
			png_handle_sBIT(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sRGB_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sRGB, 4))
			png_handle_sRGB(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_iCCP_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_iCCP, 4))
			png_handle_iCCP(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sPLT_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sPLT, 4))
			png_handle_sPLT(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_tEXt_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_tEXt, 4))
			png_handle_tEXt(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_tIME_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_tIME, 4))
			png_handle_tIME(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_tRNS_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_tRNS, 4))
			png_handle_tRNS(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_zTXt_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_zTXt, 4))
			png_handle_zTXt(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_iTXt_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_iTXt, 4))
			png_handle_iTXt(png_ptr, info_ptr, length);
#endif
		else
			png_handle_unknown(png_ptr, info_ptr, length);
	}
}
#endif
void PNGAPI
png_read_update_info(png_structp png_ptr, png_infop info_ptr)
{
	png_debug(1, "in png_read_update_info\n");
	if(!(png_ptr->flags & PNG_FLAG_ROW_INIT))
		png_read_start_row(png_ptr);
	else
		png_warning(png_ptr,
		"Ignoring extra png_read_update_info() call; row buffer not reallocated");
	png_read_transform_info(png_ptr, info_ptr);
}
#ifndef PNG_NO_SEQUENTIAL_READ_SUPPORTED
void PNGAPI
png_start_read_image(png_structp png_ptr)
{
	png_debug(1, "in png_start_read_image\n");
	if(!(png_ptr->flags & PNG_FLAG_ROW_INIT))
		png_read_start_row(png_ptr);
}
#endif
#ifndef PNG_NO_SEQUENTIAL_READ_SUPPORTED
void PNGAPI
png_read_row(png_structp png_ptr, png_bytep row, png_bytep dsp_row)
{
#ifdef PNG_USE_LOCAL_ARRAYS
	PNG_IDAT;
	const int png_pass_dsp_mask[7] = {0xff, 0x0f, 0xff, 0x33, 0xff, 0x55, 0xff};
	const int png_pass_mask[7] = {0x80, 0x08, 0x88, 0x22, 0xaa, 0x55, 0xff};
#endif
	int ret;
	png_debug2(1, "in png_read_row (row %lu, pass %d)\n",
		png_ptr->row_number, png_ptr->pass);
	if(!(png_ptr->flags & PNG_FLAG_ROW_INIT))
		png_read_start_row(png_ptr);
	if(png_ptr->row_number == 0 && png_ptr->pass == 0)
	{
#if defined(PNG_WRITE_INVERT_SUPPORTED) && !defined(PNG_READ_INVERT_SUPPORTED)
		if(png_ptr->transformations & PNG_INVERT_MONO)
			png_warning(png_ptr, "PNG_READ_INVERT_SUPPORTED is not defined.");
#endif
#if defined(PNG_WRITE_FILLER_SUPPORTED) && !defined(PNG_READ_FILLER_SUPPORTED)
		if(png_ptr->transformations & PNG_FILLER)
			png_warning(png_ptr, "PNG_READ_FILLER_SUPPORTED is not defined.");
#endif
#if defined(PNG_WRITE_PACKSWAP_SUPPORTED) && !defined(PNG_READ_PACKSWAP_SUPPORTED)
		if(png_ptr->transformations & PNG_PACKSWAP)
			png_warning(png_ptr, "PNG_READ_PACKSWAP_SUPPORTED is not defined.");
#endif
#if defined(PNG_WRITE_PACK_SUPPORTED) && !defined(PNG_READ_PACK_SUPPORTED)
		if(png_ptr->transformations & PNG_PACK)
			png_warning(png_ptr, "PNG_READ_PACK_SUPPORTED is not defined.");
#endif
#if defined(PNG_WRITE_SHIFT_SUPPORTED) && !defined(PNG_READ_SHIFT_SUPPORTED)
		if(png_ptr->transformations & PNG_SHIFT)
			png_warning(png_ptr, "PNG_READ_SHIFT_SUPPORTED is not defined.");
#endif
#if defined(PNG_WRITE_BGR_SUPPORTED) && !defined(PNG_READ_BGR_SUPPORTED)
		if(png_ptr->transformations & PNG_BGR)
			png_warning(png_ptr, "PNG_READ_BGR_SUPPORTED is not defined.");
#endif
#if defined(PNG_WRITE_SWAP_SUPPORTED) && !defined(PNG_READ_SWAP_SUPPORTED)
		if(png_ptr->transformations & PNG_SWAP_BYTES)
			png_warning(png_ptr, "PNG_READ_SWAP_SUPPORTED is not defined.");
#endif
	}
#if defined(PNG_READ_INTERLACING_SUPPORTED)
	if(png_ptr->interlaced && (png_ptr->transformations & PNG_INTERLACE))
	{
		switch (png_ptr->pass)
		{
		case 0:
			if(png_ptr->row_number & 0x07)
			{
				if(dsp_row != NULL)
					png_combine_row(png_ptr, dsp_row,
					png_pass_dsp_mask[png_ptr->pass]);
				png_read_finish_row(png_ptr);
				return;
			}
			break;
		case 1:
			if((png_ptr->row_number & 0x07) || png_ptr->width < 5)
			{
				if(dsp_row != NULL)
					png_combine_row(png_ptr, dsp_row,
					png_pass_dsp_mask[png_ptr->pass]);
				png_read_finish_row(png_ptr);
				return;
			}
			break;
		case 2:
			if((png_ptr->row_number & 0x07) != 4)
			{
				if(dsp_row != NULL && (png_ptr->row_number & 4))
					png_combine_row(png_ptr, dsp_row,
					png_pass_dsp_mask[png_ptr->pass]);
				png_read_finish_row(png_ptr);
				return;
			}
			break;
		case 3:
			if((png_ptr->row_number & 3) || png_ptr->width < 3)
			{
				if(dsp_row != NULL)
					png_combine_row(png_ptr, dsp_row,
					png_pass_dsp_mask[png_ptr->pass]);
				png_read_finish_row(png_ptr);
				return;
			}
			break;
		case 4:
			if((png_ptr->row_number & 3) != 2)
			{
				if(dsp_row != NULL && (png_ptr->row_number & 2))
					png_combine_row(png_ptr, dsp_row,
					png_pass_dsp_mask[png_ptr->pass]);
				png_read_finish_row(png_ptr);
				return;
			}
			break;
		case 5:
			if((png_ptr->row_number & 1) || png_ptr->width < 2)
			{
				if(dsp_row != NULL)
					png_combine_row(png_ptr, dsp_row,
					png_pass_dsp_mask[png_ptr->pass]);
				png_read_finish_row(png_ptr);
				return;
			}
			break;
		case 6:
			if(!(png_ptr->row_number & 1))
			{
				png_read_finish_row(png_ptr);
				return;
			}
			break;
		}
	}
#endif
	if(!(png_ptr->mode & PNG_HAVE_IDAT))
		png_error(png_ptr, "Invalid attempt to read row data");
	png_ptr->zstream.next_out = png_ptr->row_buf;
	png_ptr->zstream.avail_out = (uInt)png_ptr->irowbytes;
	do
	{
		if(!(png_ptr->zstream.avail_in))
		{
			while(!png_ptr->idat_size)
			{
				png_byte chunk_length[4];
				png_crc_finish(png_ptr, 0);
				png_read_data(png_ptr, chunk_length, 4);
				png_ptr->idat_size = png_get_uint_31(png_ptr,chunk_length);
				png_reset_crc(png_ptr);
				png_crc_read(png_ptr, png_ptr->chunk_name, 4);
				if(png_memcmp(png_ptr->chunk_name, png_IDAT, 4))
					png_error(png_ptr, "Not enough image data");
			}
			png_ptr->zstream.avail_in = (uInt)png_ptr->zbuf_size;
			png_ptr->zstream.next_in = png_ptr->zbuf;
			if(png_ptr->zbuf_size > png_ptr->idat_size)
				png_ptr->zstream.avail_in = (uInt)png_ptr->idat_size;
			png_crc_read(png_ptr, png_ptr->zbuf,
				(png_size_t)png_ptr->zstream.avail_in);
			png_ptr->idat_size -= png_ptr->zstream.avail_in;
		}
		ret = inflate(&png_ptr->zstream, Z_PARTIAL_FLUSH);
		if(ret == Z_STREAM_END)
		{
			if(png_ptr->zstream.avail_out || png_ptr->zstream.avail_in ||
				png_ptr->idat_size)
				png_error(png_ptr, "Extra compressed data");
			png_ptr->mode |= PNG_AFTER_IDAT;
			png_ptr->flags |= PNG_FLAG_ZLIB_FINISHED;
			break;
		}
		if(ret != Z_OK)
			png_error(png_ptr, png_ptr->zstream.msg ? png_ptr->zstream.msg :
		"Decompression error");
	} while(png_ptr->zstream.avail_out);
	png_ptr->row_info.color_type = png_ptr->color_type;
	png_ptr->row_info.width = png_ptr->iwidth;
	png_ptr->row_info.channels = png_ptr->channels;
	png_ptr->row_info.bit_depth = png_ptr->bit_depth;
	png_ptr->row_info.pixel_depth = png_ptr->pixel_depth;
	png_ptr->row_info.rowbytes = PNG_ROWBYTES(png_ptr->row_info.pixel_depth,
		png_ptr->row_info.width);
	if(png_ptr->row_buf[0])
		png_read_filter_row(png_ptr, &(png_ptr->row_info),
		png_ptr->row_buf + 1, png_ptr->prev_row + 1,
		(int)(png_ptr->row_buf[0]));
	png_memcpy_check(png_ptr, png_ptr->prev_row, png_ptr->row_buf,
		png_ptr->rowbytes + 1);
#if defined(PNG_MNG_FEATURES_SUPPORTED)
	if((png_ptr->mng_features_permitted & PNG_FLAG_MNG_FILTER_64) &&
		(png_ptr->filter_type == PNG_INTRAPIXEL_DIFFERENCING))
	{
		png_do_read_intrapixel(&(png_ptr->row_info), png_ptr->row_buf + 1);
	}
#endif
	if(png_ptr->transformations || (png_ptr->flags&PNG_FLAG_STRIP_ALPHA))
		png_do_read_transformations(png_ptr);
#if defined(PNG_READ_INTERLACING_SUPPORTED)
	if(png_ptr->interlaced &&
		(png_ptr->transformations & PNG_INTERLACE))
	{
		if(png_ptr->pass < 6)
			png_do_read_interlace(png_ptr);
		if(dsp_row != NULL)
			png_combine_row(png_ptr, dsp_row,
			png_pass_dsp_mask[png_ptr->pass]);
		if(row != NULL)
			png_combine_row(png_ptr, row,
			png_pass_mask[png_ptr->pass]);
	}
	else
#endif
	{
		if(row != NULL)
			png_combine_row(png_ptr, row, 0xff);
		if(dsp_row != NULL)
			png_combine_row(png_ptr, dsp_row, 0xff);
	}
	png_read_finish_row(png_ptr);
	if(png_ptr->read_row_fn != NULL)
		(*(png_ptr->read_row_fn))(png_ptr, png_ptr->row_number, png_ptr->pass);
}
#endif
#ifndef PNG_NO_SEQUENTIAL_READ_SUPPORTED
void PNGAPI
png_read_rows(png_structp png_ptr, png_bytepp row,
			  png_bytepp display_row, png_uint_32 num_rows)
{
	png_uint_32 i;
	png_bytepp rp;
	png_bytepp dp;
	png_debug(1, "in png_read_rows\n");
	rp = row;
	dp = display_row;
	if(rp != NULL && dp != NULL)
		for(i = 0; i < num_rows; i++)
		{
			png_bytep rptr = *rp++;
			png_bytep dptr = *dp++;
			png_read_row(png_ptr, rptr, dptr);
		}
	else if(rp != NULL)
		for(i = 0; i < num_rows; i++)
		{
			png_bytep rptr = *rp;
			png_read_row(png_ptr, rptr, png_bytep_NULL);
			rp++;
		}
	else if(dp != NULL)
		for(i = 0; i < num_rows; i++)
		{
			png_bytep dptr = *dp;
			png_read_row(png_ptr, png_bytep_NULL, dptr);
			dp++;
		}
}
#endif
#ifndef PNG_NO_SEQUENTIAL_READ_SUPPORTED
void PNGAPI
png_read_image(png_structp png_ptr, png_bytepp image)
{
	png_uint_32 i,image_height;
	int pass, j;
	png_bytepp rp;
	png_debug(1, "in png_read_image\n");
#ifdef PNG_READ_INTERLACING_SUPPORTED
	pass = png_set_interlace_handling(png_ptr);
#else
	if(png_ptr->interlaced)
		png_error(png_ptr,
		"Cannot read interlaced image -- interlace handler disabled.");
	pass = 1;
#endif
	image_height = png_ptr->height;
	png_ptr->num_rows = image_height;
	for(j = 0; j < pass; j++)
	{
		rp = image;
		for(i = 0; i < image_height; i++)
		{
			png_read_row(png_ptr, *rp, png_bytep_NULL);
			rp++;
		}
	}
}
#endif
#ifndef PNG_NO_SEQUENTIAL_READ_SUPPORTED
void PNGAPI
png_read_end(png_structp png_ptr, png_infop info_ptr)
{
	png_byte chunk_length[4];
	png_uint_32 length;
	png_debug(1, "in png_read_end\n");
	png_crc_finish(png_ptr, 0);
	do
	{
#ifdef PNG_USE_LOCAL_ARRAYS
		PNG_IHDR;
		PNG_IDAT;
		PNG_IEND;
		PNG_PLTE;
#if defined(PNG_READ_bKGD_SUPPORTED)
		PNG_bKGD;
#endif
#if defined(PNG_READ_cHRM_SUPPORTED)
		PNG_cHRM;
#endif
#if defined(PNG_READ_gAMA_SUPPORTED)
		PNG_gAMA;
#endif
#if defined(PNG_READ_hIST_SUPPORTED)
		PNG_hIST;
#endif
#if defined(PNG_READ_iCCP_SUPPORTED)
		PNG_iCCP;
#endif
#if defined(PNG_READ_iTXt_SUPPORTED)
		PNG_iTXt;
#endif
#if defined(PNG_READ_oFFs_SUPPORTED)
		PNG_oFFs;
#endif
#if defined(PNG_READ_pCAL_SUPPORTED)
		PNG_pCAL;
#endif
#if defined(PNG_READ_pHYs_SUPPORTED)
		PNG_pHYs;
#endif
#if defined(PNG_READ_sBIT_SUPPORTED)
		PNG_sBIT;
#endif
#if defined(PNG_READ_sCAL_SUPPORTED)
		PNG_sCAL;
#endif
#if defined(PNG_READ_sPLT_SUPPORTED)
		PNG_sPLT;
#endif
#if defined(PNG_READ_sRGB_SUPPORTED)
		PNG_sRGB;
#endif
#if defined(PNG_READ_tEXt_SUPPORTED)
		PNG_tEXt;
#endif
#if defined(PNG_READ_tIME_SUPPORTED)
		PNG_tIME;
#endif
#if defined(PNG_READ_tRNS_SUPPORTED)
		PNG_tRNS;
#endif
#if defined(PNG_READ_zTXt_SUPPORTED)
		PNG_zTXt;
#endif
#endif
		png_read_data(png_ptr, chunk_length, 4);
		length = png_get_uint_31(png_ptr,chunk_length);
		png_reset_crc(png_ptr);
		png_crc_read(png_ptr, png_ptr->chunk_name, 4);
		png_debug1(0, "Reading %s chunk.\n", png_ptr->chunk_name);
		if(!png_memcmp(png_ptr->chunk_name, png_IHDR, 4))
			png_handle_IHDR(png_ptr, info_ptr, length);
		else if(!png_memcmp(png_ptr->chunk_name, png_IEND, 4))
			png_handle_IEND(png_ptr, info_ptr, length);
#ifdef PNG_HANDLE_AS_UNKNOWN_SUPPORTED
		else if(png_handle_as_unknown(png_ptr, png_ptr->chunk_name))
		{
			if(!png_memcmp(png_ptr->chunk_name, png_IDAT, 4))
			{
				if(length > 0 || png_ptr->mode & PNG_AFTER_IDAT)
					png_error(png_ptr, "Too many IDAT's found");
			}
			else
				png_ptr->mode |= PNG_AFTER_IDAT;
			png_handle_unknown(png_ptr, info_ptr, length);
			if(!png_memcmp(png_ptr->chunk_name, png_PLTE, 4))
				png_ptr->mode |= PNG_HAVE_PLTE;
		}
#endif
		else if(!png_memcmp(png_ptr->chunk_name, png_IDAT, 4))
		{
			if(length > 0 || png_ptr->mode & PNG_AFTER_IDAT)
				png_error(png_ptr, "Too many IDAT's found");
			png_crc_finish(png_ptr, length);
		}
		else if(!png_memcmp(png_ptr->chunk_name, png_PLTE, 4))
			png_handle_PLTE(png_ptr, info_ptr, length);
#if defined(PNG_READ_bKGD_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_bKGD, 4))
			png_handle_bKGD(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_cHRM_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_cHRM, 4))
			png_handle_cHRM(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_gAMA_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_gAMA, 4))
			png_handle_gAMA(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_hIST_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_hIST, 4))
			png_handle_hIST(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_oFFs_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_oFFs, 4))
			png_handle_oFFs(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_pCAL_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_pCAL, 4))
			png_handle_pCAL(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sCAL_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sCAL, 4))
			png_handle_sCAL(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_pHYs_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_pHYs, 4))
			png_handle_pHYs(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sBIT_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sBIT, 4))
			png_handle_sBIT(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sRGB_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sRGB, 4))
			png_handle_sRGB(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_iCCP_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_iCCP, 4))
			png_handle_iCCP(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_sPLT_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_sPLT, 4))
			png_handle_sPLT(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_tEXt_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_tEXt, 4))
			png_handle_tEXt(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_tIME_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_tIME, 4))
			png_handle_tIME(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_tRNS_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_tRNS, 4))
			png_handle_tRNS(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_zTXt_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_zTXt, 4))
			png_handle_zTXt(png_ptr, info_ptr, length);
#endif
#if defined(PNG_READ_iTXt_SUPPORTED)
		else if(!png_memcmp(png_ptr->chunk_name, png_iTXt, 4))
			png_handle_iTXt(png_ptr, info_ptr, length);
#endif
		else
			png_handle_unknown(png_ptr, info_ptr, length);
	} while(!(png_ptr->mode & PNG_HAVE_IEND));
}
#endif
void PNGAPI
png_destroy_read_struct(png_structpp png_ptr_ptr, png_infopp info_ptr_ptr,
						png_infopp end_info_ptr_ptr)
{
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL, end_info_ptr = NULL;
#ifdef PNG_USER_MEM_SUPPORTED
	png_free_ptr free_fn;
	png_voidp mem_ptr;
#endif
	png_debug(1, "in png_destroy_read_struct\n");
	if(png_ptr_ptr != NULL)
		png_ptr = *png_ptr_ptr;
	if(info_ptr_ptr != NULL)
		info_ptr = *info_ptr_ptr;
	if(end_info_ptr_ptr != NULL)
		end_info_ptr = *end_info_ptr_ptr;
#ifdef PNG_USER_MEM_SUPPORTED
	free_fn = png_ptr->free_fn;
	mem_ptr = png_ptr->mem_ptr;
#endif
	png_read_destroy(png_ptr, info_ptr, end_info_ptr);
	if(info_ptr != NULL)
	{
#if defined(PNG_TEXT_SUPPORTED)
		png_free_data(png_ptr, info_ptr, PNG_FREE_TEXT, -1);
#endif
#ifdef PNG_USER_MEM_SUPPORTED
		png_destroy_struct_2((png_voidp)info_ptr, (png_free_ptr)free_fn,
			(png_voidp)mem_ptr);
#else
		png_destroy_struct((png_voidp)info_ptr);
#endif
		*info_ptr_ptr = NULL;
	}
	if(end_info_ptr != NULL)
	{
#if defined(PNG_READ_TEXT_SUPPORTED)
		png_free_data(png_ptr, end_info_ptr, PNG_FREE_TEXT, -1);
#endif
#ifdef PNG_USER_MEM_SUPPORTED
		png_destroy_struct_2((png_voidp)end_info_ptr, (png_free_ptr)free_fn,
			(png_voidp)mem_ptr);
#else
		png_destroy_struct((png_voidp)end_info_ptr);
#endif
		*end_info_ptr_ptr = NULL;
	}
	if(png_ptr != NULL)
	{
#ifdef PNG_USER_MEM_SUPPORTED
		png_destroy_struct_2((png_voidp)png_ptr, (png_free_ptr)free_fn,
			(png_voidp)mem_ptr);
#else
		png_destroy_struct((png_voidp)png_ptr);
#endif
		*png_ptr_ptr = NULL;
	}
}
void
png_read_destroy(png_structp png_ptr, png_infop info_ptr, png_infop end_info_ptr)
{
#ifdef PNG_SETJMP_SUPPORTED
	jmp_buf tmp_jmp;
#endif
	png_error_ptr error_fn;
	png_error_ptr warning_fn;
	png_voidp error_ptr;
#ifdef PNG_USER_MEM_SUPPORTED
	png_free_ptr free_fn;
#endif
	png_debug(1, "in png_read_destroy\n");
	if(info_ptr != NULL)
		png_info_destroy(png_ptr, info_ptr);
	if(end_info_ptr != NULL)
		png_info_destroy(png_ptr, end_info_ptr);
	png_free(png_ptr, png_ptr->zbuf);
	png_free(png_ptr, png_ptr->big_row_buf);
	png_free(png_ptr, png_ptr->prev_row);
#if defined(PNG_READ_DITHER_SUPPORTED)
	png_free(png_ptr, png_ptr->palette_lookup);
	png_free(png_ptr, png_ptr->dither_index);
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED)
	png_free(png_ptr, png_ptr->gamma_table);
#endif
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
	png_free(png_ptr, png_ptr->gamma_from_1);
	png_free(png_ptr, png_ptr->gamma_to_1);
#endif
#ifdef PNG_FREE_ME_SUPPORTED
	if(png_ptr->free_me & PNG_FREE_PLTE)
		png_zfree(png_ptr, png_ptr->palette);
	png_ptr->free_me &= ~PNG_FREE_PLTE;
#else
	if(png_ptr->flags & PNG_FLAG_FREE_PLTE)
		png_zfree(png_ptr, png_ptr->palette);
	png_ptr->flags &= ~PNG_FLAG_FREE_PLTE;
#endif
#if defined(PNG_tRNS_SUPPORTED) || \
	defined(PNG_READ_EXPAND_SUPPORTED) || defined(PNG_READ_BACKGROUND_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if(png_ptr->free_me & PNG_FREE_TRNS)
		png_free(png_ptr, png_ptr->trans);
	png_ptr->free_me &= ~PNG_FREE_TRNS;
#else
	if(png_ptr->flags & PNG_FLAG_FREE_TRNS)
		png_free(png_ptr, png_ptr->trans);
	png_ptr->flags &= ~PNG_FLAG_FREE_TRNS;
#endif
#endif
#if defined(PNG_READ_hIST_SUPPORTED)
#ifdef PNG_FREE_ME_SUPPORTED
	if(png_ptr->free_me & PNG_FREE_HIST)
		png_free(png_ptr, png_ptr->hist);
	png_ptr->free_me &= ~PNG_FREE_HIST;
#else
	if(png_ptr->flags & PNG_FLAG_FREE_HIST)
		png_free(png_ptr, png_ptr->hist);
	png_ptr->flags &= ~PNG_FLAG_FREE_HIST;
#endif
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED)
	if(png_ptr->gamma_16_table != NULL)
	{
		int i;
		int istop = (1 << (8 - png_ptr->gamma_shift));
		for(i = 0; i < istop; i++)
		{
			png_free(png_ptr, png_ptr->gamma_16_table[i]);
		}
		png_free(png_ptr, png_ptr->gamma_16_table);
	}
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
	if(png_ptr->gamma_16_from_1 != NULL)
	{
		int i;
		int istop = (1 << (8 - png_ptr->gamma_shift));
		for(i = 0; i < istop; i++)
		{
			png_free(png_ptr, png_ptr->gamma_16_from_1[i]);
		}
		png_free(png_ptr, png_ptr->gamma_16_from_1);
	}
	if(png_ptr->gamma_16_to_1 != NULL)
	{
		int i;
		int istop = (1 << (8 - png_ptr->gamma_shift));
		for(i = 0; i < istop; i++)
		{
			png_free(png_ptr, png_ptr->gamma_16_to_1[i]);
		}
		png_free(png_ptr, png_ptr->gamma_16_to_1);
	}
#endif
#endif
#if defined(PNG_TIME_RFC1123_SUPPORTED)
	png_free(png_ptr, png_ptr->time_buffer);
#endif
	inflateEnd(&png_ptr->zstream);
#ifdef PNG_PROGRESSIVE_READ_SUPPORTED
	png_free(png_ptr, png_ptr->save_buffer);
#endif
#ifdef PNG_PROGRESSIVE_READ_SUPPORTED
#ifdef PNG_TEXT_SUPPORTED
	png_free(png_ptr, png_ptr->current_text);
#endif
#endif
#ifdef PNG_SETJMP_SUPPORTED
	png_memcpy(tmp_jmp, png_ptr->jmpbuf, png_sizeof (jmp_buf));
#endif
	error_fn = png_ptr->error_fn;
	warning_fn = png_ptr->warning_fn;
	error_ptr = png_ptr->error_ptr;
#ifdef PNG_USER_MEM_SUPPORTED
	free_fn = png_ptr->free_fn;
#endif
	png_memset(png_ptr, 0, png_sizeof (png_struct));
	png_ptr->error_fn = error_fn;
	png_ptr->warning_fn = warning_fn;
	png_ptr->error_ptr = error_ptr;
#ifdef PNG_USER_MEM_SUPPORTED
	png_ptr->free_fn = free_fn;
#endif
#ifdef PNG_SETJMP_SUPPORTED
	png_memcpy(png_ptr->jmpbuf, tmp_jmp, png_sizeof (jmp_buf));
#endif
}
void PNGAPI
png_set_read_status_fn(png_structp png_ptr, png_read_status_ptr read_row_fn)
{
	png_ptr->read_row_fn = read_row_fn;
}
#ifndef PNG_NO_SEQUENTIAL_READ_SUPPORTED
#if defined(PNG_INFO_IMAGE_SUPPORTED)
void PNGAPI
png_read_png(png_structp png_ptr, png_infop info_ptr,
			 int transforms,
			 voidp params)
{
	int row;
#if defined(PNG_READ_INVERT_ALPHA_SUPPORTED)
	if(transforms & PNG_TRANSFORM_INVERT_ALPHA)
		png_set_invert_alpha(png_ptr);
#endif
	png_read_info(png_ptr, info_ptr);
	if(info_ptr->height > PNG_UINT_32_MAX/png_sizeof(png_bytep))
		png_error(png_ptr,"Image is too high to process with png_read_png()");
#if defined(PNG_READ_16_TO_8_SUPPORTED)
	if(transforms & PNG_TRANSFORM_STRIP_16)
		png_set_strip_16(png_ptr);
#endif
#if defined(PNG_READ_STRIP_ALPHA_SUPPORTED)
	if(transforms & PNG_TRANSFORM_STRIP_ALPHA)
		png_set_strip_alpha(png_ptr);
#endif
#if defined(PNG_READ_PACK_SUPPORTED) && !defined(PNG_READ_EXPAND_SUPPORTED)
	if(transforms & PNG_TRANSFORM_PACKING)
		png_set_packing(png_ptr);
#endif
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
	if(transforms & PNG_TRANSFORM_PACKSWAP)
		png_set_packswap(png_ptr);
#endif
#if defined(PNG_READ_EXPAND_SUPPORTED)
	if(transforms & PNG_TRANSFORM_EXPAND)
		if((png_ptr->bit_depth < 8) ||
			(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE) ||
			(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)))
			png_set_expand(png_ptr);
#endif
#if defined(PNG_READ_INVERT_SUPPORTED)
	if(transforms & PNG_TRANSFORM_INVERT_MONO)
		png_set_invert_mono(png_ptr);
#endif
#if defined(PNG_READ_SHIFT_SUPPORTED)
	if((transforms & PNG_TRANSFORM_SHIFT)
		&& png_get_valid(png_ptr, info_ptr, PNG_INFO_sBIT))
	{
		png_color_8p sig_bit;
		png_get_sBIT(png_ptr, info_ptr, &sig_bit);
		png_set_shift(png_ptr, sig_bit);
	}
#endif
#if defined(PNG_READ_BGR_SUPPORTED)
	if(transforms & PNG_TRANSFORM_BGR)
		png_set_bgr(png_ptr);
#endif
#if defined(PNG_READ_SWAP_ALPHA_SUPPORTED)
	if(transforms & PNG_TRANSFORM_SWAP_ALPHA)
		png_set_swap_alpha(png_ptr);
#endif
#if defined(PNG_READ_SWAP_SUPPORTED)
	if(transforms & PNG_TRANSFORM_SWAP_ENDIAN)
		png_set_swap(png_ptr);
#endif
	png_read_update_info(png_ptr, info_ptr);
#ifdef PNG_FREE_ME_SUPPORTED
	png_free_data(png_ptr, info_ptr, PNG_FREE_ROWS, 0);
#endif
	if(info_ptr->row_pointers == NULL)
	{
		info_ptr->row_pointers = (png_bytepp)png_malloc(png_ptr,
			info_ptr->height * png_sizeof(png_bytep));
#ifdef PNG_FREE_ME_SUPPORTED
		info_ptr->free_me |= PNG_FREE_ROWS;
#endif
		for(row = 0; row < (int)info_ptr->height; row++)
		{
			info_ptr->row_pointers[row] = (png_bytep)png_malloc(png_ptr,
				png_get_rowbytes(png_ptr, info_ptr));
		}
	}
	png_read_image(png_ptr, info_ptr->row_pointers);
	info_ptr->valid |= PNG_INFO_IDAT;
	png_read_end(png_ptr, info_ptr);
	if(transforms == 0 || params == NULL)
		return;
}
#endif
#endif
#define PNG_INTERNAL
#include "png.h"
void
png_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	png_debug1(4,"reading %d bytes\n", (int)length);
	if(png_ptr->read_data_fn != NULL)
		(*(png_ptr->read_data_fn))(png_ptr, data, length);
	else
		png_error(png_ptr, "Call to NULL read function");
}
#if !defined(PNG_NO_STDIO)
#ifndef USE_FAR_KEYWORD
void PNGAPI
png_default_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	png_size_t check;
#if defined(_WIN32_WCE)
	if( !ReadFile((HANDLE)(png_ptr->io_ptr), data, length, &check, NULL) )
		check = 0;
#else
	check = (png_size_t)fread(data, (png_size_t)1, length,
		(png_FILE_p)png_ptr->io_ptr);
#endif
	if(check != length)
		png_error(png_ptr, "Read Error");
}
#else
#define NEAR_BUF_SIZE 1024
#define MIN(a,b) (a <= b ? a : b)
static void
png_default_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
	int check;
	png_byte *n_data;
	png_FILE_p io_ptr;
	n_data = (png_byte *)CVT_PTR_NOCHECK(data);
	io_ptr = (png_FILE_p)CVT_PTR(png_ptr->io_ptr);
	if((png_bytep)n_data == data)
	{
#if defined(_WIN32_WCE)
		if( !ReadFile((HANDLE)(png_ptr->io_ptr), data, length, &check, NULL) )
			check = 0;
#else
		check = fread(n_data, 1, length, io_ptr);
#endif
	}
	else
	{
		png_byte buf[NEAR_BUF_SIZE];
		png_size_t read, remaining, err;
		check = 0;
		remaining = length;
		do
		{
			read = MIN(NEAR_BUF_SIZE, remaining);
#if defined(_WIN32_WCE)
			if( !ReadFile((HANDLE)(io_ptr), buf, read, &err, NULL) )
				err = 0;
#else
			err = fread(buf, (png_size_t)1, read, io_ptr);
#endif
			png_memcpy(data, buf, read);
			if(err != read)
				break;
			else
				check += err;
			data += read;
			remaining -= read;
		}
		while(remaining != 0);
	}
	if((png_uint_32)check != (png_uint_32)length)
		png_error(png_ptr, "read Error");
}
#endif
#endif
void PNGAPI
png_set_read_fn(png_structp png_ptr, png_voidp io_ptr,
				png_rw_ptr read_data_fn)
{
	png_ptr->io_ptr = io_ptr;
#if !defined(PNG_NO_STDIO)
	if(read_data_fn != NULL)
		png_ptr->read_data_fn = read_data_fn;
	else
		png_ptr->read_data_fn = png_default_read_data;
#else
	png_ptr->read_data_fn = read_data_fn;
#endif
	if(png_ptr->write_data_fn != NULL)
	{
		png_ptr->write_data_fn = NULL;
		png_warning(png_ptr,
			"It's an error to set both read_data_fn and write_data_fn in the ");
		png_warning(png_ptr,
			"same structure.  Resetting write_data_fn to NULL.");
	}
#if defined(PNG_WRITE_FLUSH_SUPPORTED)
	png_ptr->output_flush_fn = NULL;
#endif
}
#define PNG_INTERNAL
#include "png.h"
void PNGAPI
png_set_crc_action(png_structp png_ptr, int crit_action, int ancil_action)
{
	png_debug(1, "in png_set_crc_action\n");
	switch (crit_action)
	{
	case PNG_CRC_NO_CHANGE:
		break;
	case PNG_CRC_WARN_USE:
		png_ptr->flags &= ~PNG_FLAG_CRC_CRITICAL_MASK;
		png_ptr->flags |= PNG_FLAG_CRC_CRITICAL_USE;
		break;
	case PNG_CRC_QUIET_USE:
		png_ptr->flags &= ~PNG_FLAG_CRC_CRITICAL_MASK;
		png_ptr->flags |= PNG_FLAG_CRC_CRITICAL_USE |
			PNG_FLAG_CRC_CRITICAL_IGNORE;
		break;
	case PNG_CRC_WARN_DISCARD:
		png_warning(png_ptr, "Can't discard critical data on CRC error.");
	case PNG_CRC_ERROR_QUIT:
	case PNG_CRC_DEFAULT:
	default:
		png_ptr->flags &= ~PNG_FLAG_CRC_CRITICAL_MASK;
		break;
	}
	switch (ancil_action)
	{
	case PNG_CRC_NO_CHANGE:
		break;
	case PNG_CRC_WARN_USE:
		png_ptr->flags &= ~PNG_FLAG_CRC_ANCILLARY_MASK;
		png_ptr->flags |= PNG_FLAG_CRC_ANCILLARY_USE;
		break;
	case PNG_CRC_QUIET_USE:
		png_ptr->flags &= ~PNG_FLAG_CRC_ANCILLARY_MASK;
		png_ptr->flags |= PNG_FLAG_CRC_ANCILLARY_USE |
			PNG_FLAG_CRC_ANCILLARY_NOWARN;
		break;
	case PNG_CRC_ERROR_QUIT:
		png_ptr->flags &= ~PNG_FLAG_CRC_ANCILLARY_MASK;
		png_ptr->flags |= PNG_FLAG_CRC_ANCILLARY_NOWARN;
		break;
	case PNG_CRC_WARN_DISCARD:
	case PNG_CRC_DEFAULT:
	default:
		png_ptr->flags &= ~PNG_FLAG_CRC_ANCILLARY_MASK;
		break;
	}
}
#if defined(PNG_READ_BACKGROUND_SUPPORTED) && \
	defined(PNG_FLOATING_POINT_SUPPORTED)
void PNGAPI
png_set_background(png_structp png_ptr,
				   png_color_16p background_color, int background_gamma_code,
				   int need_expand, double background_gamma)
{
	png_debug(1, "in png_set_background\n");
	if(background_gamma_code == PNG_BACKGROUND_GAMMA_UNKNOWN)
	{
		png_warning(png_ptr, "Application must supply a known background gamma");
		return;
	}
	png_ptr->transformations |= PNG_BACKGROUND;
	png_memcpy(&(png_ptr->background), background_color,
		png_sizeof(png_color_16));
	png_ptr->background_gamma = (float)background_gamma;
	png_ptr->background_gamma_type = (png_byte)(background_gamma_code);
	png_ptr->transformations |= (need_expand ? PNG_BACKGROUND_EXPAND : 0);
	if((need_expand && !(png_ptr->color_type & PNG_COLOR_MASK_COLOR)) ||
		(!need_expand && background_color->red == background_color->green &&
		background_color->red == background_color->blue))
		png_ptr->mode |= PNG_BACKGROUND_IS_GRAY;
}
#endif
#if defined(PNG_READ_16_TO_8_SUPPORTED)
void PNGAPI
png_set_strip_16(png_structp png_ptr)
{
	png_debug(1, "in png_set_strip_16\n");
	png_ptr->transformations |= PNG_16_TO_8;
}
#endif
#if defined(PNG_READ_STRIP_ALPHA_SUPPORTED)
void PNGAPI
png_set_strip_alpha(png_structp png_ptr)
{
	png_debug(1, "in png_set_strip_alpha\n");
	png_ptr->flags |= PNG_FLAG_STRIP_ALPHA;
}
#endif
#if defined(PNG_READ_DITHER_SUPPORTED)
typedef struct png_dsort_struct
{
	struct png_dsort_struct FAR * next;
	png_byte left;
	png_byte right;
} png_dsort;
typedef png_dsort FAR *       png_dsortp;
typedef png_dsort FAR * FAR * png_dsortpp;
void PNGAPI
png_set_dither(png_structp png_ptr, png_colorp palette,
			   int num_palette, int maximum_colors, png_uint_16p histogram,
			   int full_dither)
{
	png_debug(1, "in png_set_dither\n");
	png_ptr->transformations |= PNG_DITHER;
	if(!full_dither)
	{
		int i;
		png_ptr->dither_index = (png_bytep)png_malloc(png_ptr,
			(png_uint_32)(num_palette * png_sizeof (png_byte)));
		for(i = 0; i < num_palette; i++)
			png_ptr->dither_index[i] = (png_byte)i;
	}
	if(num_palette > maximum_colors)
	{
		if(histogram != NULL)
		{
			int i;
			png_ptr->dither_sort = (png_bytep)png_malloc(png_ptr,
				(png_uint_32)(num_palette * png_sizeof (png_byte)));
			for(i = 0; i < num_palette; i++)
				png_ptr->dither_sort[i] = (png_byte)i;
			for(i = num_palette - 1; i >= maximum_colors; i--)
			{
				int done;
				int j;
				done = 1;
				for(j = 0; j < i; j++)
				{
					if(histogram[png_ptr->dither_sort[j]]
					< histogram[png_ptr->dither_sort[j + 1]])
					{
						png_byte t;
						t = png_ptr->dither_sort[j];
						png_ptr->dither_sort[j] = png_ptr->dither_sort[j + 1];
						png_ptr->dither_sort[j + 1] = t;
						done = 0;
					}
				}
				if(done)
					break;
			}
			if(full_dither)
			{
				int j = num_palette;
				for(i = 0; i < maximum_colors; i++)
				{
					if((int)png_ptr->dither_sort[i] >= maximum_colors)
					{
						do
						j--;
						while((int)png_ptr->dither_sort[j] >= maximum_colors);
						palette[i] = palette[j];
					}
				}
			}
			else
			{
				int j = num_palette;
				for(i = 0; i < maximum_colors; i++)
				{
					if((int)png_ptr->dither_sort[i] >= maximum_colors)
					{
						png_color tmp_color;
						do
						j--;
						while((int)png_ptr->dither_sort[j] >= maximum_colors);
						tmp_color = palette[j];
						palette[j] = palette[i];
						palette[i] = tmp_color;
						png_ptr->dither_index[j] = (png_byte)i;
						png_ptr->dither_index[i] = (png_byte)j;
					}
				}
				for(i = 0; i < num_palette; i++)
				{
					if((int)png_ptr->dither_index[i] >= maximum_colors)
					{
						int min_d, k, min_k, d_index;
						d_index = png_ptr->dither_index[i];
						min_d = PNG_COLOR_DIST(palette[d_index], palette[0]);
						for(k = 1, min_k = 0; k < maximum_colors; k++)
						{
							int d;
							d = PNG_COLOR_DIST(palette[d_index], palette[k]);
							if(d < min_d)
							{
								min_d = d;
								min_k = k;
							}
						}
						png_ptr->dither_index[i] = (png_byte)min_k;
					}
				}
			}
			png_free(png_ptr, png_ptr->dither_sort);
			png_ptr->dither_sort = NULL;
		}
		else
		{
			int i;
			int max_d;
			int num_new_palette;
			png_dsortp t;
			png_dsortpp hash;
			t = NULL;
			png_ptr->index_to_palette = (png_bytep)png_malloc(png_ptr,
				(png_uint_32)(num_palette * png_sizeof (png_byte)));
			png_ptr->palette_to_index = (png_bytep)png_malloc(png_ptr,
				(png_uint_32)(num_palette * png_sizeof (png_byte)));
			for(i = 0; i < num_palette; i++)
			{
				png_ptr->index_to_palette[i] = (png_byte)i;
				png_ptr->palette_to_index[i] = (png_byte)i;
			}
			hash = (png_dsortpp)png_malloc(png_ptr, (png_uint_32)(769 *
				png_sizeof (png_dsortp)));
			for(i = 0; i < 769; i++)
				hash[i] = NULL;
			num_new_palette = num_palette;
			max_d = 96;
			while(num_new_palette > maximum_colors)
			{
				for(i = 0; i < num_new_palette - 1; i++)
				{
					int j;
					for(j = i + 1; j < num_new_palette; j++)
					{
						int d;
						d = PNG_COLOR_DIST(palette[i], palette[j]);
						if(d <= max_d)
						{
							t = (png_dsortp)png_malloc_warn(png_ptr,
								(png_uint_32)(png_sizeof(png_dsort)));
							if(t == NULL)
								break;
							t->next = hash[d];
							t->left = (png_byte)i;
							t->right = (png_byte)j;
							hash[d] = t;
						}
					}
					if(t == NULL)
						break;
				}
				if(t != NULL)
					for(i = 0; i <= max_d; i++)
					{
						if(hash[i] != NULL)
						{
							png_dsortp p;
							for(p = hash[i]; p; p = p->next)
							{
								if((int)png_ptr->index_to_palette[p->left]
								< num_new_palette &&
									(int)png_ptr->index_to_palette[p->right]
									< num_new_palette)
									{
										int j, next_j;
										if(num_new_palette & 0x01)
										{
											j = p->left;
											next_j = p->right;
										}
										else
										{
											j = p->right;
											next_j = p->left;
										}
										num_new_palette--;
										palette[png_ptr->index_to_palette[j]]
										= palette[num_new_palette];
										if(!full_dither)
										{
											int k;
											for(k = 0; k < num_palette; k++)
											{
												if(png_ptr->dither_index[k] ==
													png_ptr->index_to_palette[j])
													png_ptr->dither_index[k] =
													png_ptr->index_to_palette[next_j];
												if((int)png_ptr->dither_index[k] ==
													num_new_palette)
													png_ptr->dither_index[k] =
													png_ptr->index_to_palette[j];
											}
										}
										png_ptr->index_to_palette[png_ptr->palette_to_index
											[num_new_palette]] = png_ptr->index_to_palette[j];
										png_ptr->palette_to_index[png_ptr->index_to_palette[j]]
										= png_ptr->palette_to_index[num_new_palette];
										png_ptr->index_to_palette[j] = (png_byte)num_new_palette;
										png_ptr->palette_to_index[num_new_palette] = (png_byte)j;
									}
									if(num_new_palette <= maximum_colors)
										break;
							}
							if(num_new_palette <= maximum_colors)
								break;
						}
					}
					for(i = 0; i < 769; i++)
					{
						if(hash[i] != NULL)
						{
							png_dsortp p = hash[i];
							while(p)
							{
								t = p->next;
								png_free(png_ptr, p);
								p = t;
							}
						}
						hash[i] = 0;
					}
					max_d += 96;
			}
			png_free(png_ptr, hash);
			png_free(png_ptr, png_ptr->palette_to_index);
			png_free(png_ptr, png_ptr->index_to_palette);
			png_ptr->palette_to_index = NULL;
			png_ptr->index_to_palette = NULL;
		}
		num_palette = maximum_colors;
	}
	if(png_ptr->palette == NULL)
	{
		png_ptr->palette = palette;
	}
	png_ptr->num_palette = (png_uint_16)num_palette;
	if(full_dither)
	{
		int i;
		png_bytep distance;
		int total_bits = PNG_DITHER_RED_BITS + PNG_DITHER_GREEN_BITS +
			PNG_DITHER_BLUE_BITS;
		int num_red = (1 << PNG_DITHER_RED_BITS);
		int num_green = (1 << PNG_DITHER_GREEN_BITS);
		int num_blue = (1 << PNG_DITHER_BLUE_BITS);
		png_size_t num_entries = ((png_size_t)1 << total_bits);
		png_ptr->palette_lookup = (png_bytep )png_malloc(png_ptr,
			(png_uint_32)(num_entries * png_sizeof (png_byte)));
		png_memset(png_ptr->palette_lookup, 0, num_entries *
			png_sizeof (png_byte));
		distance = (png_bytep)png_malloc(png_ptr, (png_uint_32)(num_entries *
			png_sizeof(png_byte)));
		png_memset(distance, 0xff, num_entries * png_sizeof(png_byte));
		for(i = 0; i < num_palette; i++)
		{
			int ir, ig, ib;
			int r = (palette[i].red >> (8 - PNG_DITHER_RED_BITS));
			int g = (palette[i].green >> (8 - PNG_DITHER_GREEN_BITS));
			int b = (palette[i].blue >> (8 - PNG_DITHER_BLUE_BITS));
			for(ir = 0; ir < num_red; ir++)
			{
				int dr = ((ir > r) ? ir - r : r - ir);
				int index_r = (ir << (PNG_DITHER_BLUE_BITS + PNG_DITHER_GREEN_BITS));
				for(ig = 0; ig < num_green; ig++)
				{
					int dg = ((ig > g) ? ig - g : g - ig);
					int dt = dr + dg;
					int dm = ((dr > dg) ? dr : dg);
					int index_g = index_r | (ig << PNG_DITHER_BLUE_BITS);
					for(ib = 0; ib < num_blue; ib++)
					{
						int d_index = index_g | ib;
						int db = ((ib > b) ? ib - b : b - ib);
						int dmax = ((dm > db) ? dm : db);
						int d = dmax + dt + db;
						if(d < (int)distance[d_index])
						{
							distance[d_index] = (png_byte)d;
							png_ptr->palette_lookup[d_index] = (png_byte)i;
						}
					}
				}
			}
		}
		png_free(png_ptr, distance);
	}
}
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED) && defined(PNG_FLOATING_POINT_SUPPORTED)
void PNGAPI
png_set_gamma(png_structp png_ptr, double scrn_gamma, double file_gamma)
{
	png_debug(1, "in png_set_gamma\n");
	if((fabs(scrn_gamma * file_gamma - 1.0) > PNG_GAMMA_THRESHOLD) ||
		(png_ptr->color_type & PNG_COLOR_MASK_ALPHA) ||
		(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE))
		png_ptr->transformations |= PNG_GAMMA;
	png_ptr->gamma = (float)file_gamma;
	png_ptr->screen_gamma = (float)scrn_gamma;
}
#endif
#if defined(PNG_READ_EXPAND_SUPPORTED)
void PNGAPI
png_set_expand(png_structp png_ptr)
{
	png_debug(1, "in png_set_expand\n");
	png_ptr->transformations |= PNG_EXPAND;
}
void PNGAPI
png_set_palette_to_rgb(png_structp png_ptr)
{
	png_debug(1, "in png_set_expand\n");
	png_ptr->transformations |= PNG_EXPAND;
}
void PNGAPI
png_set_gray_1_2_4_to_8(png_structp png_ptr)
{
	png_debug(1, "in png_set_expand\n");
	png_ptr->transformations |= PNG_EXPAND;
}
void PNGAPI
png_set_tRNS_to_alpha(png_structp png_ptr)
{
	png_debug(1, "in png_set_expand\n");
	png_ptr->transformations |= PNG_EXPAND;
}
#endif
#if defined(PNG_READ_GRAY_TO_RGB_SUPPORTED)
void PNGAPI
png_set_gray_to_rgb(png_structp png_ptr)
{
	png_debug(1, "in png_set_gray_to_rgb\n");
	png_ptr->transformations |= PNG_GRAY_TO_RGB;
}
#endif
#if defined(PNG_READ_RGB_TO_GRAY_SUPPORTED)
#if defined(PNG_FLOATING_POINT_SUPPORTED)
void PNGAPI
png_set_rgb_to_gray(png_structp png_ptr, int error_action, double red,
					double green)
{
	int red_fixed = (int)((float)red*100000.0 + 0.5);
	int green_fixed = (int)((float)green*100000.0 + 0.5);
	png_set_rgb_to_gray_fixed(png_ptr, error_action, red_fixed, green_fixed);
}
#endif
void PNGAPI
png_set_rgb_to_gray_fixed(png_structp png_ptr, int error_action,
						  png_fixed_point red, png_fixed_point green)
{
	png_debug(1, "in png_set_rgb_to_gray\n");
	switch(error_action)
	{
	case 1: png_ptr->transformations |= PNG_RGB_TO_GRAY;
		break;
	case 2: png_ptr->transformations |= PNG_RGB_TO_GRAY_WARN;
		break;
	case 3: png_ptr->transformations |= PNG_RGB_TO_GRAY_ERR;
	}
	if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
#if defined(PNG_READ_EXPAND_SUPPORTED)
		png_ptr->transformations |= PNG_EXPAND;
#else
	{
		png_warning(png_ptr, "Cannot do RGB_TO_GRAY without EXPAND_SUPPORTED.");
		png_ptr->transformations &= ~PNG_RGB_TO_GRAY;
	}
#endif
	{
		png_uint_16 red_int, green_int;
		if(red < 0 || green < 0)
		{
			red_int   =  6968;
			green_int = 23434;
		}
		else if(red + green < 100000L)
		{
			red_int = (png_uint_16)(((png_uint_32)red*32768L)/100000L);
			green_int = (png_uint_16)(((png_uint_32)green*32768L)/100000L);
		}
		else
		{
			png_warning(png_ptr, "ignoring out of range rgb_to_gray coefficients");
			red_int   =  6968;
			green_int = 23434;
		}
		png_ptr->rgb_to_gray_red_coeff   = red_int;
		png_ptr->rgb_to_gray_green_coeff = green_int;
		png_ptr->rgb_to_gray_blue_coeff  = (png_uint_16)(32768-red_int-green_int);
	}
}
#endif
#if defined(PNG_READ_USER_TRANSFORM_SUPPORTED) || \
	defined(PNG_WRITE_USER_TRANSFORM_SUPPORTED) || \
	defined(PNG_LEGACY_SUPPORTED)
void PNGAPI
png_set_read_user_transform_fn(png_structp png_ptr, png_user_transform_ptr
							   read_user_transform_fn)
{
	png_debug(1, "in png_set_read_user_transform_fn\n");
#if defined(PNG_READ_USER_TRANSFORM_SUPPORTED)
	png_ptr->transformations |= PNG_USER_TRANSFORM;
	png_ptr->read_user_transform_fn = read_user_transform_fn;
#endif
#ifdef PNG_LEGACY_SUPPORTED
	if(read_user_transform_fn)
		png_warning(png_ptr,
		"This version of libpng does not support user transforms");
#endif
}
#endif
void
png_init_read_transformations(png_structp png_ptr)
{
	png_debug(1, "in png_init_read_transformations\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(png_ptr != NULL)
#endif
	{
#if defined(PNG_READ_BACKGROUND_SUPPORTED) || defined(PNG_READ_SHIFT_SUPPORTED) \
	|| defined(PNG_READ_GAMMA_SUPPORTED)
		int color_type = png_ptr->color_type;
#endif
#if defined(PNG_READ_EXPAND_SUPPORTED) && defined(PNG_READ_BACKGROUND_SUPPORTED)
		if((png_ptr->transformations & PNG_BACKGROUND_EXPAND) &&
			(png_ptr->transformations & PNG_EXPAND))
		{
			if(!(color_type & PNG_COLOR_MASK_COLOR))
			{
				switch (png_ptr->bit_depth)
				{
				case 1:
					png_ptr->background.gray *= (png_uint_16)0xff;
					png_ptr->background.red = png_ptr->background.green
						=  png_ptr->background.blue = png_ptr->background.gray;
					break;
				case 2:
					png_ptr->background.gray *= (png_uint_16)0x55;
					png_ptr->background.red = png_ptr->background.green
						= png_ptr->background.blue = png_ptr->background.gray;
					break;
				case 4:
					png_ptr->background.gray *= (png_uint_16)0x11;
					png_ptr->background.red = png_ptr->background.green
						= png_ptr->background.blue = png_ptr->background.gray;
					break;
				case 8:
				case 16:
					png_ptr->background.red = png_ptr->background.green
						= png_ptr->background.blue = png_ptr->background.gray;
					break;
				}
			}
			else if(color_type == PNG_COLOR_TYPE_PALETTE)
			{
				png_ptr->background.red   =
					png_ptr->palette[png_ptr->background.index].red;
				png_ptr->background.green =
					png_ptr->palette[png_ptr->background.index].green;
				png_ptr->background.blue  =
					png_ptr->palette[png_ptr->background.index].blue;
#if defined(PNG_READ_INVERT_ALPHA_SUPPORTED)
				if(png_ptr->transformations & PNG_INVERT_ALPHA)
				{
#if defined(PNG_READ_EXPAND_SUPPORTED)
					if(!(png_ptr->transformations & PNG_EXPAND))
#endif
					{
						int i,istop;
						istop = (int)png_ptr->num_trans;
						for(i = 0; i<istop; i++)
							png_ptr->trans[i] = (png_byte)(255 - png_ptr->trans[i]);
					}
				}
#endif
			}
		}
#endif
#if defined(PNG_READ_BACKGROUND_SUPPORTED) && defined(PNG_READ_GAMMA_SUPPORTED)
		png_ptr->background_1 = png_ptr->background;
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED) && defined(PNG_FLOATING_POINT_SUPPORTED)
		if((color_type == PNG_COLOR_TYPE_PALETTE && png_ptr->num_trans != 0)
			&& (fabs(png_ptr->screen_gamma * png_ptr->gamma - 1.0)
			< PNG_GAMMA_THRESHOLD))
		{
			int i,k;
			k = 0;
			for(i = 0; i<png_ptr->num_trans; i++)
			{
				if(png_ptr->trans[i] != 0 && png_ptr->trans[i] != 0xff)
					k = 1;
			}
			if(k == 0)
				png_ptr->transformations &= (~PNG_GAMMA);
		}
		if(png_ptr->transformations & (PNG_GAMMA | PNG_RGB_TO_GRAY))
		{
			png_build_gamma_table(png_ptr);
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
			if(png_ptr->transformations & PNG_BACKGROUND)
			{
				if(color_type == PNG_COLOR_TYPE_PALETTE)
				{
					png_color back, back_1;
					png_colorp palette = png_ptr->palette;
					int num_palette = png_ptr->num_palette;
					int i;
					if(png_ptr->background_gamma_type == PNG_BACKGROUND_GAMMA_FILE)
					{
						back.red = png_ptr->gamma_table[png_ptr->background.red];
						back.green = png_ptr->gamma_table[png_ptr->background.green];
						back.blue = png_ptr->gamma_table[png_ptr->background.blue];
						back_1.red = png_ptr->gamma_to_1[png_ptr->background.red];
						back_1.green = png_ptr->gamma_to_1[png_ptr->background.green];
						back_1.blue = png_ptr->gamma_to_1[png_ptr->background.blue];
					}
					else
					{
						double g, gs;
						switch (png_ptr->background_gamma_type)
						{
						case PNG_BACKGROUND_GAMMA_SCREEN:
							g = (png_ptr->screen_gamma);
							gs = 1.0;
							break;
						case PNG_BACKGROUND_GAMMA_FILE:
							g = 1.0 / (png_ptr->gamma);
							gs = 1.0 / (png_ptr->gamma * png_ptr->screen_gamma);
							break;
						case PNG_BACKGROUND_GAMMA_UNIQUE:
							g = 1.0 / (png_ptr->background_gamma);
							gs = 1.0 / (png_ptr->background_gamma *
								png_ptr->screen_gamma);
							break;
						default:
							g = 1.0;
							gs = 1.0;
						}
						if( fabs(gs - 1.0) < PNG_GAMMA_THRESHOLD)
						{
							back.red   = (png_byte)png_ptr->background.red;
							back.green = (png_byte)png_ptr->background.green;
							back.blue  = (png_byte)png_ptr->background.blue;
						}
						else
						{
							back.red = (png_byte)(pow(
								(double)png_ptr->background.red/255, gs) * 255.0 + .5);
							back.green = (png_byte)(pow(
								(double)png_ptr->background.green/255, gs) * 255.0 + .5);
							back.blue = (png_byte)(pow(
								(double)png_ptr->background.blue/255, gs) * 255.0 + .5);
						}
						back_1.red = (png_byte)(pow(
							(double)png_ptr->background.red/255, g) * 255.0 + .5);
						back_1.green = (png_byte)(pow(
							(double)png_ptr->background.green/255, g) * 255.0 + .5);
						back_1.blue = (png_byte)(pow(
							(double)png_ptr->background.blue/255, g) * 255.0 + .5);
					}
					for(i = 0; i < num_palette; i++)
					{
						if(i < (int)png_ptr->num_trans && png_ptr->trans[i] != 0xff)
						{
							if(png_ptr->trans[i] == 0)
							{
								palette[i] = back;
							}
							else
							{
								png_byte v, w;
								v = png_ptr->gamma_to_1[palette[i].red];
								png_composite(w, v, png_ptr->trans[i], back_1.red);
								palette[i].red = png_ptr->gamma_from_1[w];
								v = png_ptr->gamma_to_1[palette[i].green];
								png_composite(w, v, png_ptr->trans[i], back_1.green);
								palette[i].green = png_ptr->gamma_from_1[w];
								v = png_ptr->gamma_to_1[palette[i].blue];
								png_composite(w, v, png_ptr->trans[i], back_1.blue);
								palette[i].blue = png_ptr->gamma_from_1[w];
							}
						}
						else
						{
							palette[i].red = png_ptr->gamma_table[palette[i].red];
							palette[i].green = png_ptr->gamma_table[palette[i].green];
							palette[i].blue = png_ptr->gamma_table[palette[i].blue];
						}
					}
				}
				else
				{
					double m = (double)(((png_uint_32)1 << png_ptr->bit_depth) - 1);
					double g = 1.0;
					double gs = 1.0;
					switch (png_ptr->background_gamma_type)
					{
					case PNG_BACKGROUND_GAMMA_SCREEN:
						g = (png_ptr->screen_gamma);
						gs = 1.0;
						break;
					case PNG_BACKGROUND_GAMMA_FILE:
						g = 1.0 / (png_ptr->gamma);
						gs = 1.0 / (png_ptr->gamma * png_ptr->screen_gamma);
						break;
					case PNG_BACKGROUND_GAMMA_UNIQUE:
						g = 1.0 / (png_ptr->background_gamma);
						gs = 1.0 / (png_ptr->background_gamma *
							png_ptr->screen_gamma);
						break;
					}
					png_ptr->background_1.gray = (png_uint_16)(pow(
						(double)png_ptr->background.gray / m, g) * m + .5);
					png_ptr->background.gray = (png_uint_16)(pow(
						(double)png_ptr->background.gray / m, gs) * m + .5);
					if((png_ptr->background.red != png_ptr->background.green) ||
						(png_ptr->background.red != png_ptr->background.blue) ||
						(png_ptr->background.red != png_ptr->background.gray))
					{
						png_ptr->background_1.red = (png_uint_16)(pow(
							(double)png_ptr->background.red / m, g) * m + .5);
						png_ptr->background_1.green = (png_uint_16)(pow(
							(double)png_ptr->background.green / m, g) * m + .5);
						png_ptr->background_1.blue = (png_uint_16)(pow(
							(double)png_ptr->background.blue / m, g) * m + .5);
						png_ptr->background.red = (png_uint_16)(pow(
							(double)png_ptr->background.red / m, gs) * m + .5);
						png_ptr->background.green = (png_uint_16)(pow(
							(double)png_ptr->background.green / m, gs) * m + .5);
						png_ptr->background.blue = (png_uint_16)(pow(
							(double)png_ptr->background.blue / m, gs) * m + .5);
					}
					else
					{
						png_ptr->background_1.red = png_ptr->background_1.green
							= png_ptr->background_1.blue = png_ptr->background_1.gray;
						png_ptr->background.red = png_ptr->background.green
							= png_ptr->background.blue = png_ptr->background.gray;
					}
				}
			}
			else
#endif
				if(color_type == PNG_COLOR_TYPE_PALETTE)
				{
					png_colorp palette = png_ptr->palette;
					int num_palette = png_ptr->num_palette;
					int i;
					for(i = 0; i < num_palette; i++)
					{
						palette[i].red = png_ptr->gamma_table[palette[i].red];
						palette[i].green = png_ptr->gamma_table[palette[i].green];
						palette[i].blue = png_ptr->gamma_table[palette[i].blue];
					}
				}
		}
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
		else
#endif
#endif
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
			if((png_ptr->transformations & PNG_BACKGROUND) &&
				(color_type == PNG_COLOR_TYPE_PALETTE))
			{
				int i;
				int istop = (int)png_ptr->num_trans;
				png_color back;
				png_colorp palette = png_ptr->palette;
				back.red   = (png_byte)png_ptr->background.red;
				back.green = (png_byte)png_ptr->background.green;
				back.blue  = (png_byte)png_ptr->background.blue;
				for(i = 0; i < istop; i++)
				{
					if(png_ptr->trans[i] == 0)
					{
						palette[i] = back;
					}
					else if(png_ptr->trans[i] != 0xff)
					{
						png_composite(palette[i].red, palette[i].red,
							png_ptr->trans[i], back.red);
						png_composite(palette[i].green, palette[i].green,
							png_ptr->trans[i], back.green);
						png_composite(palette[i].blue, palette[i].blue,
							png_ptr->trans[i], back.blue);
					}
				}
			}
#endif
#if defined(PNG_READ_SHIFT_SUPPORTED)
			if((png_ptr->transformations & PNG_SHIFT) &&
				(color_type == PNG_COLOR_TYPE_PALETTE))
			{
				png_uint_16 i;
				png_uint_16 istop = png_ptr->num_palette;
				int sr = 8 - png_ptr->sig_bit.red;
				int sg = 8 - png_ptr->sig_bit.green;
				int sb = 8 - png_ptr->sig_bit.blue;
				if(sr < 0 || sr > 8)
					sr = 0;
				if(sg < 0 || sg > 8)
					sg = 0;
				if(sb < 0 || sb > 8)
					sb = 0;
				for(i = 0; i < istop; i++)
				{
					png_ptr->palette[i].red >>= sr;
					png_ptr->palette[i].green >>= sg;
					png_ptr->palette[i].blue >>= sb;
				}
			}
#endif
	}
#if !defined(PNG_READ_GAMMA_SUPPORTED) && !defined(PNG_READ_SHIFT_SUPPORTED) \
	&& !defined(PNG_READ_BACKGROUND_SUPPORTED)
	if(png_ptr)
		return;
#endif
}
void
png_read_transform_info(png_structp png_ptr, png_infop info_ptr)
{
	png_debug(1, "in png_read_transform_info\n");
#if defined(PNG_READ_EXPAND_SUPPORTED)
	if(png_ptr->transformations & PNG_EXPAND)
	{
		if(info_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		{
			if(png_ptr->num_trans)
				info_ptr->color_type = PNG_COLOR_TYPE_RGB_ALPHA;
			else
				info_ptr->color_type = PNG_COLOR_TYPE_RGB;
			info_ptr->bit_depth = 8;
			info_ptr->num_trans = 0;
		}
		else
		{
			if(png_ptr->num_trans)
				info_ptr->color_type |= PNG_COLOR_MASK_ALPHA;
			if(info_ptr->bit_depth < 8)
				info_ptr->bit_depth = 8;
			info_ptr->num_trans = 0;
		}
	}
#endif
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
	if(png_ptr->transformations & PNG_BACKGROUND)
	{
		info_ptr->color_type &= ~PNG_COLOR_MASK_ALPHA;
		info_ptr->num_trans = 0;
		info_ptr->background = png_ptr->background;
	}
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED)
	if(png_ptr->transformations & PNG_GAMMA)
	{
#ifdef PNG_FLOATING_POINT_SUPPORTED
		info_ptr->gamma = png_ptr->gamma;
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
		info_ptr->int_gamma = png_ptr->int_gamma;
#endif
	}
#endif
#if defined(PNG_READ_16_TO_8_SUPPORTED)
	if((png_ptr->transformations & PNG_16_TO_8) && (info_ptr->bit_depth == 16))
		info_ptr->bit_depth = 8;
#endif
#if defined(PNG_READ_DITHER_SUPPORTED)
	if(png_ptr->transformations & PNG_DITHER)
	{
		if(((info_ptr->color_type == PNG_COLOR_TYPE_RGB) ||
			(info_ptr->color_type == PNG_COLOR_TYPE_RGB_ALPHA)) &&
			png_ptr->palette_lookup && info_ptr->bit_depth == 8)
		{
			info_ptr->color_type = PNG_COLOR_TYPE_PALETTE;
		}
	}
#endif
#if defined(PNG_READ_PACK_SUPPORTED)
	if((png_ptr->transformations & PNG_PACK) && (info_ptr->bit_depth < 8))
		info_ptr->bit_depth = 8;
#endif
#if defined(PNG_READ_GRAY_TO_RGB_SUPPORTED)
	if(png_ptr->transformations & PNG_GRAY_TO_RGB)
		info_ptr->color_type |= PNG_COLOR_MASK_COLOR;
#endif
#if defined(PNG_READ_RGB_TO_GRAY_SUPPORTED)
	if(png_ptr->transformations & PNG_RGB_TO_GRAY)
		info_ptr->color_type &= ~PNG_COLOR_MASK_COLOR;
#endif
	if(info_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		info_ptr->channels = 1;
	else if(info_ptr->color_type & PNG_COLOR_MASK_COLOR)
		info_ptr->channels = 3;
	else
		info_ptr->channels = 1;
#if defined(PNG_READ_STRIP_ALPHA_SUPPORTED)
	if(png_ptr->flags & PNG_FLAG_STRIP_ALPHA)
		info_ptr->color_type &= ~PNG_COLOR_MASK_ALPHA;
#endif
	if(info_ptr->color_type & PNG_COLOR_MASK_ALPHA)
		info_ptr->channels++;
#if defined(PNG_READ_FILLER_SUPPORTED)
	if((png_ptr->transformations & PNG_FILLER) &&
		((info_ptr->color_type == PNG_COLOR_TYPE_RGB) ||
		(info_ptr->color_type == PNG_COLOR_TYPE_GRAY)))
	{
		info_ptr->channels++;
#if !defined(PNG_1_0_X)
		if(png_ptr->transformations & PNG_ADD_ALPHA)
			info_ptr->color_type |= PNG_COLOR_MASK_ALPHA;
#endif
	}
#endif
#if defined(PNG_USER_TRANSFORM_PTR_SUPPORTED) && \
	defined(PNG_READ_USER_TRANSFORM_SUPPORTED)
	if(png_ptr->transformations & PNG_USER_TRANSFORM)
	{
		if(info_ptr->bit_depth < png_ptr->user_transform_depth)
			info_ptr->bit_depth = png_ptr->user_transform_depth;
		if(info_ptr->channels < png_ptr->user_transform_channels)
			info_ptr->channels = png_ptr->user_transform_channels;
	}
#endif
	info_ptr->pixel_depth = (png_byte)(info_ptr->channels *
		info_ptr->bit_depth);
	info_ptr->rowbytes = PNG_ROWBYTES(info_ptr->pixel_depth,info_ptr->width);
#if !defined(PNG_READ_EXPAND_SUPPORTED)
	if(png_ptr)
		return;
#endif
}
void
png_do_read_transformations(png_structp png_ptr)
{
	png_debug(1, "in png_do_read_transformations\n");
#if !defined(PNG_USELESS_TESTS_SUPPORTED)
	if(png_ptr->row_buf == NULL)
	{
#if !defined(PNG_NO_STDIO) && !defined(_WIN32_WCE)
		char msg[50];
		sprintf(msg, "NULL row buffer for row %ld, pass %d", png_ptr->row_number,
			png_ptr->pass);
		png_error(png_ptr, msg);
#else
		png_error(png_ptr, "NULL row buffer");
#endif
	}
#endif
#if defined(PNG_READ_EXPAND_SUPPORTED)
	if(png_ptr->transformations & PNG_EXPAND)
	{
		if(png_ptr->row_info.color_type == PNG_COLOR_TYPE_PALETTE)
		{
			png_do_expand_palette(&(png_ptr->row_info), png_ptr->row_buf + 1,
				png_ptr->palette, png_ptr->trans, png_ptr->num_trans);
		}
		else
		{
			if(png_ptr->num_trans)
				png_do_expand(&(png_ptr->row_info), png_ptr->row_buf + 1,
				&(png_ptr->trans_values));
			else
				png_do_expand(&(png_ptr->row_info), png_ptr->row_buf + 1,
				NULL);
		}
	}
#endif
#if defined(PNG_READ_STRIP_ALPHA_SUPPORTED)
	if(png_ptr->flags & PNG_FLAG_STRIP_ALPHA)
		png_do_strip_filler(&(png_ptr->row_info), png_ptr->row_buf + 1,
		PNG_FLAG_FILLER_AFTER | (png_ptr->flags & PNG_FLAG_STRIP_ALPHA));
#endif
#if defined(PNG_READ_RGB_TO_GRAY_SUPPORTED)
	if(png_ptr->transformations & PNG_RGB_TO_GRAY)
	{
		int rgb_error =
			png_do_rgb_to_gray(png_ptr, &(png_ptr->row_info), png_ptr->row_buf + 1);
		if(rgb_error)
		{
			png_ptr->rgb_to_gray_status = 1;
			if(png_ptr->transformations == PNG_RGB_TO_GRAY_WARN)
				png_warning(png_ptr, "png_do_rgb_to_gray found nongray pixel");
			if(png_ptr->transformations == PNG_RGB_TO_GRAY_ERR)
				png_error(png_ptr, "png_do_rgb_to_gray found nongray pixel");
		}
	}
#endif
#if defined(PNG_READ_GRAY_TO_RGB_SUPPORTED)
	if((png_ptr->transformations & PNG_GRAY_TO_RGB) &&
		!(png_ptr->mode & PNG_BACKGROUND_IS_GRAY))
		png_do_gray_to_rgb(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
	if((png_ptr->transformations & PNG_BACKGROUND) &&
		((png_ptr->num_trans != 0 ) ||
		(png_ptr->color_type & PNG_COLOR_MASK_ALPHA)))
		png_do_background(&(png_ptr->row_info), png_ptr->row_buf + 1,
		&(png_ptr->trans_values), &(png_ptr->background)
#if defined(PNG_READ_GAMMA_SUPPORTED)
		, &(png_ptr->background_1),
		png_ptr->gamma_table, png_ptr->gamma_from_1,
		png_ptr->gamma_to_1, png_ptr->gamma_16_table,
		png_ptr->gamma_16_from_1, png_ptr->gamma_16_to_1,
		png_ptr->gamma_shift
#endif
		);
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED)
	if((png_ptr->transformations & PNG_GAMMA) &&
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
		!((png_ptr->transformations & PNG_BACKGROUND) &&
		((png_ptr->num_trans != 0) ||
		(png_ptr->color_type & PNG_COLOR_MASK_ALPHA))) &&
#endif
		(png_ptr->color_type != PNG_COLOR_TYPE_PALETTE))
		png_do_gamma(&(png_ptr->row_info), png_ptr->row_buf + 1,
		png_ptr->gamma_table, png_ptr->gamma_16_table,
		png_ptr->gamma_shift);
#endif
#if defined(PNG_READ_16_TO_8_SUPPORTED)
	if(png_ptr->transformations & PNG_16_TO_8)
		png_do_chop(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_DITHER_SUPPORTED)
	if(png_ptr->transformations & PNG_DITHER)
	{
		png_do_dither((png_row_infop)&(png_ptr->row_info), png_ptr->row_buf + 1,
			png_ptr->palette_lookup, png_ptr->dither_index);
		if(png_ptr->row_info.rowbytes == (png_uint_32)0)
			png_error(png_ptr, "png_do_dither returned rowbytes = 0");
	}
#endif
#if defined(PNG_READ_INVERT_SUPPORTED)
	if(png_ptr->transformations & PNG_INVERT_MONO)
		png_do_invert(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_SHIFT_SUPPORTED)
	if(png_ptr->transformations & PNG_SHIFT)
		png_do_unshift(&(png_ptr->row_info), png_ptr->row_buf + 1,
		&(png_ptr->shift));
#endif
#if defined(PNG_READ_PACK_SUPPORTED)
	if(png_ptr->transformations & PNG_PACK)
		png_do_unpack(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_BGR_SUPPORTED)
	if(png_ptr->transformations & PNG_BGR)
		png_do_bgr(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
	if(png_ptr->transformations & PNG_PACKSWAP)
		png_do_packswap(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_GRAY_TO_RGB_SUPPORTED)
	if((png_ptr->transformations & PNG_GRAY_TO_RGB) &&
		(png_ptr->mode & PNG_BACKGROUND_IS_GRAY))
		png_do_gray_to_rgb(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_FILLER_SUPPORTED)
	if(png_ptr->transformations & PNG_FILLER)
		png_do_read_filler(&(png_ptr->row_info), png_ptr->row_buf + 1,
		(png_uint_32)png_ptr->filler, png_ptr->flags);
#endif
#if defined(PNG_READ_INVERT_ALPHA_SUPPORTED)
	if(png_ptr->transformations & PNG_INVERT_ALPHA)
		png_do_read_invert_alpha(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_SWAP_ALPHA_SUPPORTED)
	if(png_ptr->transformations & PNG_SWAP_ALPHA)
		png_do_read_swap_alpha(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_SWAP_SUPPORTED)
	if(png_ptr->transformations & PNG_SWAP_BYTES)
		png_do_swap(&(png_ptr->row_info), png_ptr->row_buf + 1);
#endif
#if defined(PNG_READ_USER_TRANSFORM_SUPPORTED)
	if(png_ptr->transformations & PNG_USER_TRANSFORM)
	{
		if(png_ptr->read_user_transform_fn != NULL)
			(*(png_ptr->read_user_transform_fn))
			(png_ptr,
			&(png_ptr->row_info),
			png_ptr->row_buf + 1);
#if defined(PNG_USER_TRANSFORM_PTR_SUPPORTED)
		if(png_ptr->user_transform_depth)
			png_ptr->row_info.bit_depth = png_ptr->user_transform_depth;
		if(png_ptr->user_transform_channels)
			png_ptr->row_info.channels = png_ptr->user_transform_channels;
#endif
		png_ptr->row_info.pixel_depth = (png_byte)(png_ptr->row_info.bit_depth *
			png_ptr->row_info.channels);
		png_ptr->row_info.rowbytes = PNG_ROWBYTES(png_ptr->row_info.pixel_depth,
			png_ptr->row_info.width);
	}
#endif
}
#if defined(PNG_READ_PACK_SUPPORTED)
void
png_do_unpack(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_unpack\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL && row_info->bit_depth < 8)
#else
	if(row_info->bit_depth < 8)
#endif
	{
		png_uint_32 i;
		png_uint_32 row_width = row_info->width;
		switch (row_info->bit_depth)
		{
		case 1:
			{
				png_bytep sp = row + (png_size_t)((row_width - 1) >> 3);
				png_bytep dp = row + (png_size_t)row_width - 1;
				png_uint_32 shift = 7 - (int)((row_width + 7) & 0x07);
				for(i = 0; i < row_width; i++)
				{
					*dp = (png_byte)((*sp >> shift) & 0x01);
					if(shift == 7)
					{
						shift = 0;
						sp--;
					}
					else
						shift++;
					dp--;
				}
				break;
			}
		case 2:
			{
				png_bytep sp = row + (png_size_t)((row_width - 1) >> 2);
				png_bytep dp = row + (png_size_t)row_width - 1;
				png_uint_32 shift = (int)((3 - ((row_width + 3) & 0x03)) << 1);
				for(i = 0; i < row_width; i++)
				{
					*dp = (png_byte)((*sp >> shift) & 0x03);
					if(shift == 6)
					{
						shift = 0;
						sp--;
					}
					else
						shift += 2;
					dp--;
				}
				break;
			}
		case 4:
			{
				png_bytep sp = row + (png_size_t)((row_width - 1) >> 1);
				png_bytep dp = row + (png_size_t)row_width - 1;
				png_uint_32 shift = (int)((1 - ((row_width + 1) & 0x01)) << 2);
				for(i = 0; i < row_width; i++)
				{
					*dp = (png_byte)((*sp >> shift) & 0x0f);
					if(shift == 4)
					{
						shift = 0;
						sp--;
					}
					else
						shift = 4;
					dp--;
				}
				break;
			}
		}
		row_info->bit_depth = 8;
		row_info->pixel_depth = (png_byte)(8 * row_info->channels);
		row_info->rowbytes = row_width * row_info->channels;
	}
}
#endif
#if defined(PNG_READ_SHIFT_SUPPORTED)
void
png_do_unshift(png_row_infop row_info, png_bytep row, png_color_8p sig_bits)
{
	png_debug(1, "in png_do_unshift\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL && sig_bits != NULL &&
#endif
		row_info->color_type != PNG_COLOR_TYPE_PALETTE)
	{
		int shift[4];
		int channels = 0;
		int c;
		png_uint_16 value = 0;
		png_uint_32 row_width = row_info->width;
		if(row_info->color_type & PNG_COLOR_MASK_COLOR)
		{
			shift[channels++] = row_info->bit_depth - sig_bits->red;
			shift[channels++] = row_info->bit_depth - sig_bits->green;
			shift[channels++] = row_info->bit_depth - sig_bits->blue;
		}
		else
		{
			shift[channels++] = row_info->bit_depth - sig_bits->gray;
		}
		if(row_info->color_type & PNG_COLOR_MASK_ALPHA)
		{
			shift[channels++] = row_info->bit_depth - sig_bits->alpha;
		}
		for(c = 0; c < channels; c++)
		{
			if(shift[c] <= 0)
				shift[c] = 0;
			else
				value = 1;
		}
		if(!value)
			return;
		switch (row_info->bit_depth)
		{
		case 2:
			{
				png_bytep bp;
				png_uint_32 i;
				png_uint_32 istop = row_info->rowbytes;
				for(bp = row, i = 0; i < istop; i++)
				{
					*bp >>= 1;
					*bp++ &= 0x55;
				}
				break;
			}
		case 4:
			{
				png_bytep bp = row;
				png_uint_32 i;
				png_uint_32 istop = row_info->rowbytes;
				png_byte mask = (png_byte)((((int)0xf0 >> shift[0]) & (int)0xf0) |
					(png_byte)((int)0xf >> shift[0]));
				for(i = 0; i < istop; i++)
				{
					*bp >>= shift[0];
					*bp++ &= mask;
				}
				break;
			}
		case 8:
			{
				png_bytep bp = row;
				png_uint_32 i;
				png_uint_32 istop = row_width * channels;
				for(i = 0; i < istop; i++)
				{
					*bp++ >>= shift[i%channels];
				}
				break;
			}
		case 16:
			{
				png_bytep bp = row;
				png_uint_32 i;
				png_uint_32 istop = channels * row_width;
				for(i = 0; i < istop; i++)
				{
					value = (png_uint_16)((*bp << 8) + *(bp + 1));
					value >>= shift[i%channels];
					*bp++ = (png_byte)(value >> 8);
					*bp++ = (png_byte)(value & 0xff);
				}
				break;
			}
		}
	}
}
#endif
#if defined(PNG_READ_16_TO_8_SUPPORTED)
void
png_do_chop(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_chop\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL && row_info->bit_depth == 16)
#else
	if(row_info->bit_depth == 16)
#endif
	{
		png_bytep sp = row;
		png_bytep dp = row;
		png_uint_32 i;
		png_uint_32 istop = row_info->width * row_info->channels;
		for(i = 0; i<istop; i++, sp += 2, dp++)
		{
#if defined(PNG_READ_16_TO_8_ACCURATE_SCALE_SUPPORTED)
			*dp = *sp + ((((int)(*(sp + 1)) - *sp) > 128) ? 1 : 0);
#else
			*dp = *sp;
#endif
		}
		row_info->bit_depth = 8;
		row_info->pixel_depth = (png_byte)(8 * row_info->channels);
		row_info->rowbytes = row_info->width * row_info->channels;
	}
}
#endif
#if defined(PNG_READ_SWAP_ALPHA_SUPPORTED)
void
png_do_read_swap_alpha(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_read_swap_alpha\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL)
#endif
	{
		png_uint_32 row_width = row_info->width;
		if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_byte save;
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					save = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = save;
				}
			}
			else
			{
				png_bytep sp = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_byte save[2];
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					save[0] = *(--sp);
					save[1] = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = save[0];
					*(--dp) = save[1];
				}
			}
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_byte save;
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					save = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = save;
				}
			}
			else
			{
				png_bytep sp = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_byte save[2];
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					save[0] = *(--sp);
					save[1] = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = save[0];
					*(--dp) = save[1];
				}
			}
		}
	}
}
#endif
#if defined(PNG_READ_INVERT_ALPHA_SUPPORTED)
void
png_do_read_invert_alpha(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_read_invert_alpha\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL)
#endif
	{
		png_uint_32 row_width = row_info->width;
		if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = (png_byte)(255 - *(--sp));
					sp -= 3;
					dp = sp;
				}
			}
			else
			{
				png_bytep sp = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = (png_byte)(255 - *(--sp));
					*(--dp) = (png_byte)(255 - *(--sp));
					sp -= 6;
					dp = sp;
				}
			}
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = (png_byte)(255 - *(--sp));
					*(--dp) = *(--sp);
				}
			}
			else
			{
				png_bytep sp  = row + row_info->rowbytes;
				png_bytep dp = sp;
				png_uint_32 i;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = (png_byte)(255 - *(--sp));
					*(--dp) = (png_byte)(255 - *(--sp));
					sp -= 2;
					dp = sp;
				}
			}
		}
	}
}
#endif
#if defined(PNG_READ_FILLER_SUPPORTED)
void
png_do_read_filler(png_row_infop row_info, png_bytep row,
				   png_uint_32 filler, png_uint_32 flags)
{
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	png_byte hi_filler = (png_byte)((filler>>8) & 0xff);
	png_byte lo_filler = (png_byte)(filler & 0xff);
	png_debug(1, "in png_do_read_filler\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL  && row_info != NULL &&
#endif
		row_info->color_type == PNG_COLOR_TYPE_GRAY)
	{
		if(row_info->bit_depth == 8)
		{
			if(flags & PNG_FLAG_FILLER_AFTER)
			{
				png_bytep sp = row + (png_size_t)row_width;
				png_bytep dp =  sp + (png_size_t)row_width;
				for(i = 1; i < row_width; i++)
				{
					*(--dp) = lo_filler;
					*(--dp) = *(--sp);
				}
				*(--dp) = lo_filler;
				row_info->channels = 2;
				row_info->pixel_depth = 16;
				row_info->rowbytes = row_width * 2;
			}
			else
			{
				png_bytep sp = row + (png_size_t)row_width;
				png_bytep dp = sp  + (png_size_t)row_width;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = *(--sp);
					*(--dp) = lo_filler;
				}
				row_info->channels = 2;
				row_info->pixel_depth = 16;
				row_info->rowbytes = row_width * 2;
			}
		}
		else if(row_info->bit_depth == 16)
		{
			if(flags & PNG_FLAG_FILLER_AFTER)
			{
				png_bytep sp = row + (png_size_t)row_width * 2;
				png_bytep dp = sp  + (png_size_t)row_width * 2;
				for(i = 1; i < row_width; i++)
				{
					*(--dp) = hi_filler;
					*(--dp) = lo_filler;
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
				}
				*(--dp) = hi_filler;
				*(--dp) = lo_filler;
				row_info->channels = 2;
				row_info->pixel_depth = 32;
				row_info->rowbytes = row_width * 4;
			}
			else
			{
				png_bytep sp = row + (png_size_t)row_width * 2;
				png_bytep dp = sp  + (png_size_t)row_width * 2;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = hi_filler;
					*(--dp) = lo_filler;
				}
				row_info->channels = 2;
				row_info->pixel_depth = 32;
				row_info->rowbytes = row_width * 4;
			}
		}
	}
	else if(row_info->color_type == PNG_COLOR_TYPE_RGB)
	{
		if(row_info->bit_depth == 8)
		{
			if(flags & PNG_FLAG_FILLER_AFTER)
			{
				png_bytep sp = row + (png_size_t)row_width * 3;
				png_bytep dp = sp  + (png_size_t)row_width;
				for(i = 1; i < row_width; i++)
				{
					*(--dp) = lo_filler;
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
				}
				*(--dp) = lo_filler;
				row_info->channels = 4;
				row_info->pixel_depth = 32;
				row_info->rowbytes = row_width * 4;
			}
			else
			{
				png_bytep sp = row + (png_size_t)row_width * 3;
				png_bytep dp = sp + (png_size_t)row_width;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = lo_filler;
				}
				row_info->channels = 4;
				row_info->pixel_depth = 32;
				row_info->rowbytes = row_width * 4;
			}
		}
		else if(row_info->bit_depth == 16)
		{
			if(flags & PNG_FLAG_FILLER_AFTER)
			{
				png_bytep sp = row + (png_size_t)row_width * 6;
				png_bytep dp = sp  + (png_size_t)row_width * 2;
				for(i = 1; i < row_width; i++)
				{
					*(--dp) = hi_filler;
					*(--dp) = lo_filler;
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
				}
				*(--dp) = hi_filler;
				*(--dp) = lo_filler;
				row_info->channels = 4;
				row_info->pixel_depth = 64;
				row_info->rowbytes = row_width * 8;
			}
			else
			{
				png_bytep sp = row + (png_size_t)row_width * 6;
				png_bytep dp = sp  + (png_size_t)row_width * 2;
				for(i = 0; i < row_width; i++)
				{
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = *(--sp);
					*(--dp) = hi_filler;
					*(--dp) = lo_filler;
				}
				row_info->channels = 4;
				row_info->pixel_depth = 64;
				row_info->rowbytes = row_width * 8;
			}
		}
	}
}
#endif
#if defined(PNG_READ_GRAY_TO_RGB_SUPPORTED)
void
png_do_gray_to_rgb(png_row_infop row_info, png_bytep row)
{
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	png_debug(1, "in png_do_gray_to_rgb\n");
	if(row_info->bit_depth >= 8 &&
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		!(row_info->color_type & PNG_COLOR_MASK_COLOR))
	{
		if(row_info->color_type == PNG_COLOR_TYPE_GRAY)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp = row + (png_size_t)row_width - 1;
				png_bytep dp = sp  + (png_size_t)row_width * 2;
				for(i = 0; i < row_width; i++)
				{
					*(dp--) = *sp;
					*(dp--) = *sp;
					*(dp--) = *(sp--);
				}
			}
			else
			{
				png_bytep sp = row + (png_size_t)row_width * 2 - 1;
				png_bytep dp = sp  + (png_size_t)row_width * 4;
				for(i = 0; i < row_width; i++)
				{
					*(dp--) = *sp;
					*(dp--) = *(sp - 1);
					*(dp--) = *sp;
					*(dp--) = *(sp - 1);
					*(dp--) = *(sp--);
					*(dp--) = *(sp--);
				}
			}
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
				png_bytep sp = row + (png_size_t)row_width * 2 - 1;
				png_bytep dp = sp  + (png_size_t)row_width * 2;
				for(i = 0; i < row_width; i++)
				{
					*(dp--) = *(sp--);
					*(dp--) = *sp;
					*(dp--) = *sp;
					*(dp--) = *(sp--);
				}
			}
			else
			{
				png_bytep sp = row + (png_size_t)row_width * 4 - 1;
				png_bytep dp = sp  + (png_size_t)row_width * 4;
				for(i = 0; i < row_width; i++)
				{
					*(dp--) = *(sp--);
					*(dp--) = *(sp--);
					*(dp--) = *sp;
					*(dp--) = *(sp - 1);
					*(dp--) = *sp;
					*(dp--) = *(sp - 1);
					*(dp--) = *(sp--);
					*(dp--) = *(sp--);
				}
			}
		}
		row_info->channels += (png_byte)2;
		row_info->color_type |= PNG_COLOR_MASK_COLOR;
		row_info->pixel_depth = (png_byte)(row_info->channels *
			row_info->bit_depth);
		row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,row_width);
	}
}
#endif
#if defined(PNG_READ_RGB_TO_GRAY_SUPPORTED)
int
png_do_rgb_to_gray(png_structp png_ptr, png_row_infop row_info, png_bytep row)
{
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	int rgb_error = 0;
	png_debug(1, "in png_do_rgb_to_gray\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		(row_info->color_type & PNG_COLOR_MASK_COLOR))
	{
		png_uint_32 rc = png_ptr->rgb_to_gray_red_coeff;
		png_uint_32 gc = png_ptr->rgb_to_gray_green_coeff;
		png_uint_32 bc = png_ptr->rgb_to_gray_blue_coeff;
		if(row_info->color_type == PNG_COLOR_TYPE_RGB)
		{
			if(row_info->bit_depth == 8)
			{
#if defined(PNG_READ_GAMMA_SUPPORTED) || defined(PNG_READ_BACKGROUND_SUPPORTED)
				if(png_ptr->gamma_from_1 != NULL && png_ptr->gamma_to_1 != NULL)
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_byte red   = png_ptr->gamma_to_1[*(sp++)];
						png_byte green = png_ptr->gamma_to_1[*(sp++)];
						png_byte blue  = png_ptr->gamma_to_1[*(sp++)];
						if(red != green || red != blue)
						{
							rgb_error |= 1;
							*(dp++) = png_ptr->gamma_from_1[
								(rc*red+gc*green+bc*blue)>>15];
						}
						else
							*(dp++) = *(sp-1);
					}
				}
				else
#endif
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_byte red   = *(sp++);
						png_byte green = *(sp++);
						png_byte blue  = *(sp++);
						if(red != green || red != blue)
						{
							rgb_error |= 1;
							*(dp++) = (png_byte)((rc*red+gc*green+bc*blue)>>15);
						}
						else
							*(dp++) = *(sp-1);
					}
				}
			}
			else
			{
#if defined(PNG_READ_GAMMA_SUPPORTED) || defined(PNG_READ_BACKGROUND_SUPPORTED)
				if(png_ptr->gamma_16_to_1 != NULL &&
					png_ptr->gamma_16_from_1 != NULL)
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 red, green, blue, w;
						red   = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						green = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						blue  = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						if(red == green && red == blue)
							w = red;
						else
						{
							png_uint_16 red_1   = png_ptr->gamma_16_to_1[(red&0xff) >>
								png_ptr->gamma_shift][red>>8];
							png_uint_16 green_1 = png_ptr->gamma_16_to_1[(green&0xff) >>
								png_ptr->gamma_shift][green>>8];
							png_uint_16 blue_1  = png_ptr->gamma_16_to_1[(blue&0xff) >>
								png_ptr->gamma_shift][blue>>8];
							png_uint_16 gray16  = (png_uint_16)((rc*red_1 + gc*green_1
								+ bc*blue_1)>>15);
							w = png_ptr->gamma_16_from_1[(gray16&0xff) >>
								png_ptr->gamma_shift][gray16 >> 8];
							rgb_error |= 1;
						}
						*(dp++) = (png_byte)((w>>8) & 0xff);
						*(dp++) = (png_byte)(w & 0xff);
					}
				}
				else
#endif
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 red, green, blue, gray16;
						red   = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						green = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						blue  = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						if(red != green || red != blue)
							rgb_error |= 1;
						gray16  = (png_uint_16)((rc*red + gc*green + bc*blue)>>15);
						*(dp++) = (png_byte)((gray16>>8) & 0xff);
						*(dp++) = (png_byte)(gray16 & 0xff);
					}
				}
			}
		}
		if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
		{
			if(row_info->bit_depth == 8)
			{
#if defined(PNG_READ_GAMMA_SUPPORTED) || defined(PNG_READ_BACKGROUND_SUPPORTED)
				if(png_ptr->gamma_from_1 != NULL && png_ptr->gamma_to_1 != NULL)
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_byte red   = png_ptr->gamma_to_1[*(sp++)];
						png_byte green = png_ptr->gamma_to_1[*(sp++)];
						png_byte blue  = png_ptr->gamma_to_1[*(sp++)];
						if(red != green || red != blue)
							rgb_error |= 1;
						*(dp++) =  png_ptr->gamma_from_1
							[(rc*red + gc*green + bc*blue)>>15];
						*(dp++) = *(sp++);
					}
				}
				else
#endif
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_byte red   = *(sp++);
						png_byte green = *(sp++);
						png_byte blue  = *(sp++);
						if(red != green || red != blue)
							rgb_error |= 1;
						*(dp++) =  (png_byte)((rc*red + gc*green + bc*blue)>>15);
						*(dp++) = *(sp++);
					}
				}
			}
			else
			{
#if defined(PNG_READ_GAMMA_SUPPORTED) || defined(PNG_READ_BACKGROUND_SUPPORTED)
				if(png_ptr->gamma_16_to_1 != NULL &&
					png_ptr->gamma_16_from_1 != NULL)
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 red, green, blue, w;
						red   = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						green = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						blue  = (png_uint_16)(((*(sp))<<8) | *(sp+1)); sp += 2;
						if(red == green && red == blue)
							w = red;
						else
						{
							png_uint_16 red_1   = png_ptr->gamma_16_to_1[(red&0xff) >>
								png_ptr->gamma_shift][red>>8];
							png_uint_16 green_1 = png_ptr->gamma_16_to_1[(green&0xff) >>
								png_ptr->gamma_shift][green>>8];
							png_uint_16 blue_1  = png_ptr->gamma_16_to_1[(blue&0xff) >>
								png_ptr->gamma_shift][blue>>8];
							png_uint_16 gray16  = (png_uint_16)((rc * red_1
								+ gc * green_1 + bc * blue_1)>>15);
							w = png_ptr->gamma_16_from_1[(gray16&0xff) >>
								png_ptr->gamma_shift][gray16 >> 8];
							rgb_error |= 1;
						}
						*(dp++) = (png_byte)((w>>8) & 0xff);
						*(dp++) = (png_byte)(w & 0xff);
						*(dp++) = *(sp++);
						*(dp++) = *(sp++);
					}
				}
				else
#endif
				{
					png_bytep sp = row;
					png_bytep dp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 red, green, blue, gray16;
						red   = (png_uint_16)((*(sp)<<8) | *(sp+1)); sp += 2;
						green = (png_uint_16)((*(sp)<<8) | *(sp+1)); sp += 2;
						blue  = (png_uint_16)((*(sp)<<8) | *(sp+1)); sp += 2;
						if(red != green || red != blue)
							rgb_error |= 1;
						gray16  = (png_uint_16)((rc*red + gc*green + bc*blue)>>15);
						*(dp++) = (png_byte)((gray16>>8) & 0xff);
						*(dp++) = (png_byte)(gray16 & 0xff);
						*(dp++) = *(sp++);
						*(dp++) = *(sp++);
					}
				}
			}
		}
		row_info->channels -= (png_byte)2;
		row_info->color_type &= ~PNG_COLOR_MASK_COLOR;
		row_info->pixel_depth = (png_byte)(row_info->channels *
			row_info->bit_depth);
		row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,row_width);
	}
	return rgb_error;
}
#endif
void PNGAPI
png_build_grayscale_palette(int bit_depth, png_colorp palette)
{
	int num_palette;
	int color_inc;
	int i;
	int v;
	png_debug(1, "in png_do_build_grayscale_palette\n");
	if(palette == NULL)
		return;
	switch (bit_depth)
	{
	case 1:
		num_palette = 2;
		color_inc = 0xff;
		break;
	case 2:
		num_palette = 4;
		color_inc = 0x55;
		break;
	case 4:
		num_palette = 16;
		color_inc = 0x11;
		break;
	case 8:
		num_palette = 256;
		color_inc = 1;
		break;
	default:
		num_palette = 0;
		color_inc = 0;
		break;
	}
	for(i = 0, v = 0; i < num_palette; i++, v += color_inc)
	{
		palette[i].red = (png_byte)v;
		palette[i].green = (png_byte)v;
		palette[i].blue = (png_byte)v;
	}
}
#if defined(PNG_READ_DITHER_SUPPORTED) && defined(PNG_CORRECT_PALETTE_SUPPORTED)
void
png_correct_palette(png_structp png_ptr, png_colorp palette,
					int num_palette)
{
	png_debug(1, "in png_correct_palette\n");
#if defined(PNG_READ_BACKGROUND_SUPPORTED) && \
	defined(PNG_READ_GAMMA_SUPPORTED) && defined(PNG_FLOATING_POINT_SUPPORTED)
	if(png_ptr->transformations & (PNG_GAMMA | PNG_BACKGROUND))
	{
		png_color back, back_1;
		if(png_ptr->background_gamma_type == PNG_BACKGROUND_GAMMA_FILE)
		{
			back.red = png_ptr->gamma_table[png_ptr->background.red];
			back.green = png_ptr->gamma_table[png_ptr->background.green];
			back.blue = png_ptr->gamma_table[png_ptr->background.blue];
			back_1.red = png_ptr->gamma_to_1[png_ptr->background.red];
			back_1.green = png_ptr->gamma_to_1[png_ptr->background.green];
			back_1.blue = png_ptr->gamma_to_1[png_ptr->background.blue];
		}
		else
		{
			double g;
			g = 1.0 / (png_ptr->background_gamma * png_ptr->screen_gamma);
			if(png_ptr->background_gamma_type == PNG_BACKGROUND_GAMMA_SCREEN ||
				fabs(g - 1.0) < PNG_GAMMA_THRESHOLD)
			{
				back.red = png_ptr->background.red;
				back.green = png_ptr->background.green;
				back.blue = png_ptr->background.blue;
			}
			else
			{
				back.red =
					(png_byte)(pow((double)png_ptr->background.red/255, g) *
					255.0 + 0.5);
				back.green =
					(png_byte)(pow((double)png_ptr->background.green/255, g) *
					255.0 + 0.5);
				back.blue =
					(png_byte)(pow((double)png_ptr->background.blue/255, g) *
					255.0 + 0.5);
			}
			g = 1.0 / png_ptr->background_gamma;
			back_1.red =
				(png_byte)(pow((double)png_ptr->background.red/255, g) *
				255.0 + 0.5);
			back_1.green =
				(png_byte)(pow((double)png_ptr->background.green/255, g) *
				255.0 + 0.5);
			back_1.blue =
				(png_byte)(pow((double)png_ptr->background.blue/255, g) *
				255.0 + 0.5);
		}
		if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		{
			png_uint_32 i;
			for(i = 0; i < (png_uint_32)num_palette; i++)
			{
				if(i < png_ptr->num_trans && png_ptr->trans[i] == 0)
				{
					palette[i] = back;
				}
				else if(i < png_ptr->num_trans && png_ptr->trans[i] != 0xff)
				{
					png_byte v, w;
					v = png_ptr->gamma_to_1[png_ptr->palette[i].red];
					png_composite(w, v, png_ptr->trans[i], back_1.red);
					palette[i].red = png_ptr->gamma_from_1[w];
					v = png_ptr->gamma_to_1[png_ptr->palette[i].green];
					png_composite(w, v, png_ptr->trans[i], back_1.green);
					palette[i].green = png_ptr->gamma_from_1[w];
					v = png_ptr->gamma_to_1[png_ptr->palette[i].blue];
					png_composite(w, v, png_ptr->trans[i], back_1.blue);
					palette[i].blue = png_ptr->gamma_from_1[w];
				}
				else
				{
					palette[i].red = png_ptr->gamma_table[palette[i].red];
					palette[i].green = png_ptr->gamma_table[palette[i].green];
					palette[i].blue = png_ptr->gamma_table[palette[i].blue];
				}
			}
		}
		else
		{
			int i;
			for(i = 0; i < num_palette; i++)
			{
				if(palette[i].red == (png_byte)png_ptr->trans_values.gray)
				{
					palette[i] = back;
				}
				else
				{
					palette[i].red = png_ptr->gamma_table[palette[i].red];
					palette[i].green = png_ptr->gamma_table[palette[i].green];
					palette[i].blue = png_ptr->gamma_table[palette[i].blue];
				}
			}
		}
	}
	else
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED)
		if(png_ptr->transformations & PNG_GAMMA)
		{
			int i;
			for(i = 0; i < num_palette; i++)
			{
				palette[i].red = png_ptr->gamma_table[palette[i].red];
				palette[i].green = png_ptr->gamma_table[palette[i].green];
				palette[i].blue = png_ptr->gamma_table[palette[i].blue];
			}
		}
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
		else
#endif
#endif
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
			if(png_ptr->transformations & PNG_BACKGROUND)
			{
				if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
				{
					png_color back;
					back.red   = (png_byte)png_ptr->background.red;
					back.green = (png_byte)png_ptr->background.green;
					back.blue  = (png_byte)png_ptr->background.blue;
					for(i = 0; i < (int)png_ptr->num_trans; i++)
					{
						if(png_ptr->trans[i] == 0)
						{
							palette[i].red = back.red;
							palette[i].green = back.green;
							palette[i].blue = back.blue;
						}
						else if(png_ptr->trans[i] != 0xff)
						{
							png_composite(palette[i].red, png_ptr->palette[i].red,
								png_ptr->trans[i], back.red);
							png_composite(palette[i].green, png_ptr->palette[i].green,
								png_ptr->trans[i], back.green);
							png_composite(palette[i].blue, png_ptr->palette[i].blue,
								png_ptr->trans[i], back.blue);
						}
					}
				}
				else
				{
					int i;
					for(i = 0; i < num_palette; i++)
					{
						if(i == (png_byte)png_ptr->trans_values.gray)
						{
							palette[i].red = (png_byte)png_ptr->background.red;
							palette[i].green = (png_byte)png_ptr->background.green;
							palette[i].blue = (png_byte)png_ptr->background.blue;
						}
					}
				}
			}
#endif
}
#endif
#if defined(PNG_READ_BACKGROUND_SUPPORTED)
void
png_do_background(png_row_infop row_info, png_bytep row,
				  png_color_16p trans_values, png_color_16p background
#if defined(PNG_READ_GAMMA_SUPPORTED)
				  , png_color_16p background_1,
				  png_bytep gamma_table, png_bytep gamma_from_1, png_bytep gamma_to_1,
				  png_uint_16pp gamma_16, png_uint_16pp gamma_16_from_1,
				  png_uint_16pp gamma_16_to_1, int gamma_shift
#endif
				  )
{
	png_bytep sp, dp;
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	int shift;
	png_debug(1, "in png_do_background\n");
	if(background != NULL &&
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		(!(row_info->color_type & PNG_COLOR_MASK_ALPHA) ||
		(row_info->color_type != PNG_COLOR_TYPE_PALETTE && trans_values)))
	{
		switch (row_info->color_type)
		{
		case PNG_COLOR_TYPE_GRAY:
			{
				switch (row_info->bit_depth)
				{
				case 1:
					{
						sp = row;
						shift = 7;
						for(i = 0; i < row_width; i++)
						{
							if((png_uint_16)((*sp >> shift) & 0x01)
								== trans_values->gray)
							{
								*sp &= (png_byte)((0x7f7f >> (7 - shift)) & 0xff);
								*sp |= (png_byte)(background->gray << shift);
							}
							if(!shift)
							{
								shift = 7;
								sp++;
							}
							else
								shift--;
						}
						break;
					}
				case 2:
					{
#if defined(PNG_READ_GAMMA_SUPPORTED)
						if(gamma_table != NULL)
						{
							sp = row;
							shift = 6;
							for(i = 0; i < row_width; i++)
							{
								if((png_uint_16)((*sp >> shift) & 0x03)
									== trans_values->gray)
								{
									*sp &= (png_byte)((0x3f3f >> (6 - shift)) & 0xff);
									*sp |= (png_byte)(background->gray << shift);
								}
								else
								{
									png_byte p = (png_byte)((*sp >> shift) & 0x03);
									png_byte g = (png_byte)((gamma_table [p | (p << 2) |
										(p << 4) | (p << 6)] >> 6) & 0x03);
									*sp &= (png_byte)((0x3f3f >> (6 - shift)) & 0xff);
									*sp |= (png_byte)(g << shift);
								}
								if(!shift)
								{
									shift = 6;
									sp++;
								}
								else
									shift -= 2;
							}
						}
						else
#endif
						{
							sp = row;
							shift = 6;
							for(i = 0; i < row_width; i++)
							{
								if((png_uint_16)((*sp >> shift) & 0x03)
									== trans_values->gray)
								{
									*sp &= (png_byte)((0x3f3f >> (6 - shift)) & 0xff);
									*sp |= (png_byte)(background->gray << shift);
								}
								if(!shift)
								{
									shift = 6;
									sp++;
								}
								else
									shift -= 2;
							}
						}
						break;
					}
				case 4:
					{
#if defined(PNG_READ_GAMMA_SUPPORTED)
						if(gamma_table != NULL)
						{
							sp = row;
							shift = 4;
							for(i = 0; i < row_width; i++)
							{
								if((png_uint_16)((*sp >> shift) & 0x0f)
									== trans_values->gray)
								{
									*sp &= (png_byte)((0xf0f >> (4 - shift)) & 0xff);
									*sp |= (png_byte)(background->gray << shift);
								}
								else
								{
									png_byte p = (png_byte)((*sp >> shift) & 0x0f);
									png_byte g = (png_byte)((gamma_table[p |
										(p << 4)] >> 4) & 0x0f);
									*sp &= (png_byte)((0xf0f >> (4 - shift)) & 0xff);
									*sp |= (png_byte)(g << shift);
								}
								if(!shift)
								{
									shift = 4;
									sp++;
								}
								else
									shift -= 4;
							}
						}
						else
#endif
						{
							sp = row;
							shift = 4;
							for(i = 0; i < row_width; i++)
							{
								if((png_uint_16)((*sp >> shift) & 0x0f)
									== trans_values->gray)
								{
									*sp &= (png_byte)((0xf0f >> (4 - shift)) & 0xff);
									*sp |= (png_byte)(background->gray << shift);
								}
								if(!shift)
								{
									shift = 4;
									sp++;
								}
								else
									shift -= 4;
							}
						}
						break;
					}
				case 8:
					{
#if defined(PNG_READ_GAMMA_SUPPORTED)
						if(gamma_table != NULL)
						{
							sp = row;
							for(i = 0; i < row_width; i++, sp++)
							{
								if(*sp == trans_values->gray)
								{
									*sp = (png_byte)background->gray;
								}
								else
								{
									*sp = gamma_table[*sp];
								}
							}
						}
						else
#endif
						{
							sp = row;
							for(i = 0; i < row_width; i++, sp++)
							{
								if(*sp == trans_values->gray)
								{
									*sp = (png_byte)background->gray;
								}
							}
						}
						break;
					}
				case 16:
					{
#if defined(PNG_READ_GAMMA_SUPPORTED)
						if(gamma_16 != NULL)
						{
							sp = row;
							for(i = 0; i < row_width; i++, sp += 2)
							{
								png_uint_16 v;
								v = (png_uint_16)(((*sp) << 8) + *(sp + 1));
								if(v == trans_values->gray)
								{
									*sp = (png_byte)((background->gray >> 8) & 0xff);
									*(sp + 1) = (png_byte)(background->gray & 0xff);
								}
								else
								{
									v = gamma_16[*(sp + 1) >> gamma_shift][*sp];
									*sp = (png_byte)((v >> 8) & 0xff);
									*(sp + 1) = (png_byte)(v & 0xff);
								}
							}
						}
						else
#endif
						{
							sp = row;
							for(i = 0; i < row_width; i++, sp += 2)
							{
								png_uint_16 v;
								v = (png_uint_16)(((*sp) << 8) + *(sp + 1));
								if(v == trans_values->gray)
								{
									*sp = (png_byte)((background->gray >> 8) & 0xff);
									*(sp + 1) = (png_byte)(background->gray & 0xff);
								}
							}
						}
						break;
					}
				}
				break;
			}
		case PNG_COLOR_TYPE_RGB:
			{
				if(row_info->bit_depth == 8)
				{
#if defined(PNG_READ_GAMMA_SUPPORTED)
					if(gamma_table != NULL)
					{
						sp = row;
						for(i = 0; i < row_width; i++, sp += 3)
						{
							if(*sp == trans_values->red &&
								*(sp + 1) == trans_values->green &&
								*(sp + 2) == trans_values->blue)
							{
								*sp = (png_byte)background->red;
								*(sp + 1) = (png_byte)background->green;
								*(sp + 2) = (png_byte)background->blue;
							}
							else
							{
								*sp = gamma_table[*sp];
								*(sp + 1) = gamma_table[*(sp + 1)];
								*(sp + 2) = gamma_table[*(sp + 2)];
							}
						}
					}
					else
#endif
					{
						sp = row;
						for(i = 0; i < row_width; i++, sp += 3)
						{
							if(*sp == trans_values->red &&
								*(sp + 1) == trans_values->green &&
								*(sp + 2) == trans_values->blue)
							{
								*sp = (png_byte)background->red;
								*(sp + 1) = (png_byte)background->green;
								*(sp + 2) = (png_byte)background->blue;
							}
						}
					}
				}
				else
				{
#if defined(PNG_READ_GAMMA_SUPPORTED)
					if(gamma_16 != NULL)
					{
						sp = row;
						for(i = 0; i < row_width; i++, sp += 6)
						{
							png_uint_16 r = (png_uint_16)(((*sp) << 8) + *(sp + 1));
							png_uint_16 g = (png_uint_16)(((*(sp+2)) << 8) + *(sp+3));
							png_uint_16 b = (png_uint_16)(((*(sp+4)) << 8) + *(sp+5));
							if(r == trans_values->red && g == trans_values->green &&
								b == trans_values->blue)
							{
								*sp = (png_byte)((background->red >> 8) & 0xff);
								*(sp + 1) = (png_byte)(background->red & 0xff);
								*(sp + 2) = (png_byte)((background->green >> 8) & 0xff);
								*(sp + 3) = (png_byte)(background->green & 0xff);
								*(sp + 4) = (png_byte)((background->blue >> 8) & 0xff);
								*(sp + 5) = (png_byte)(background->blue & 0xff);
							}
							else
							{
								png_uint_16 v = gamma_16[*(sp + 1) >> gamma_shift][*sp];
								*sp = (png_byte)((v >> 8) & 0xff);
								*(sp + 1) = (png_byte)(v & 0xff);
								v = gamma_16[*(sp + 3) >> gamma_shift][*(sp + 2)];
								*(sp + 2) = (png_byte)((v >> 8) & 0xff);
								*(sp + 3) = (png_byte)(v & 0xff);
								v = gamma_16[*(sp + 5) >> gamma_shift][*(sp + 4)];
								*(sp + 4) = (png_byte)((v >> 8) & 0xff);
								*(sp + 5) = (png_byte)(v & 0xff);
							}
						}
					}
					else
#endif
					{
						sp = row;
						for(i = 0; i < row_width; i++, sp += 6)
						{
							png_uint_16 r = (png_uint_16)(((*sp) << 8) + *(sp+1));
							png_uint_16 g = (png_uint_16)(((*(sp+2)) << 8) + *(sp+3));
							png_uint_16 b = (png_uint_16)(((*(sp+4)) << 8) + *(sp+5));
							if(r == trans_values->red && g == trans_values->green &&
								b == trans_values->blue)
							{
								*sp = (png_byte)((background->red >> 8) & 0xff);
								*(sp + 1) = (png_byte)(background->red & 0xff);
								*(sp + 2) = (png_byte)((background->green >> 8) & 0xff);
								*(sp + 3) = (png_byte)(background->green & 0xff);
								*(sp + 4) = (png_byte)((background->blue >> 8) & 0xff);
								*(sp + 5) = (png_byte)(background->blue & 0xff);
							}
						}
					}
				}
				break;
			}
		case PNG_COLOR_TYPE_GRAY_ALPHA:
			{
				if(row_info->bit_depth == 8)
				{
#if defined(PNG_READ_GAMMA_SUPPORTED)
					if(gamma_to_1 != NULL && gamma_from_1 != NULL &&
						gamma_table != NULL)
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 2, dp++)
						{
							png_uint_16 a = *(sp + 1);
							if(a == 0xff)
							{
								*dp = gamma_table[*sp];
							}
							else if(a == 0)
							{
								*dp = (png_byte)background->gray;
							}
							else
							{
								png_byte v, w;
								v = gamma_to_1[*sp];
								png_composite(w, v, a, background_1->gray);
								*dp = gamma_from_1[w];
							}
						}
					}
					else
#endif
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 2, dp++)
						{
							png_byte a = *(sp + 1);
							if(a == 0xff)
							{
								*dp = *sp;
							}
#if defined(PNG_READ_GAMMA_SUPPORTED)
							else if(a == 0)
							{
								*dp = (png_byte)background->gray;
							}
							else
							{
								png_composite(*dp, *sp, a, background_1->gray);
							}
#else
							*dp = (png_byte)background->gray;
#endif
						}
					}
				}
				else
				{
#if defined(PNG_READ_GAMMA_SUPPORTED)
					if(gamma_16 != NULL && gamma_16_from_1 != NULL &&
						gamma_16_to_1 != NULL)
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 4, dp += 2)
						{
							png_uint_16 a = (png_uint_16)(((*(sp+2)) << 8) + *(sp+3));
							if(a == (png_uint_16)0xffff)
							{
								png_uint_16 v;
								v = gamma_16[*(sp + 1) >> gamma_shift][*sp];
								*dp = (png_byte)((v >> 8) & 0xff);
								*(dp + 1) = (png_byte)(v & 0xff);
							}
#if defined(PNG_READ_GAMMA_SUPPORTED)
							else if(a == 0)
#else
							else
#endif
							{
								*dp = (png_byte)((background->gray >> 8) & 0xff);
								*(dp + 1) = (png_byte)(background->gray & 0xff);
							}
#if defined(PNG_READ_GAMMA_SUPPORTED)
					else
					{
						png_uint_16 g, v, w;
						g = gamma_16_to_1[*(sp + 1) >> gamma_shift][*sp];
						png_composite_16(v, g, a, background_1->gray);
						w = gamma_16_from_1[(v&0xff) >> gamma_shift][v >> 8];
						*dp = (png_byte)((w >> 8) & 0xff);
						*(dp + 1) = (png_byte)(w & 0xff);
					}
#endif
						}
					}
					else
#endif
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 4, dp += 2)
						{
							png_uint_16 a = (png_uint_16)(((*(sp+2)) << 8) + *(sp+3));
							if(a == (png_uint_16)0xffff)
							{
								png_memcpy(dp, sp, 2);
							}
#if defined(PNG_READ_GAMMA_SUPPORTED)
							else if(a == 0)
#else
							else
#endif
							{
								*dp = (png_byte)((background->gray >> 8) & 0xff);
								*(dp + 1) = (png_byte)(background->gray & 0xff);
							}
#if defined(PNG_READ_GAMMA_SUPPORTED)
	else
	{
		png_uint_16 g, v;
		g = (png_uint_16)(((*sp) << 8) + *(sp + 1));
		png_composite_16(v, g, a, background_1->gray);
		*dp = (png_byte)((v >> 8) & 0xff);
		*(dp + 1) = (png_byte)(v & 0xff);
	}
#endif
						}
					}
				}
				break;
			}
		case PNG_COLOR_TYPE_RGB_ALPHA:
			{
				if(row_info->bit_depth == 8)
				{
#if defined(PNG_READ_GAMMA_SUPPORTED)
					if(gamma_to_1 != NULL && gamma_from_1 != NULL &&
						gamma_table != NULL)
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 4, dp += 3)
						{
							png_byte a = *(sp + 3);
							if(a == 0xff)
							{
								*dp = gamma_table[*sp];
								*(dp + 1) = gamma_table[*(sp + 1)];
								*(dp + 2) = gamma_table[*(sp + 2)];
							}
							else if(a == 0)
							{
								*dp = (png_byte)background->red;
								*(dp + 1) = (png_byte)background->green;
								*(dp + 2) = (png_byte)background->blue;
							}
							else
							{
								png_byte v, w;
								v = gamma_to_1[*sp];
								png_composite(w, v, a, background_1->red);
								*dp = gamma_from_1[w];
								v = gamma_to_1[*(sp + 1)];
								png_composite(w, v, a, background_1->green);
								*(dp + 1) = gamma_from_1[w];
								v = gamma_to_1[*(sp + 2)];
								png_composite(w, v, a, background_1->blue);
								*(dp + 2) = gamma_from_1[w];
							}
						}
					}
					else
#endif
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 4, dp += 3)
						{
							png_byte a = *(sp + 3);
							if(a == 0xff)
							{
								*dp = *sp;
								*(dp + 1) = *(sp + 1);
								*(dp + 2) = *(sp + 2);
							}
							else if(a == 0)
							{
								*dp = (png_byte)background->red;
								*(dp + 1) = (png_byte)background->green;
								*(dp + 2) = (png_byte)background->blue;
							}
							else
							{
								png_composite(*dp, *sp, a, background->red);
								png_composite(*(dp + 1), *(sp + 1), a,
									background->green);
								png_composite(*(dp + 2), *(sp + 2), a,
									background->blue);
							}
						}
					}
				}
				else
				{
#if defined(PNG_READ_GAMMA_SUPPORTED)
					if(gamma_16 != NULL && gamma_16_from_1 != NULL &&
						gamma_16_to_1 != NULL)
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 8, dp += 6)
						{
							png_uint_16 a = (png_uint_16)(((png_uint_16)(*(sp + 6))
								<< 8) + (png_uint_16)(*(sp + 7)));
							if(a == (png_uint_16)0xffff)
							{
								png_uint_16 v;
								v = gamma_16[*(sp + 1) >> gamma_shift][*sp];
								*dp = (png_byte)((v >> 8) & 0xff);
								*(dp + 1) = (png_byte)(v & 0xff);
								v = gamma_16[*(sp + 3) >> gamma_shift][*(sp + 2)];
								*(dp + 2) = (png_byte)((v >> 8) & 0xff);
								*(dp + 3) = (png_byte)(v & 0xff);
								v = gamma_16[*(sp + 5) >> gamma_shift][*(sp + 4)];
								*(dp + 4) = (png_byte)((v >> 8) & 0xff);
								*(dp + 5) = (png_byte)(v & 0xff);
							}
							else if(a == 0)
							{
								*dp = (png_byte)((background->red >> 8) & 0xff);
								*(dp + 1) = (png_byte)(background->red & 0xff);
								*(dp + 2) = (png_byte)((background->green >> 8) & 0xff);
								*(dp + 3) = (png_byte)(background->green & 0xff);
								*(dp + 4) = (png_byte)((background->blue >> 8) & 0xff);
								*(dp + 5) = (png_byte)(background->blue & 0xff);
							}
							else
							{
								png_uint_16 v, w, x;
								v = gamma_16_to_1[*(sp + 1) >> gamma_shift][*sp];
								png_composite_16(w, v, a, background_1->red);
								x = gamma_16_from_1[((w&0xff) >> gamma_shift)][w >> 8];
								*dp = (png_byte)((x >> 8) & 0xff);
								*(dp + 1) = (png_byte)(x & 0xff);
								v = gamma_16_to_1[*(sp + 3) >> gamma_shift][*(sp + 2)];
								png_composite_16(w, v, a, background_1->green);
								x = gamma_16_from_1[((w&0xff) >> gamma_shift)][w >> 8];
								*(dp + 2) = (png_byte)((x >> 8) & 0xff);
								*(dp + 3) = (png_byte)(x & 0xff);
								v = gamma_16_to_1[*(sp + 5) >> gamma_shift][*(sp + 4)];
								png_composite_16(w, v, a, background_1->blue);
								x = gamma_16_from_1[(w & 0xff) >> gamma_shift][w >> 8];
								*(dp + 4) = (png_byte)((x >> 8) & 0xff);
								*(dp + 5) = (png_byte)(x & 0xff);
							}
						}
					}
					else
#endif
					{
						sp = row;
						dp = row;
						for(i = 0; i < row_width; i++, sp += 8, dp += 6)
						{
							png_uint_16 a = (png_uint_16)(((png_uint_16)(*(sp + 6))
								<< 8) + (png_uint_16)(*(sp + 7)));
							if(a == (png_uint_16)0xffff)
							{
								png_memcpy(dp, sp, 6);
							}
							else if(a == 0)
							{
								*dp = (png_byte)((background->red >> 8) & 0xff);
								*(dp + 1) = (png_byte)(background->red & 0xff);
								*(dp + 2) = (png_byte)((background->green >> 8) & 0xff);
								*(dp + 3) = (png_byte)(background->green & 0xff);
								*(dp + 4) = (png_byte)((background->blue >> 8) & 0xff);
								*(dp + 5) = (png_byte)(background->blue & 0xff);
							}
							else
							{
								png_uint_16 v;
								png_uint_16 r = (png_uint_16)(((*sp) << 8) + *(sp + 1));
								png_uint_16 g = (png_uint_16)(((*(sp + 2)) << 8)
									+ *(sp + 3));
								png_uint_16 b = (png_uint_16)(((*(sp + 4)) << 8)
									+ *(sp + 5));
								png_composite_16(v, r, a, background->red);
								*dp = (png_byte)((v >> 8) & 0xff);
								*(dp + 1) = (png_byte)(v & 0xff);
								png_composite_16(v, g, a, background->green);
								*(dp + 2) = (png_byte)((v >> 8) & 0xff);
								*(dp + 3) = (png_byte)(v & 0xff);
								png_composite_16(v, b, a, background->blue);
								*(dp + 4) = (png_byte)((v >> 8) & 0xff);
								*(dp + 5) = (png_byte)(v & 0xff);
							}
						}
					}
				}
				break;
			}
		}
		if(row_info->color_type & PNG_COLOR_MASK_ALPHA)
		{
			row_info->color_type &= ~PNG_COLOR_MASK_ALPHA;
			row_info->channels--;
			row_info->pixel_depth = (png_byte)(row_info->channels *
				row_info->bit_depth);
			row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,row_width);
		}
	}
}
#endif
#if defined(PNG_READ_GAMMA_SUPPORTED)
void
png_do_gamma(png_row_infop row_info, png_bytep row,
			 png_bytep gamma_table, png_uint_16pp gamma_16_table,
			 int gamma_shift)
{
	png_bytep sp;
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	png_debug(1, "in png_do_gamma\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		((row_info->bit_depth <= 8 && gamma_table != NULL) ||
		(row_info->bit_depth == 16 && gamma_16_table != NULL)))
	{
		switch (row_info->color_type)
		{
		case PNG_COLOR_TYPE_RGB:
			{
				if(row_info->bit_depth == 8)
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						*sp = gamma_table[*sp];
						sp++;
						*sp = gamma_table[*sp];
						sp++;
						*sp = gamma_table[*sp];
						sp++;
					}
				}
				else
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 v;
						v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 2;
						v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 2;
						v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 2;
					}
				}
				break;
			}
		case PNG_COLOR_TYPE_RGB_ALPHA:
			{
				if(row_info->bit_depth == 8)
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						*sp = gamma_table[*sp];
						sp++;
						*sp = gamma_table[*sp];
						sp++;
						*sp = gamma_table[*sp];
						sp++;
						sp++;
					}
				}
				else
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 2;
						v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 2;
						v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 4;
					}
				}
				break;
			}
		case PNG_COLOR_TYPE_GRAY_ALPHA:
			{
				if(row_info->bit_depth == 8)
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						*sp = gamma_table[*sp];
						sp += 2;
					}
				}
				else
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 4;
					}
				}
				break;
			}
		case PNG_COLOR_TYPE_GRAY:
			{
				if(row_info->bit_depth == 2)
				{
					sp = row;
					for(i = 0; i < row_width; i += 4)
					{
						int a = *sp & 0xc0;
						int b = *sp & 0x30;
						int c = *sp & 0x0c;
						int d = *sp & 0x03;
						*sp = (png_byte)(
							((((int)gamma_table[a|(a>>2)|(a>>4)|(a>>6)])   ) & 0xc0)|
							((((int)gamma_table[(b<<2)|b|(b>>2)|(b>>4)])>>2) & 0x30)|
							((((int)gamma_table[(c<<4)|(c<<2)|c|(c>>2)])>>4) & 0x0c)|
							((((int)gamma_table[(d<<6)|(d<<4)|(d<<2)|d])>>6) ));
						sp++;
					}
				}
				if(row_info->bit_depth == 4)
				{
					sp = row;
					for(i = 0; i < row_width; i += 2)
					{
						int msb = *sp & 0xf0;
						int lsb = *sp & 0x0f;
						*sp = (png_byte)((((int)gamma_table[msb | (msb >> 4)]) & 0xf0)
							| (((int)gamma_table[(lsb << 4) | lsb]) >> 4));
						sp++;
					}
				}
				else if(row_info->bit_depth == 8)
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						*sp = gamma_table[*sp];
						sp++;
					}
				}
				else if(row_info->bit_depth == 16)
				{
					sp = row;
					for(i = 0; i < row_width; i++)
					{
						png_uint_16 v = gamma_16_table[*(sp + 1) >> gamma_shift][*sp];
						*sp = (png_byte)((v >> 8) & 0xff);
						*(sp + 1) = (png_byte)(v & 0xff);
						sp += 2;
					}
				}
				break;
			}
		}
	}
}
#endif
#if defined(PNG_READ_EXPAND_SUPPORTED)
void
png_do_expand_palette(png_row_infop row_info, png_bytep row,
					  png_colorp palette, png_bytep trans, int num_trans)
{
	int shift, value;
	png_bytep sp, dp;
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	png_debug(1, "in png_do_expand_palette\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		row_info->color_type == PNG_COLOR_TYPE_PALETTE)
	{
		if(row_info->bit_depth < 8)
		{
			switch (row_info->bit_depth)
			{
			case 1:
				{
					sp = row + (png_size_t)((row_width - 1) >> 3);
					dp = row + (png_size_t)row_width - 1;
					shift = 7 - (int)((row_width + 7) & 0x07);
					for(i = 0; i < row_width; i++)
					{
						if((*sp >> shift) & 0x01)
							*dp = 1;
						else
							*dp = 0;
						if(shift == 7)
						{
							shift = 0;
							sp--;
						}
						else
							shift++;
						dp--;
					}
					break;
				}
			case 2:
				{
					sp = row + (png_size_t)((row_width - 1) >> 2);
					dp = row + (png_size_t)row_width - 1;
					shift = (int)((3 - ((row_width + 3) & 0x03)) << 1);
					for(i = 0; i < row_width; i++)
					{
						value = (*sp >> shift) & 0x03;
						*dp = (png_byte)value;
						if(shift == 6)
						{
							shift = 0;
							sp--;
						}
						else
							shift += 2;
						dp--;
					}
					break;
				}
			case 4:
				{
					sp = row + (png_size_t)((row_width - 1) >> 1);
					dp = row + (png_size_t)row_width - 1;
					shift = (int)((row_width & 0x01) << 2);
					for(i = 0; i < row_width; i++)
					{
						value = (*sp >> shift) & 0x0f;
						*dp = (png_byte)value;
						if(shift == 4)
						{
							shift = 0;
							sp--;
						}
						else
							shift += 4;
						dp--;
					}
					break;
				}
			}
			row_info->bit_depth = 8;
			row_info->pixel_depth = 8;
			row_info->rowbytes = row_width;
		}
		switch (row_info->bit_depth)
		{
		case 8:
			{
				if(trans != NULL)
				{
					sp = row + (png_size_t)row_width - 1;
					dp = row + (png_size_t)(row_width << 2) - 1;
					for(i = 0; i < row_width; i++)
					{
						if((int)(*sp) >= num_trans)
							*dp-- = 0xff;
						else
							*dp-- = trans[*sp];
						*dp-- = palette[*sp].blue;
						*dp-- = palette[*sp].green;
						*dp-- = palette[*sp].red;
						sp--;
					}
					row_info->bit_depth = 8;
					row_info->pixel_depth = 32;
					row_info->rowbytes = row_width * 4;
					row_info->color_type = 6;
					row_info->channels = 4;
				}
				else
				{
					sp = row + (png_size_t)row_width - 1;
					dp = row + (png_size_t)(row_width * 3) - 1;
					for(i = 0; i < row_width; i++)
					{
						*dp-- = palette[*sp].blue;
						*dp-- = palette[*sp].green;
						*dp-- = palette[*sp].red;
						sp--;
					}
					row_info->bit_depth = 8;
					row_info->pixel_depth = 24;
					row_info->rowbytes = row_width * 3;
					row_info->color_type = 2;
					row_info->channels = 3;
				}
				break;
			}
		}
	}
}
void
png_do_expand(png_row_infop row_info, png_bytep row,
			  png_color_16p trans_value)
{
	int shift, value;
	png_bytep sp, dp;
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	png_debug(1, "in png_do_expand\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL)
#endif
	{
		if(row_info->color_type == PNG_COLOR_TYPE_GRAY)
		{
			png_uint_16 gray = (png_uint_16)(trans_value ? trans_value->gray : 0);
			if(row_info->bit_depth < 8)
			{
				switch (row_info->bit_depth)
				{
				case 1:
					{
						gray = (png_uint_16)(gray*0xff);
						sp = row + (png_size_t)((row_width - 1) >> 3);
						dp = row + (png_size_t)row_width - 1;
						shift = 7 - (int)((row_width + 7) & 0x07);
						for(i = 0; i < row_width; i++)
						{
							if((*sp >> shift) & 0x01)
								*dp = 0xff;
							else
								*dp = 0;
							if(shift == 7)
							{
								shift = 0;
								sp--;
							}
							else
								shift++;
							dp--;
						}
						break;
					}
				case 2:
					{
						gray = (png_uint_16)(gray*0x55);
						sp = row + (png_size_t)((row_width - 1) >> 2);
						dp = row + (png_size_t)row_width - 1;
						shift = (int)((3 - ((row_width + 3) & 0x03)) << 1);
						for(i = 0; i < row_width; i++)
						{
							value = (*sp >> shift) & 0x03;
							*dp = (png_byte)(value | (value << 2) | (value << 4) |
								(value << 6));
							if(shift == 6)
							{
								shift = 0;
								sp--;
							}
							else
								shift += 2;
							dp--;
						}
						break;
					}
				case 4:
					{
						gray = (png_uint_16)(gray*0x11);
						sp = row + (png_size_t)((row_width - 1) >> 1);
						dp = row + (png_size_t)row_width - 1;
						shift = (int)((1 - ((row_width + 1) & 0x01)) << 2);
						for(i = 0; i < row_width; i++)
						{
							value = (*sp >> shift) & 0x0f;
							*dp = (png_byte)(value | (value << 4));
							if(shift == 4)
							{
								shift = 0;
								sp--;
							}
							else
								shift = 4;
							dp--;
						}
						break;
					}
				}
				row_info->bit_depth = 8;
				row_info->pixel_depth = 8;
				row_info->rowbytes = row_width;
			}
			if(trans_value != NULL)
			{
				if(row_info->bit_depth == 8)
				{
					sp = row + (png_size_t)row_width - 1;
					dp = row + (png_size_t)(row_width << 1) - 1;
					for(i = 0; i < row_width; i++)
					{
						if(*sp == gray)
							*dp-- = 0;
						else
							*dp-- = 0xff;
						*dp-- = *sp--;
					}
				}
				else if(row_info->bit_depth == 16)
				{
					sp = row + row_info->rowbytes - 1;
					dp = row + (row_info->rowbytes << 1) - 1;
					for(i = 0; i < row_width; i++)
					{
						if(((png_uint_16)*(sp) |
							((png_uint_16)*(sp - 1) << 8)) == gray)
						{
							*dp-- = 0;
							*dp-- = 0;
						}
						else
						{
							*dp-- = 0xff;
							*dp-- = 0xff;
						}
						*dp-- = *sp--;
						*dp-- = *sp--;
					}
				}
				row_info->color_type = PNG_COLOR_TYPE_GRAY_ALPHA;
				row_info->channels = 2;
				row_info->pixel_depth = (png_byte)(row_info->bit_depth << 1);
				row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,
					row_width);
			}
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_RGB && trans_value)
		{
			if(row_info->bit_depth == 8)
			{
				sp = row + (png_size_t)row_info->rowbytes - 1;
				dp = row + (png_size_t)(row_width << 2) - 1;
				for(i = 0; i < row_width; i++)
				{
					if(*(sp - 2) == trans_value->red &&
						*(sp - 1) == trans_value->green &&
						*(sp - 0) == trans_value->blue)
						*dp-- = 0;
					else
						*dp-- = 0xff;
					*dp-- = *sp--;
					*dp-- = *sp--;
					*dp-- = *sp--;
				}
			}
			else if(row_info->bit_depth == 16)
			{
				sp = row + row_info->rowbytes - 1;
				dp = row + (png_size_t)(row_width << 3) - 1;
				for(i = 0; i < row_width; i++)
				{
					if((((png_uint_16)*(sp - 4) |
						((png_uint_16)*(sp - 5) << 8)) == trans_value->red) &&
						(((png_uint_16)*(sp - 2) |
						((png_uint_16)*(sp - 3) << 8)) == trans_value->green) &&
						(((png_uint_16)*(sp - 0) |
						((png_uint_16)*(sp - 1) << 8)) == trans_value->blue))
					{
						*dp-- = 0;
						*dp-- = 0;
					}
					else
					{
						*dp-- = 0xff;
						*dp-- = 0xff;
					}
					*dp-- = *sp--;
					*dp-- = *sp--;
					*dp-- = *sp--;
					*dp-- = *sp--;
					*dp-- = *sp--;
					*dp-- = *sp--;
				}
			}
			row_info->color_type = PNG_COLOR_TYPE_RGB_ALPHA;
			row_info->channels = 4;
			row_info->pixel_depth = (png_byte)(row_info->bit_depth << 2);
			row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,row_width);
		}
	}
}
#endif
#if defined(PNG_READ_DITHER_SUPPORTED)
void
png_do_dither(png_row_infop row_info, png_bytep row,
			  png_bytep palette_lookup, png_bytep dither_lookup)
{
	png_bytep sp, dp;
	png_uint_32 i;
	png_uint_32 row_width = row_info->width;
	png_debug(1, "in png_do_dither\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL)
#endif
	{
		if(row_info->color_type == PNG_COLOR_TYPE_RGB &&
			palette_lookup && row_info->bit_depth == 8)
		{
			int r, g, b, p;
			sp = row;
			dp = row;
			for(i = 0; i < row_width; i++)
			{
				r = *sp++;
				g = *sp++;
				b = *sp++;
				p = (((r >> (8 - PNG_DITHER_RED_BITS)) &
					((1 << PNG_DITHER_RED_BITS) - 1)) <<
					(PNG_DITHER_GREEN_BITS + PNG_DITHER_BLUE_BITS)) |
					(((g >> (8 - PNG_DITHER_GREEN_BITS)) &
					((1 << PNG_DITHER_GREEN_BITS) - 1)) <<
					(PNG_DITHER_BLUE_BITS)) |
					((b >> (8 - PNG_DITHER_BLUE_BITS)) &
					((1 << PNG_DITHER_BLUE_BITS) - 1));
				*dp++ = palette_lookup[p];
			}
			row_info->color_type = PNG_COLOR_TYPE_PALETTE;
			row_info->channels = 1;
			row_info->pixel_depth = row_info->bit_depth;
			row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,row_width);
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA &&
			palette_lookup != NULL && row_info->bit_depth == 8)
		{
			int r, g, b, p;
			sp = row;
			dp = row;
			for(i = 0; i < row_width; i++)
			{
				r = *sp++;
				g = *sp++;
				b = *sp++;
				sp++;
				p = (((r >> (8 - PNG_DITHER_RED_BITS)) &
					((1 << PNG_DITHER_RED_BITS) - 1)) <<
					(PNG_DITHER_GREEN_BITS + PNG_DITHER_BLUE_BITS)) |
					(((g >> (8 - PNG_DITHER_GREEN_BITS)) &
					((1 << PNG_DITHER_GREEN_BITS) - 1)) <<
					(PNG_DITHER_BLUE_BITS)) |
					((b >> (8 - PNG_DITHER_BLUE_BITS)) &
					((1 << PNG_DITHER_BLUE_BITS) - 1));
				*dp++ = palette_lookup[p];
			}
			row_info->color_type = PNG_COLOR_TYPE_PALETTE;
			row_info->channels = 1;
			row_info->pixel_depth = row_info->bit_depth;
			row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,row_width);
		}
		else if(row_info->color_type == PNG_COLOR_TYPE_PALETTE &&
			dither_lookup && row_info->bit_depth == 8)
		{
			sp = row;
			for(i = 0; i < row_width; i++, sp++)
			{
				*sp = dither_lookup[*sp];
			}
		}
	}
}
#endif
#ifdef PNG_FLOATING_POINT_SUPPORTED
#if defined(PNG_READ_GAMMA_SUPPORTED)
static int png_gamma_shift[] =
{0x10, 0x21, 0x42, 0x84, 0x110, 0x248, 0x550, 0xff0};
void
png_build_gamma_table(png_structp png_ptr)
{
	png_debug(1, "in png_build_gamma_table\n");
	if(png_ptr->gamma != 0.0)
	{
		if(png_ptr->bit_depth <= 8)
		{
			int i;
			double g;
			if(png_ptr->screen_gamma > .000001)
				g = 1.0 / (png_ptr->gamma * png_ptr->screen_gamma);
			else
				g = 1.0;
			png_ptr->gamma_table = (png_bytep)png_malloc(png_ptr,
				(png_uint_32)256);
			for(i = 0; i < 256; i++)
			{
				png_ptr->gamma_table[i] = (png_byte)(pow((double)i / 255.0,
					g) * 255.0 + .5);
			}
#if defined(PNG_READ_BACKGROUND_SUPPORTED) || \
	defined(PNG_READ_RGB_TO_GRAY_SUPPORTED)
			if(png_ptr->transformations & ((PNG_BACKGROUND) | PNG_RGB_TO_GRAY))
			{
				g = 1.0 / (png_ptr->gamma);
				png_ptr->gamma_to_1 = (png_bytep)png_malloc(png_ptr,
					(png_uint_32)256);
				for(i = 0; i < 256; i++)
				{
					png_ptr->gamma_to_1[i] = (png_byte)(pow((double)i / 255.0,
						g) * 255.0 + .5);
				}
				png_ptr->gamma_from_1 = (png_bytep)png_malloc(png_ptr,
					(png_uint_32)256);
				if(png_ptr->screen_gamma > 0.000001)
					g = 1.0 / png_ptr->screen_gamma;
				else
					g = png_ptr->gamma;
				for(i = 0; i < 256; i++)
				{
					png_ptr->gamma_from_1[i] = (png_byte)(pow((double)i / 255.0,
						g) * 255.0 + .5);
				}
			}
#endif
		}
		else
		{
			double g;
			int i, j, shift, num;
			int sig_bit;
			png_uint_32 ig;
			if(png_ptr->color_type & PNG_COLOR_MASK_COLOR)
			{
				sig_bit = (int)png_ptr->sig_bit.red;
				if((int)png_ptr->sig_bit.green > sig_bit)
					sig_bit = png_ptr->sig_bit.green;
				if((int)png_ptr->sig_bit.blue > sig_bit)
					sig_bit = png_ptr->sig_bit.blue;
			}
			else
			{
				sig_bit = (int)png_ptr->sig_bit.gray;
			}
			if(sig_bit > 0)
				shift = 16 - sig_bit;
			else
				shift = 0;
			if(png_ptr->transformations & PNG_16_TO_8)
			{
				if(shift < (16 - PNG_MAX_GAMMA_8))
					shift = (16 - PNG_MAX_GAMMA_8);
			}
			if(shift > 8)
				shift = 8;
			if(shift < 0)
				shift = 0;
			png_ptr->gamma_shift = (png_byte)shift;
			num = (1 << (8 - shift));
			if(png_ptr->screen_gamma > .000001)
				g = 1.0 / (png_ptr->gamma * png_ptr->screen_gamma);
			else
				g = 1.0;
			png_ptr->gamma_16_table = (png_uint_16pp)png_malloc(png_ptr,
				(png_uint_32)(num * png_sizeof (png_uint_16p)));
			if(png_ptr->transformations & (PNG_16_TO_8 | PNG_BACKGROUND))
			{
				double fin, fout;
				png_uint_32 last, max;
				for(i = 0; i < num; i++)
				{
					png_ptr->gamma_16_table[i] = (png_uint_16p)png_malloc(png_ptr,
						(png_uint_32)(256 * png_sizeof (png_uint_16)));
				}
				g = 1.0 / g;
				last = 0;
				for(i = 0; i < 256; i++)
				{
					fout = ((double)i + 0.5) / 256.0;
					fin = pow(fout, g);
					max = (png_uint_32)(fin * (double)((png_uint_32)num << 8));
					while(last <= max)
					{
						png_ptr->gamma_16_table[(int)(last & (0xff >> shift))]
						[(int)(last >> (8 - shift))] = (png_uint_16)(
							(png_uint_16)i | ((png_uint_16)i << 8));
						last++;
					}
				}
				while(last < ((png_uint_32)num << 8))
				{
					png_ptr->gamma_16_table[(int)(last & (0xff >> shift))]
					[(int)(last >> (8 - shift))] = (png_uint_16)65535L;
					last++;
				}
			}
			else
			{
				for(i = 0; i < num; i++)
				{
					png_ptr->gamma_16_table[i] = (png_uint_16p)png_malloc(png_ptr,
						(png_uint_32)(256 * png_sizeof (png_uint_16)));
					ig = (((png_uint_32)i * (png_uint_32)png_gamma_shift[shift]) >> 4);
					for(j = 0; j < 256; j++)
					{
						png_ptr->gamma_16_table[i][j] =
							(png_uint_16)(pow((double)(ig + ((png_uint_32)j << 8)) /
							65535.0, g) * 65535.0 + .5);
					}
				}
			}
#if defined(PNG_READ_BACKGROUND_SUPPORTED) || \
	defined(PNG_READ_RGB_TO_GRAY_SUPPORTED)
			if(png_ptr->transformations & (PNG_BACKGROUND | PNG_RGB_TO_GRAY))
			{
				g = 1.0 / (png_ptr->gamma);
				png_ptr->gamma_16_to_1 = (png_uint_16pp)png_malloc(png_ptr,
					(png_uint_32)(num * png_sizeof (png_uint_16p )));
				for(i = 0; i < num; i++)
				{
					png_ptr->gamma_16_to_1[i] = (png_uint_16p)png_malloc(png_ptr,
						(png_uint_32)(256 * png_sizeof (png_uint_16)));
					ig = (((png_uint_32)i *
						(png_uint_32)png_gamma_shift[shift]) >> 4);
					for(j = 0; j < 256; j++)
					{
						png_ptr->gamma_16_to_1[i][j] =
							(png_uint_16)(pow((double)(ig + ((png_uint_32)j << 8)) /
							65535.0, g) * 65535.0 + .5);
					}
				}
				if(png_ptr->screen_gamma > 0.000001)
					g = 1.0 / png_ptr->screen_gamma;
				else
					g = png_ptr->gamma;
				png_ptr->gamma_16_from_1 = (png_uint_16pp)png_malloc(png_ptr,
					(png_uint_32)(num * png_sizeof (png_uint_16p)));
				for(i = 0; i < num; i++)
				{
					png_ptr->gamma_16_from_1[i] = (png_uint_16p)png_malloc(png_ptr,
						(png_uint_32)(256 * png_sizeof (png_uint_16)));
					ig = (((png_uint_32)i *
						(png_uint_32)png_gamma_shift[shift]) >> 4);
					for(j = 0; j < 256; j++)
					{
						png_ptr->gamma_16_from_1[i][j] =
							(png_uint_16)(pow((double)(ig + ((png_uint_32)j << 8)) /
							65535.0, g) * 65535.0 + .5);
					}
				}
			}
#endif
		}
	}
}
#endif
#endif
#if defined(PNG_MNG_FEATURES_SUPPORTED)
void
png_do_read_intrapixel(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_read_intrapixel\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		(row_info->color_type & PNG_COLOR_MASK_COLOR))
	{
		int bytes_per_pixel;
		png_uint_32 row_width = row_info->width;
		if(row_info->bit_depth == 8)
		{
			png_bytep rp;
			png_uint_32 i;
			if(row_info->color_type == PNG_COLOR_TYPE_RGB)
				bytes_per_pixel = 3;
			else if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
				bytes_per_pixel = 4;
			else
				return;
			for(i = 0, rp = row; i < row_width; i++, rp += bytes_per_pixel)
			{
				*(rp) = (png_byte)((256 + *rp + *(rp+1))&0xff);
				*(rp+2) = (png_byte)((256 + *(rp+2) + *(rp+1))&0xff);
			}
		}
		else if(row_info->bit_depth == 16)
		{
			png_bytep rp;
			png_uint_32 i;
			if(row_info->color_type == PNG_COLOR_TYPE_RGB)
				bytes_per_pixel = 6;
			else if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
				bytes_per_pixel = 8;
			else
				return;
			for(i = 0, rp = row; i < row_width; i++, rp += bytes_per_pixel)
			{
				png_uint_32 s0   = (*(rp  ) << 8) | *(rp+1);
				png_uint_32 s1   = (*(rp+2) << 8) | *(rp+3);
				png_uint_32 s2   = (*(rp+4) << 8) | *(rp+5);
				png_uint_32 red  = (png_uint_32)((s0+s1+65536L) & 0xffffL);
				png_uint_32 blue = (png_uint_32)((s2+s1+65536L) & 0xffffL);
				*(rp  ) = (png_byte)((red >> 8) & 0xff);
				*(rp+1) = (png_byte)(red & 0xff);
				*(rp+4) = (png_byte)((blue >> 8) & 0xff);
				*(rp+5) = (png_byte)(blue & 0xff);
			}
		}
	}
}
#endif
#define PNG_INTERNAL
#include "png.h"
#if defined(_WIN32_WCE)
#  ifdef PNG_FLOATING_POINT_SUPPORTED
__inline double strtod(const char *nptr, char **endptr)
{
	double result = 0;
	int len;
	wchar_t *str, *end;
	len = MultiByteToWideChar(CP_ACP, 0, nptr, -1, NULL, 0);
	str = (wchar_t *)malloc(len * sizeof(wchar_t));
	if( NULL != str )
	{
		MultiByteToWideChar(CP_ACP, 0, nptr, -1, str, len);
		result = wcstod(str, &end);
		len = WideCharToMultiByte(CP_ACP, 0, end, -1, NULL, 0, NULL, NULL);
		*endptr = (char *)nptr + (png_strlen(nptr) - len + 1);
		free(str);
	}
	return result;
}
#  endif
#endif
png_uint_32
png_get_uint_31(png_structp png_ptr, png_bytep buf)
{
	png_uint_32 i = png_get_uint_32(buf);
	if(i > PNG_UINT_31_MAX)
		png_error(png_ptr, "PNG unsigned integer out of range.\n");
	return (i);
}
#ifndef PNG_READ_BIG_ENDIAN_SUPPORTED
png_uint_32
png_get_uint_32(png_bytep buf)
{
	png_uint_32 i = ((png_uint_32)(*buf) << 24) +
		((png_uint_32)(*(buf + 1)) << 16) +
		((png_uint_32)(*(buf + 2)) << 8) +
		(png_uint_32)(*(buf + 3));
	return (i);
}
#if defined(PNG_READ_pCAL_SUPPORTED) || defined(PNG_READ_oFFs_SUPPORTED)
png_int_32
png_get_int_32(png_bytep buf)
{
	png_int_32 i = ((png_int_32)(*buf) << 24) +
		((png_int_32)(*(buf + 1)) << 16) +
		((png_int_32)(*(buf + 2)) << 8) +
		(png_int_32)(*(buf + 3));
	return (i);
}
#endif
png_uint_16
png_get_uint_16(png_bytep buf)
{
	png_uint_16 i = (png_uint_16)(((png_uint_16)(*buf) << 8) +
		(png_uint_16)(*(buf + 1)));
	return (i);
}
#endif
void
png_crc_read(png_structp png_ptr, png_bytep buf, png_size_t length)
{
	png_read_data(png_ptr, buf, length);
	png_calculate_crc(png_ptr, buf, length);
}
int
png_crc_finish(png_structp png_ptr, png_uint_32 skip)
{
	png_size_t i;
	png_size_t istop = png_ptr->zbuf_size;
	for(i = (png_size_t)skip; i > istop; i -= istop)
	{
		png_crc_read(png_ptr, png_ptr->zbuf, png_ptr->zbuf_size);
	}
	if(i)
	{
		png_crc_read(png_ptr, png_ptr->zbuf, i);
	}
	if(png_crc_error(png_ptr))
	{
		if(((png_ptr->chunk_name[0] & 0x20) &&
			!(png_ptr->flags & PNG_FLAG_CRC_ANCILLARY_NOWARN)) ||
			(!(png_ptr->chunk_name[0] & 0x20) &&
			(png_ptr->flags & PNG_FLAG_CRC_CRITICAL_USE)))
		{
			png_chunk_warning(png_ptr, "CRC error");
		}
		else
		{
			png_chunk_error(png_ptr, "CRC error");
		}
		return (1);
	}
	return (0);
}
int
png_crc_error(png_structp png_ptr)
{
	png_byte crc_bytes[4];
	png_uint_32 crc;
	int need_crc = 1;
	if(png_ptr->chunk_name[0] & 0x20)
	{
		if((png_ptr->flags & PNG_FLAG_CRC_ANCILLARY_MASK) ==
			(PNG_FLAG_CRC_ANCILLARY_USE | PNG_FLAG_CRC_ANCILLARY_NOWARN))
			need_crc = 0;
	}
	else
	{
		if(png_ptr->flags & PNG_FLAG_CRC_CRITICAL_IGNORE)
			need_crc = 0;
	}
	png_read_data(png_ptr, crc_bytes, 4);
	if(need_crc)
	{
		crc = png_get_uint_32(crc_bytes);
		return ((int)(crc != png_ptr->crc));
	}
	else
		return (0);
}
#if defined(PNG_READ_zTXt_SUPPORTED) || defined(PNG_READ_iTXt_SUPPORTED) || \
	defined(PNG_READ_iCCP_SUPPORTED)
png_charp
png_decompress_chunk(png_structp png_ptr, int comp_type,
					 png_charp chunkdata, png_size_t chunklength,
					 png_size_t prefix_size, png_size_t *newlength)
{
	static char msg[] = "Error decoding compressed text";
	png_charp text;
	png_size_t text_size;
	if(comp_type == PNG_COMPRESSION_TYPE_BASE)
	{
		int ret = Z_OK;
		png_ptr->zstream.next_in = (png_bytep)(chunkdata + prefix_size);
		png_ptr->zstream.avail_in = (uInt)(chunklength - prefix_size);
		png_ptr->zstream.next_out = png_ptr->zbuf;
		png_ptr->zstream.avail_out = (uInt)png_ptr->zbuf_size;
		text_size = 0;
		text = NULL;
		while(png_ptr->zstream.avail_in)
		{
			ret = inflate(&png_ptr->zstream, Z_PARTIAL_FLUSH);
			if(ret != Z_OK && ret != Z_STREAM_END)
			{
				if(png_ptr->zstream.msg != NULL)
					png_warning(png_ptr, png_ptr->zstream.msg);
				else
					png_warning(png_ptr, msg);
				inflateReset(&png_ptr->zstream);
				png_ptr->zstream.avail_in = 0;
				if(text ==  NULL)
				{
					text_size = prefix_size + png_sizeof(msg) + 1;
					text = (png_charp)png_malloc_warn(png_ptr, (png_uint_32)(text_size));
					if(text ==  NULL)
					{
						png_free(png_ptr,chunkdata);
						png_error(png_ptr,"Not enough memory to decompress chunk");
					}
					png_memcpy(text, chunkdata, prefix_size);
				}
				text[text_size - 1] = 0x00;
				text_size = (png_size_t)(chunklength - (text - chunkdata) - 1);
				text_size = png_sizeof(msg) > text_size ? text_size :
				png_sizeof(msg);
				png_memcpy(text + prefix_size, msg, text_size + 1);
				break;
			}
			if(!png_ptr->zstream.avail_out || ret == Z_STREAM_END)
			{
				if(text == NULL)
				{
					text_size = prefix_size +
						png_ptr->zbuf_size - png_ptr->zstream.avail_out;
					text = (png_charp)png_malloc_warn(png_ptr, (png_uint_32)(text_size + 1));
					if(text ==  NULL)
					{
						png_free(png_ptr,chunkdata);
						png_error(png_ptr,"Not enough memory to decompress chunk.");
					}
					png_memcpy(text + prefix_size, png_ptr->zbuf,
						text_size - prefix_size);
					png_memcpy(text, chunkdata, prefix_size);
					*(text + text_size) = 0x00;
				}
				else
				{
					png_charp tmp;
					tmp = text;
					text = (png_charp)png_malloc_warn(png_ptr,
						(png_uint_32)(text_size +
						png_ptr->zbuf_size - png_ptr->zstream.avail_out + 1));
					if(text == NULL)
					{
						png_free(png_ptr, tmp);
						png_free(png_ptr, chunkdata);
						png_error(png_ptr,"Not enough memory to decompress chunk..");
					}
					png_memcpy(text, tmp, text_size);
					png_free(png_ptr, tmp);
					png_memcpy(text + text_size, png_ptr->zbuf,
						(png_ptr->zbuf_size - png_ptr->zstream.avail_out));
					text_size += png_ptr->zbuf_size - png_ptr->zstream.avail_out;
					*(text + text_size) = 0x00;
				}
				if(ret == Z_STREAM_END)
					break;
				else
				{
					png_ptr->zstream.next_out = png_ptr->zbuf;
					png_ptr->zstream.avail_out = (uInt)png_ptr->zbuf_size;
				}
			}
		}
		if(ret != Z_STREAM_END)
		{
#if !defined(PNG_NO_STDIO) && !defined(_WIN32_WCE)
			char umsg[50];
			if(ret == Z_BUF_ERROR)
				sprintf(umsg,"Buffer error in compressed datastream in %s chunk",
				png_ptr->chunk_name);
			else if(ret == Z_DATA_ERROR)
				sprintf(umsg,"Data error in compressed datastream in %s chunk",
				png_ptr->chunk_name);
			else
				sprintf(umsg,"Incomplete compressed datastream in %s chunk",
				png_ptr->chunk_name);
			png_warning(png_ptr, umsg);
#else
			png_warning(png_ptr,
				"Incomplete compressed datastream in chunk other than IDAT");
#endif
			text_size = prefix_size;
			if(text ==  NULL)
			{
				text = (png_charp)png_malloc_warn(png_ptr, (png_uint_32)(text_size+1));
				if(text == NULL)
				{
					png_free(png_ptr, chunkdata);
					png_error(png_ptr,"Not enough memory for text.");
				}
				png_memcpy(text, chunkdata, prefix_size);
			}
			*(text + text_size) = 0x00;
		}
		inflateReset(&png_ptr->zstream);
		png_ptr->zstream.avail_in = 0;
		png_free(png_ptr, chunkdata);
		chunkdata = text;
		*newlength = text_size;
	}
	else
	{
#if !defined(PNG_NO_STDIO) && !defined(_WIN32_WCE)
		char umsg[50];
		sprintf(umsg, "Unknown zTXt compression type %d", comp_type);
		png_warning(png_ptr, umsg);
#else
		png_warning(png_ptr, "Unknown zTXt compression type");
#endif
		*(chunkdata + prefix_size) = 0x00;
		*newlength = prefix_size;
	}
	return chunkdata;
}
#endif
void
png_handle_IHDR(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_byte buf[13];
	png_uint_32 width, height;
	int bit_depth, color_type, compression_type, filter_type;
	int interlace_type;
	png_debug(1, "in png_handle_IHDR\n");
	if(png_ptr->mode & PNG_HAVE_IHDR)
		png_error(png_ptr, "Out of place IHDR");
	if(length != 13)
		png_error(png_ptr, "Invalid IHDR chunk");
	png_ptr->mode |= PNG_HAVE_IHDR;
	png_crc_read(png_ptr, buf, 13);
	png_crc_finish(png_ptr, 0);
	width = png_get_uint_31(png_ptr, buf);
	height = png_get_uint_31(png_ptr, buf + 4);
	bit_depth = buf[8];
	color_type = buf[9];
	compression_type = buf[10];
	filter_type = buf[11];
	interlace_type = buf[12];
	png_ptr->width = width;
	png_ptr->height = height;
	png_ptr->bit_depth = (png_byte)bit_depth;
	png_ptr->interlaced = (png_byte)interlace_type;
	png_ptr->color_type = (png_byte)color_type;
#if defined(PNG_MNG_FEATURES_SUPPORTED)
	png_ptr->filter_type = (png_byte)filter_type;
#endif
	png_ptr->compression_type = (png_byte)compression_type;
	switch (png_ptr->color_type)
	{
	case PNG_COLOR_TYPE_GRAY:
	case PNG_COLOR_TYPE_PALETTE:
		png_ptr->channels = 1;
		break;
	case PNG_COLOR_TYPE_RGB:
		png_ptr->channels = 3;
		break;
	case PNG_COLOR_TYPE_GRAY_ALPHA:
		png_ptr->channels = 2;
		break;
	case PNG_COLOR_TYPE_RGB_ALPHA:
		png_ptr->channels = 4;
		break;
	}
	png_ptr->pixel_depth = (png_byte)(png_ptr->bit_depth *
		png_ptr->channels);
	png_ptr->rowbytes = PNG_ROWBYTES(png_ptr->pixel_depth,png_ptr->width);
	png_debug1(3,"bit_depth = %d\n", png_ptr->bit_depth);
	png_debug1(3,"channels = %d\n", png_ptr->channels);
	png_debug1(3,"rowbytes = %lu\n", png_ptr->rowbytes);
	png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth,
		color_type, interlace_type, compression_type, filter_type);
}
void
png_handle_PLTE(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_color palette[PNG_MAX_PALETTE_LENGTH];
	int num, i;
#ifndef PNG_NO_POINTER_INDEXING
	png_colorp pal_ptr;
#endif
	png_debug(1, "in png_handle_PLTE\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before PLTE");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid PLTE after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(png_ptr->mode & PNG_HAVE_PLTE)
		png_error(png_ptr, "Duplicate PLTE chunk");
	png_ptr->mode |= PNG_HAVE_PLTE;
	if(!(png_ptr->color_type&PNG_COLOR_MASK_COLOR))
	{
		png_warning(png_ptr,
			"Ignoring PLTE chunk in grayscale PNG");
		png_crc_finish(png_ptr, length);
		return;
	}
#if !defined(PNG_READ_OPT_PLTE_SUPPORTED)
	if(png_ptr->color_type != PNG_COLOR_TYPE_PALETTE)
	{
		png_crc_finish(png_ptr, length);
		return;
	}
#endif
	if(length > 3*PNG_MAX_PALETTE_LENGTH || length % 3)
	{
		if(png_ptr->color_type != PNG_COLOR_TYPE_PALETTE)
		{
			png_warning(png_ptr, "Invalid palette chunk");
			png_crc_finish(png_ptr, length);
			return;
		}
		else
		{
			png_error(png_ptr, "Invalid palette chunk");
		}
	}
	num = (int)length / 3;
#ifndef PNG_NO_POINTER_INDEXING
	for(i = 0, pal_ptr = palette; i < num; i++, pal_ptr++)
	{
		png_byte buf[3];
		png_crc_read(png_ptr, buf, 3);
		pal_ptr->red = buf[0];
		pal_ptr->green = buf[1];
		pal_ptr->blue = buf[2];
	}
#else
	for(i = 0; i < num; i++)
	{
		png_byte buf[3];
		png_crc_read(png_ptr, buf, 3);
		palette[i].red = buf[0];
		palette[i].green = buf[1];
		palette[i].blue = buf[2];
	}
#endif
#if !defined(PNG_READ_OPT_PLTE_SUPPORTED)
	if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
#endif
	{
		png_crc_finish(png_ptr, 0);
	}
#if !defined(PNG_READ_OPT_PLTE_SUPPORTED)
	else if(png_crc_error(png_ptr))
	{
		if(!(png_ptr->flags & PNG_FLAG_CRC_ANCILLARY_USE))
		{
			if(png_ptr->flags & PNG_FLAG_CRC_ANCILLARY_NOWARN)
			{
				png_chunk_error(png_ptr, "CRC error");
			}
			else
			{
				png_chunk_warning(png_ptr, "CRC error");
				return;
			}
		}
		else if(!(png_ptr->flags & PNG_FLAG_CRC_ANCILLARY_NOWARN))
		{
			png_chunk_warning(png_ptr, "CRC error");
		}
	}
#endif
	png_set_PLTE(png_ptr, info_ptr, palette, num);
#if defined(PNG_READ_tRNS_SUPPORTED)
	if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
	{
		if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_tRNS))
		{
			if(png_ptr->num_trans > (png_uint_16)num)
			{
				png_warning(png_ptr, "Truncating incorrect tRNS chunk length");
				png_ptr->num_trans = (png_uint_16)num;
			}
			if(info_ptr->num_trans > (png_uint_16)num)
			{
				png_warning(png_ptr, "Truncating incorrect info tRNS chunk length");
				info_ptr->num_trans = (png_uint_16)num;
			}
		}
	}
#endif
}
void
png_handle_IEND(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_debug(1, "in png_handle_IEND\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR) || !(png_ptr->mode & PNG_HAVE_IDAT))
	{
		png_error(png_ptr, "No image in file");
	}
	png_ptr->mode |= (PNG_AFTER_IDAT | PNG_HAVE_IEND);
	if(length != 0)
	{
		png_warning(png_ptr, "Incorrect IEND chunk length");
	}
	png_crc_finish(png_ptr, length);
	if(&info_ptr == NULL)
		return;
}
#if defined(PNG_READ_gAMA_SUPPORTED)
void
png_handle_gAMA(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_fixed_point igamma;
#ifdef PNG_FLOATING_POINT_SUPPORTED
	float file_gamma;
#endif
	png_byte buf[4];
	png_debug(1, "in png_handle_gAMA\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before gAMA");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid gAMA after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(png_ptr->mode & PNG_HAVE_PLTE)
		png_warning(png_ptr, "Out of place gAMA chunk");
	if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_gAMA)
#if defined(PNG_READ_sRGB_SUPPORTED)
		&& !(info_ptr->valid & PNG_INFO_sRGB)
#endif
		)
	{
		png_warning(png_ptr, "Duplicate gAMA chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(length != 4)
	{
		png_warning(png_ptr, "Incorrect gAMA chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, 4);
	if(png_crc_finish(png_ptr, 0))
		return;
	igamma = (png_fixed_point)png_get_uint_32(buf);
	if(igamma == 0)
	{
		png_warning(png_ptr,
			"Ignoring gAMA chunk with gamma = 0");
		return;
	}
#if defined(PNG_READ_sRGB_SUPPORTED)
	if(info_ptr->valid & PNG_INFO_sRGB)
		if(PNG_OUT_OF_RANGE(igamma, 45500L, 500))
		{
			png_warning(png_ptr,
				"Ignoring incorrect gAMA value when sRGB is also present");
#ifndef PNG_NO_CONSOLE_IO
			fprintf(stderr, "gamma = (%d/100000)\n", (int)igamma);
#endif
			return;
		}
#endif
#ifdef PNG_FLOATING_POINT_SUPPORTED
		file_gamma = (float)igamma / (float)100000.0;
#  ifdef PNG_READ_GAMMA_SUPPORTED
		png_ptr->gamma = file_gamma;
#  endif
		png_set_gAMA(png_ptr, info_ptr, file_gamma);
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
		png_set_gAMA_fixed(png_ptr, info_ptr, igamma);
#endif
}
#endif
#if defined(PNG_READ_sBIT_SUPPORTED)
void
png_handle_sBIT(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_size_t truelen;
	png_byte buf[4];
	png_debug(1, "in png_handle_sBIT\n");
	buf[0] = buf[1] = buf[2] = buf[3] = 0;
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before sBIT");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid sBIT after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(png_ptr->mode & PNG_HAVE_PLTE)
	{
		png_warning(png_ptr, "Out of place sBIT chunk");
	}
	if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_sBIT))
	{
		png_warning(png_ptr, "Duplicate sBIT chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		truelen = 3;
	else
		truelen = (png_size_t)png_ptr->channels;
	if(length != truelen || length > 4)
	{
		png_warning(png_ptr, "Incorrect sBIT chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, truelen);
	if(png_crc_finish(png_ptr, 0))
		return;
	if(png_ptr->color_type & PNG_COLOR_MASK_COLOR)
	{
		png_ptr->sig_bit.red = buf[0];
		png_ptr->sig_bit.green = buf[1];
		png_ptr->sig_bit.blue = buf[2];
		png_ptr->sig_bit.alpha = buf[3];
	}
	else
	{
		png_ptr->sig_bit.gray = buf[0];
		png_ptr->sig_bit.red = buf[0];
		png_ptr->sig_bit.green = buf[0];
		png_ptr->sig_bit.blue = buf[0];
		png_ptr->sig_bit.alpha = buf[1];
	}
	png_set_sBIT(png_ptr, info_ptr, &(png_ptr->sig_bit));
}
#endif
#if defined(PNG_READ_cHRM_SUPPORTED)
void
png_handle_cHRM(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_byte buf[4];
#ifdef PNG_FLOATING_POINT_SUPPORTED
	float white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y;
#endif
	png_fixed_point int_x_white, int_y_white, int_x_red, int_y_red, int_x_green,
		int_y_green, int_x_blue, int_y_blue;
	png_uint_32 uint_x, uint_y;
	png_debug(1, "in png_handle_cHRM\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before cHRM");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid cHRM after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(png_ptr->mode & PNG_HAVE_PLTE)
		png_warning(png_ptr, "Missing PLTE before cHRM");
	if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_cHRM)
#if defined(PNG_READ_sRGB_SUPPORTED)
		&& !(info_ptr->valid & PNG_INFO_sRGB)
#endif
		)
	{
		png_warning(png_ptr, "Duplicate cHRM chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(length != 32)
	{
		png_warning(png_ptr, "Incorrect cHRM chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, 4);
	uint_x = png_get_uint_32(buf);
	png_crc_read(png_ptr, buf, 4);
	uint_y = png_get_uint_32(buf);
	if(uint_x > 80000L || uint_y > 80000L ||
		uint_x + uint_y > 100000L)
	{
		png_warning(png_ptr, "Invalid cHRM white point");
		png_crc_finish(png_ptr, 24);
		return;
	}
	int_x_white = (png_fixed_point)uint_x;
	int_y_white = (png_fixed_point)uint_y;
	png_crc_read(png_ptr, buf, 4);
	uint_x = png_get_uint_32(buf);
	png_crc_read(png_ptr, buf, 4);
	uint_y = png_get_uint_32(buf);
	if(uint_x > 80000L || uint_y > 80000L ||
		uint_x + uint_y > 100000L)
	{
		png_warning(png_ptr, "Invalid cHRM red point");
		png_crc_finish(png_ptr, 16);
		return;
	}
	int_x_red = (png_fixed_point)uint_x;
	int_y_red = (png_fixed_point)uint_y;
	png_crc_read(png_ptr, buf, 4);
	uint_x = png_get_uint_32(buf);
	png_crc_read(png_ptr, buf, 4);
	uint_y = png_get_uint_32(buf);
	if(uint_x > 80000L || uint_y > 80000L ||
		uint_x + uint_y > 100000L)
	{
		png_warning(png_ptr, "Invalid cHRM green point");
		png_crc_finish(png_ptr, 8);
		return;
	}
	int_x_green = (png_fixed_point)uint_x;
	int_y_green = (png_fixed_point)uint_y;
	png_crc_read(png_ptr, buf, 4);
	uint_x = png_get_uint_32(buf);
	png_crc_read(png_ptr, buf, 4);
	uint_y = png_get_uint_32(buf);
	if(uint_x > 80000L || uint_y > 80000L ||
		uint_x + uint_y > 100000L)
	{
		png_warning(png_ptr, "Invalid cHRM blue point");
		png_crc_finish(png_ptr, 0);
		return;
	}
	int_x_blue = (png_fixed_point)uint_x;
	int_y_blue = (png_fixed_point)uint_y;
#ifdef PNG_FLOATING_POINT_SUPPORTED
	white_x = (float)int_x_white / (float)100000.0;
	white_y = (float)int_y_white / (float)100000.0;
	red_x   = (float)int_x_red   / (float)100000.0;
	red_y   = (float)int_y_red   / (float)100000.0;
	green_x = (float)int_x_green / (float)100000.0;
	green_y = (float)int_y_green / (float)100000.0;
	blue_x  = (float)int_x_blue  / (float)100000.0;
	blue_y  = (float)int_y_blue  / (float)100000.0;
#endif
#if defined(PNG_READ_sRGB_SUPPORTED)
	if(info_ptr->valid & PNG_INFO_sRGB)
	{
		if(PNG_OUT_OF_RANGE(int_x_white, 31270,  1000) ||
			PNG_OUT_OF_RANGE(int_y_white, 32900,  1000) ||
			PNG_OUT_OF_RANGE(int_x_red,   64000L, 1000) ||
			PNG_OUT_OF_RANGE(int_y_red,   33000,  1000) ||
			PNG_OUT_OF_RANGE(int_x_green, 30000,  1000) ||
			PNG_OUT_OF_RANGE(int_y_green, 60000L, 1000) ||
			PNG_OUT_OF_RANGE(int_x_blue,  15000,  1000) ||
			PNG_OUT_OF_RANGE(int_y_blue,   6000,  1000))
		{
			png_warning(png_ptr,
				"Ignoring incorrect cHRM value when sRGB is also present");
#ifndef PNG_NO_CONSOLE_IO
#ifdef PNG_FLOATING_POINT_SUPPORTED
			fprintf(stderr,"wx = %f, wy = %f, rx = %f, ry = %f\n",
				white_x, white_y, red_x, red_y);
			fprintf(stderr,"gx = %f, gy = %f, bx = %f, by = %f\n",
				green_x, green_y, blue_x, blue_y);
#else
			fprintf(stderr,"wx = %ld, wy = %ld, rx = %ld, ry = %ld\n",
				int_x_white, int_y_white, int_x_red, int_y_red);
			fprintf(stderr,"gx = %ld, gy = %ld, bx = %ld, by = %ld\n",
				int_x_green, int_y_green, int_x_blue, int_y_blue);
#endif
#endif
		}
		png_crc_finish(png_ptr, 0);
		return;
	}
#endif
#ifdef PNG_FLOATING_POINT_SUPPORTED
	png_set_cHRM(png_ptr, info_ptr,
		white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y);
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
	png_set_cHRM_fixed(png_ptr, info_ptr,
		int_x_white, int_y_white, int_x_red, int_y_red, int_x_green,
		int_y_green, int_x_blue, int_y_blue);
#endif
	if(png_crc_finish(png_ptr, 0))
		return;
}
#endif
#if defined(PNG_READ_sRGB_SUPPORTED)
void
png_handle_sRGB(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	int intent;
	png_byte buf[1];
	png_debug(1, "in png_handle_sRGB\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before sRGB");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid sRGB after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(png_ptr->mode & PNG_HAVE_PLTE)
		png_warning(png_ptr, "Out of place sRGB chunk");
	if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_sRGB))
	{
		png_warning(png_ptr, "Duplicate sRGB chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(length != 1)
	{
		png_warning(png_ptr, "Incorrect sRGB chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, 1);
	if(png_crc_finish(png_ptr, 0))
		return;
	intent = buf[0];
	if(intent >= PNG_sRGB_INTENT_LAST)
	{
		png_warning(png_ptr, "Unknown sRGB intent");
		return;
	}
#if defined(PNG_READ_gAMA_SUPPORTED) && defined(PNG_READ_GAMMA_SUPPORTED)
	if((info_ptr->valid & PNG_INFO_gAMA))
	{
		png_fixed_point igamma;
#ifdef PNG_FIXED_POINT_SUPPORTED
		igamma = info_ptr->int_gamma;
#else
#  ifdef PNG_FLOATING_POINT_SUPPORTED
		igamma = (png_fixed_point)(info_ptr->gamma * 100000.);
#  endif
#endif
		if(PNG_OUT_OF_RANGE(igamma, 45500L, 500))
		{
			png_warning(png_ptr,
				"Ignoring incorrect gAMA value when sRGB is also present");
#ifndef PNG_NO_CONSOLE_IO
#  ifdef PNG_FIXED_POINT_SUPPORTED
			fprintf(stderr,"incorrect gamma = (%d/100000)\n",(int)png_ptr->int_gamma);
#  else
#    ifdef PNG_FLOATING_POINT_SUPPORTED
			fprintf(stderr,"incorrect gamma = %f\n",png_ptr->gamma);
#    endif
#  endif
#endif
		}
	}
#endif
#ifdef PNG_READ_cHRM_SUPPORTED
#ifdef PNG_FIXED_POINT_SUPPORTED
	if(info_ptr->valid & PNG_INFO_cHRM)
		if(PNG_OUT_OF_RANGE(info_ptr->int_x_white, 31270,  1000) ||
			PNG_OUT_OF_RANGE(info_ptr->int_y_white, 32900,  1000) ||
			PNG_OUT_OF_RANGE(info_ptr->int_x_red,   64000L, 1000) ||
			PNG_OUT_OF_RANGE(info_ptr->int_y_red,   33000,  1000) ||
			PNG_OUT_OF_RANGE(info_ptr->int_x_green, 30000,  1000) ||
			PNG_OUT_OF_RANGE(info_ptr->int_y_green, 60000L, 1000) ||
			PNG_OUT_OF_RANGE(info_ptr->int_x_blue,  15000,  1000) ||
			PNG_OUT_OF_RANGE(info_ptr->int_y_blue,   6000,  1000))
		{
			png_warning(png_ptr,
				"Ignoring incorrect cHRM value when sRGB is also present");
		}
#endif
#endif
		png_set_sRGB_gAMA_and_cHRM(png_ptr, info_ptr, intent);
}
#endif
#if defined(PNG_READ_iCCP_SUPPORTED)
void
png_handle_iCCP(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_charp chunkdata;
	png_byte compression_type;
	png_bytep pC;
	png_charp profile;
	png_uint_32 skip = 0;
	png_uint_32 profile_size, profile_length;
	png_size_t slength, prefix_length, data_length;
	png_debug(1, "in png_handle_iCCP\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before iCCP");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid iCCP after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(png_ptr->mode & PNG_HAVE_PLTE)
		png_warning(png_ptr, "Out of place iCCP chunk");
	if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_iCCP))
	{
		png_warning(png_ptr, "Duplicate iCCP chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
#ifdef PNG_MAX_MALLOC_64K
	if(length > (png_uint_32)65535L)
	{
		png_warning(png_ptr, "iCCP chunk too large to fit in memory");
		skip = length - (png_uint_32)65535L;
		length = (png_uint_32)65535L;
	}
#endif
	chunkdata = (png_charp)png_malloc(png_ptr, length + 1);
	slength = (png_size_t)length;
	png_crc_read(png_ptr, (png_bytep)chunkdata, slength);
	if(png_crc_finish(png_ptr, skip))
	{
		png_free(png_ptr, chunkdata);
		return;
	}
	chunkdata[slength] = 0x00;
	for(profile = chunkdata; *profile; profile++)
		;
	++profile;
	if( profile >= chunkdata + slength)
	{
		png_free(png_ptr, chunkdata);
		png_warning(png_ptr, "Malformed iCCP chunk");
		return;
	}
	compression_type = *profile++;
	if(compression_type)
	{
		png_warning(png_ptr, "Ignoring nonzero compression type in iCCP chunk");
		compression_type = 0x00;
	}
	prefix_length = profile - chunkdata;
	chunkdata = png_decompress_chunk(png_ptr, compression_type, chunkdata,
		slength, prefix_length, &data_length);
	profile_length = (png_uint_32)(data_length - prefix_length);
	if( prefix_length > data_length || profile_length < 4)
	{
		png_free(png_ptr, chunkdata);
		png_warning(png_ptr, "Profile size field missing from iCCP chunk");
		return;
	}
	pC = (png_bytep)(chunkdata+prefix_length);
	profile_size = ((*(pC  ))<<24) |
		((*(pC+1))<<16) |
		((*(pC+2))<< 8) |
		((*(pC+3))    );
	if(profile_size < profile_length)
		profile_length = profile_size;
	if(profile_size > profile_length)
	{
		png_free(png_ptr, chunkdata);
		png_warning(png_ptr, "Ignoring truncated iCCP profile.\n");
		return;
	}
	png_set_iCCP(png_ptr, info_ptr, chunkdata, compression_type,
		chunkdata + prefix_length, profile_length);
	png_free(png_ptr, chunkdata);
}
#endif
#if defined(PNG_READ_sPLT_SUPPORTED)
void
png_handle_sPLT(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_bytep chunkdata;
	png_bytep entry_start;
	png_sPLT_t new_palette;
#ifdef PNG_NO_POINTER_INDEXING
	png_sPLT_entryp pp;
#endif
	int data_length, entry_size, i;
	png_uint_32 skip = 0;
	png_size_t slength;
	png_debug(1, "in png_handle_sPLT\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before sPLT");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid sPLT after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
#ifdef PNG_MAX_MALLOC_64K
	if(length > (png_uint_32)65535L)
	{
		png_warning(png_ptr, "sPLT chunk too large to fit in memory");
		skip = length - (png_uint_32)65535L;
		length = (png_uint_32)65535L;
	}
#endif
	chunkdata = (png_bytep)png_malloc(png_ptr, length + 1);
	slength = (png_size_t)length;
	png_crc_read(png_ptr, (png_bytep)chunkdata, slength);
	if(png_crc_finish(png_ptr, skip))
	{
		png_free(png_ptr, chunkdata);
		return;
	}
	chunkdata[slength] = 0x00;
	for(entry_start = chunkdata; *entry_start; entry_start++)
		;
	++entry_start;
	if(entry_start > chunkdata + slength)
	{
		png_free(png_ptr, chunkdata);
		png_warning(png_ptr, "malformed sPLT chunk");
		return;
	}
	new_palette.depth = *entry_start++;
	entry_size = (new_palette.depth == 8 ? 6 : 10);
	data_length = (int)((slength - (entry_start - chunkdata)));
	if(data_length % entry_size)
	{
		png_free(png_ptr, chunkdata);
		png_warning(png_ptr, "sPLT chunk has bad length");
		return;
	}
	new_palette.nentries = (png_uint_32) (data_length / entry_size);
	if((png_uint_32) new_palette.nentries > (png_uint_32) (PNG_SIZE_MAX /
		png_sizeof(png_sPLT_entry)))
	{
		png_warning(png_ptr, "sPLT chunk too long");
		return;
	}
	new_palette.entries = (png_sPLT_entryp)png_malloc_warn(
		png_ptr, new_palette.nentries * png_sizeof(png_sPLT_entry));
	if(new_palette.entries == NULL)
	{
		png_warning(png_ptr, "sPLT chunk requires too much memory");
		return;
	}
#ifndef PNG_NO_POINTER_INDEXING
	for(i = 0; i < new_palette.nentries; i++)
	{
		png_sPLT_entryp pp = new_palette.entries + i;
		if(new_palette.depth == 8)
		{
			pp->red = *entry_start++;
			pp->green = *entry_start++;
			pp->blue = *entry_start++;
			pp->alpha = *entry_start++;
		}
		else
		{
			pp->red   = png_get_uint_16(entry_start); entry_start += 2;
			pp->green = png_get_uint_16(entry_start); entry_start += 2;
			pp->blue  = png_get_uint_16(entry_start); entry_start += 2;
			pp->alpha = png_get_uint_16(entry_start); entry_start += 2;
		}
		pp->frequency = png_get_uint_16(entry_start); entry_start += 2;
	}
#else
	pp = new_palette.entries;
	for(i = 0; i < new_palette.nentries; i++)
	{
		if(new_palette.depth == 8)
		{
			pp[i].red   = *entry_start++;
			pp[i].green = *entry_start++;
			pp[i].blue  = *entry_start++;
			pp[i].alpha = *entry_start++;
		}
		else
		{
			pp[i].red   = png_get_uint_16(entry_start); entry_start += 2;
			pp[i].green = png_get_uint_16(entry_start); entry_start += 2;
			pp[i].blue  = png_get_uint_16(entry_start); entry_start += 2;
			pp[i].alpha = png_get_uint_16(entry_start); entry_start += 2;
		}
		pp->frequency = png_get_uint_16(entry_start); entry_start += 2;
	}
#endif
	new_palette.name = (png_charp)chunkdata;
	png_set_sPLT(png_ptr, info_ptr, &new_palette, 1);
	png_free(png_ptr, chunkdata);
	png_free(png_ptr, new_palette.entries);
}
#endif
#if defined(PNG_READ_tRNS_SUPPORTED)
void
png_handle_tRNS(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_byte readbuf[PNG_MAX_PALETTE_LENGTH];
	png_debug(1, "in png_handle_tRNS\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before tRNS");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid tRNS after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_tRNS))
	{
		png_warning(png_ptr, "Duplicate tRNS chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(png_ptr->color_type == PNG_COLOR_TYPE_GRAY)
	{
		png_byte buf[2];
		if(length != 2)
		{
			png_warning(png_ptr, "Incorrect tRNS chunk length");
			png_crc_finish(png_ptr, length);
			return;
		}
		png_crc_read(png_ptr, buf, 2);
		png_ptr->num_trans = 1;
		png_ptr->trans_values.gray = png_get_uint_16(buf);
	}
	else if(png_ptr->color_type == PNG_COLOR_TYPE_RGB)
	{
		png_byte buf[6];
		if(length != 6)
		{
			png_warning(png_ptr, "Incorrect tRNS chunk length");
			png_crc_finish(png_ptr, length);
			return;
		}
		png_crc_read(png_ptr, buf, (png_size_t)length);
		png_ptr->num_trans = 1;
		png_ptr->trans_values.red = png_get_uint_16(buf);
		png_ptr->trans_values.green = png_get_uint_16(buf + 2);
		png_ptr->trans_values.blue = png_get_uint_16(buf + 4);
	}
	else if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
	{
		if(!(png_ptr->mode & PNG_HAVE_PLTE))
		{
			png_warning(png_ptr, "Missing PLTE before tRNS");
		}
		if(length > (png_uint_32)png_ptr->num_palette ||
			length > PNG_MAX_PALETTE_LENGTH)
		{
			png_warning(png_ptr, "Incorrect tRNS chunk length");
			png_crc_finish(png_ptr, length);
			return;
		}
		if(length == 0)
		{
			png_warning(png_ptr, "Zero length tRNS chunk");
			png_crc_finish(png_ptr, length);
			return;
		}
		png_crc_read(png_ptr, readbuf, (png_size_t)length);
		png_ptr->num_trans = (png_uint_16)length;
	}
	else
	{
		png_warning(png_ptr, "tRNS chunk not allowed with alpha channel");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(png_crc_finish(png_ptr, 0))
		return;
	png_set_tRNS(png_ptr, info_ptr, readbuf, png_ptr->num_trans,
		&(png_ptr->trans_values));
}
#endif
#if defined(PNG_READ_bKGD_SUPPORTED)
void
png_handle_bKGD(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_size_t truelen;
	png_byte buf[6];
	png_debug(1, "in png_handle_bKGD\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before bKGD");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid bKGD after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE &&
		!(png_ptr->mode & PNG_HAVE_PLTE))
	{
		png_warning(png_ptr, "Missing PLTE before bKGD");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_bKGD))
	{
		png_warning(png_ptr, "Duplicate bKGD chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		truelen = 1;
	else if(png_ptr->color_type & PNG_COLOR_MASK_COLOR)
		truelen = 6;
	else
		truelen = 2;
	if(length != truelen)
	{
		png_warning(png_ptr, "Incorrect bKGD chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, truelen);
	if(png_crc_finish(png_ptr, 0))
		return;
	if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
	{
		png_ptr->background.index = buf[0];
		if(info_ptr->num_palette)
		{
			if(buf[0] > info_ptr->num_palette)
			{
				png_warning(png_ptr, "Incorrect bKGD chunk index value");
				return;
			}
			png_ptr->background.red =
				(png_uint_16)png_ptr->palette[buf[0]].red;
			png_ptr->background.green =
				(png_uint_16)png_ptr->palette[buf[0]].green;
			png_ptr->background.blue =
				(png_uint_16)png_ptr->palette[buf[0]].blue;
		}
	}
	else if(!(png_ptr->color_type & PNG_COLOR_MASK_COLOR))
	{
		png_ptr->background.red =
			png_ptr->background.green =
			png_ptr->background.blue =
			png_ptr->background.gray = png_get_uint_16(buf);
	}
	else
	{
		png_ptr->background.red = png_get_uint_16(buf);
		png_ptr->background.green = png_get_uint_16(buf + 2);
		png_ptr->background.blue = png_get_uint_16(buf + 4);
	}
	png_set_bKGD(png_ptr, info_ptr, &(png_ptr->background));
}
#endif
#if defined(PNG_READ_hIST_SUPPORTED)
void
png_handle_hIST(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	unsigned int num, i;
	png_uint_16 readbuf[PNG_MAX_PALETTE_LENGTH];
	png_debug(1, "in png_handle_hIST\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before hIST");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid hIST after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(!(png_ptr->mode & PNG_HAVE_PLTE))
	{
		png_warning(png_ptr, "Missing PLTE before hIST");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_hIST))
	{
		png_warning(png_ptr, "Duplicate hIST chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	num = length / 2 ;
	if(num != (unsigned int) png_ptr->num_palette || num >
		(unsigned int) PNG_MAX_PALETTE_LENGTH)
	{
		png_warning(png_ptr, "Incorrect hIST chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	for(i = 0; i < num; i++)
	{
		png_byte buf[2];
		png_crc_read(png_ptr, buf, 2);
		readbuf[i] = png_get_uint_16(buf);
	}
	if(png_crc_finish(png_ptr, 0))
		return;
	png_set_hIST(png_ptr, info_ptr, readbuf);
}
#endif
#if defined(PNG_READ_pHYs_SUPPORTED)
void
png_handle_pHYs(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_byte buf[9];
	png_uint_32 res_x, res_y;
	int unit_type;
	png_debug(1, "in png_handle_pHYs\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before pHYs");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid pHYs after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_pHYs))
	{
		png_warning(png_ptr, "Duplicate pHYs chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(length != 9)
	{
		png_warning(png_ptr, "Incorrect pHYs chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, 9);
	if(png_crc_finish(png_ptr, 0))
		return;
	res_x = png_get_uint_32(buf);
	res_y = png_get_uint_32(buf + 4);
	unit_type = buf[8];
	png_set_pHYs(png_ptr, info_ptr, res_x, res_y, unit_type);
}
#endif
#if defined(PNG_READ_oFFs_SUPPORTED)
void
png_handle_oFFs(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_byte buf[9];
	png_int_32 offset_x, offset_y;
	int unit_type;
	png_debug(1, "in png_handle_oFFs\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before oFFs");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid oFFs after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_oFFs))
	{
		png_warning(png_ptr, "Duplicate oFFs chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(length != 9)
	{
		png_warning(png_ptr, "Incorrect oFFs chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, 9);
	if(png_crc_finish(png_ptr, 0))
		return;
	offset_x = png_get_int_32(buf);
	offset_y = png_get_int_32(buf + 4);
	unit_type = buf[8];
	png_set_oFFs(png_ptr, info_ptr, offset_x, offset_y, unit_type);
}
#endif
#if defined(PNG_READ_pCAL_SUPPORTED)
void
png_handle_pCAL(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_charp purpose;
	png_int_32 X0, X1;
	png_byte type, nparams;
	png_charp buf, units, endptr;
	png_charpp params;
	png_size_t slength;
	int i;
	png_debug(1, "in png_handle_pCAL\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before pCAL");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid pCAL after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_pCAL))
	{
		png_warning(png_ptr, "Duplicate pCAL chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_debug1(2, "Allocating and reading pCAL chunk data (%lu bytes)\n",
		length + 1);
	purpose = (png_charp)png_malloc_warn(png_ptr, length + 1);
	if(purpose == NULL)
	{
		png_warning(png_ptr, "No memory for pCAL purpose.");
		return;
	}
	slength = (png_size_t)length;
	png_crc_read(png_ptr, (png_bytep)purpose, slength);
	if(png_crc_finish(png_ptr, 0))
	{
		png_free(png_ptr, purpose);
		return;
	}
	purpose[slength] = 0x00;
	png_debug(3, "Finding end of pCAL purpose string\n");
	for(buf = purpose; *buf; buf++)
		;
	endptr = purpose + slength;
	if(endptr <= buf + 12)
	{
		png_warning(png_ptr, "Invalid pCAL data");
		png_free(png_ptr, purpose);
		return;
	}
	png_debug(3, "Reading pCAL X0, X1, type, nparams, and units\n");
	X0 = png_get_int_32((png_bytep)buf+1);
	X1 = png_get_int_32((png_bytep)buf+5);
	type = buf[9];
	nparams = buf[10];
	units = buf + 11;
	png_debug(3, "Checking pCAL equation type and number of parameters\n");
	if((type == PNG_EQUATION_LINEAR && nparams != 2) ||
		(type == PNG_EQUATION_BASE_E && nparams != 3) ||
		(type == PNG_EQUATION_ARBITRARY && nparams != 3) ||
		(type == PNG_EQUATION_HYPERBOLIC && nparams != 4))
	{
		png_warning(png_ptr, "Invalid pCAL parameters for equation type");
		png_free(png_ptr, purpose);
		return;
	}
	else if(type >= PNG_EQUATION_LAST)
	{
		png_warning(png_ptr, "Unrecognized equation type for pCAL chunk");
	}
	for(buf = units; *buf; buf++)
		;
	png_debug(3, "Allocating pCAL parameters array\n");
	params = (png_charpp)png_malloc_warn(png_ptr, (png_uint_32)(nparams
		*png_sizeof(png_charp))) ;
	if(params == NULL)
	{
		png_free(png_ptr, purpose);
		png_warning(png_ptr, "No memory for pCAL params.");
		return;
	}
	for(i = 0; i < (int)nparams; i++)
	{
		buf++;
		png_debug1(3, "Reading pCAL parameter %d\n", i);
		for(params[i] = buf; *buf != 0x00 && buf <= endptr; buf++)
			;
		if(buf > endptr)
		{
			png_warning(png_ptr, "Invalid pCAL data");
			png_free(png_ptr, purpose);
			png_free(png_ptr, params);
			return;
		}
	}
	png_set_pCAL(png_ptr, info_ptr, purpose, X0, X1, type, nparams,
		units, params);
	png_free(png_ptr, purpose);
	png_free(png_ptr, params);
}
#endif
#if defined(PNG_READ_sCAL_SUPPORTED)
void
png_handle_sCAL(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_charp buffer, ep;
#ifdef PNG_FLOATING_POINT_SUPPORTED
	double width, height;
	png_charp vp;
#else
#ifdef PNG_FIXED_POINT_SUPPORTED
	png_charp swidth, sheight;
#endif
#endif
	png_size_t slength;
	png_debug(1, "in png_handle_sCAL\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before sCAL");
	else if(png_ptr->mode & PNG_HAVE_IDAT)
	{
		png_warning(png_ptr, "Invalid sCAL after IDAT");
		png_crc_finish(png_ptr, length);
		return;
	}
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_sCAL))
	{
		png_warning(png_ptr, "Duplicate sCAL chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_debug1(2, "Allocating and reading sCAL chunk data (%lu bytes)\n",
		length + 1);
	buffer = (png_charp)png_malloc_warn(png_ptr, length + 1);
	if(buffer == NULL)
	{
		png_warning(png_ptr, "Out of memory while processing sCAL chunk");
		return;
	}
	slength = (png_size_t)length;
	png_crc_read(png_ptr, (png_bytep)buffer, slength);
	if(png_crc_finish(png_ptr, 0))
	{
		png_free(png_ptr, buffer);
		return;
	}
	buffer[slength] = 0x00;
	ep = buffer + 1;
#ifdef PNG_FLOATING_POINT_SUPPORTED
	width = strtod(ep, &vp);
	if(*vp)
	{
		png_warning(png_ptr, "malformed width string in sCAL chunk");
		return;
	}
#else
#ifdef PNG_FIXED_POINT_SUPPORTED
	swidth = (png_charp)png_malloc_warn(png_ptr, png_strlen(ep) + 1);
	if(swidth == NULL)
	{
		png_warning(png_ptr, "Out of memory while processing sCAL chunk width");
		return;
	}
	png_memcpy(swidth, ep, (png_size_t)png_strlen(ep));
#endif
#endif
	for(ep = buffer; *ep; ep++)
		;
	ep++;
#ifdef PNG_FLOATING_POINT_SUPPORTED
	height = strtod(ep, &vp);
	if(*vp)
	{
		png_warning(png_ptr, "malformed height string in sCAL chunk");
		return;
	}
#else
#ifdef PNG_FIXED_POINT_SUPPORTED
	sheight = (png_charp)png_malloc_warn(png_ptr, png_strlen(ep) + 1);
	if(swidth == NULL)
	{
		png_warning(png_ptr, "Out of memory while processing sCAL chunk height");
		return;
	}
	png_memcpy(sheight, ep, (png_size_t)png_strlen(ep));
#endif
#endif
	if(buffer + slength < ep
#ifdef PNG_FLOATING_POINT_SUPPORTED
		|| width <= 0. || height <= 0.
#endif
		)
	{
		png_warning(png_ptr, "Invalid sCAL data");
		png_free(png_ptr, buffer);
#if defined(PNG_FIXED_POINT_SUPPORTED) && !defined(PNG_FLOATING_POINT_SUPPORTED)
		png_free(png_ptr, swidth);
		png_free(png_ptr, sheight);
#endif
		return;
	}
#ifdef PNG_FLOATING_POINT_SUPPORTED
	png_set_sCAL(png_ptr, info_ptr, buffer[0], width, height);
#else
#ifdef PNG_FIXED_POINT_SUPPORTED
	png_set_sCAL_s(png_ptr, info_ptr, buffer[0], swidth, sheight);
#endif
#endif
	png_free(png_ptr, buffer);
#if defined(PNG_FIXED_POINT_SUPPORTED) && !defined(PNG_FLOATING_POINT_SUPPORTED)
	png_free(png_ptr, swidth);
	png_free(png_ptr, sheight);
#endif
}
#endif
#if defined(PNG_READ_tIME_SUPPORTED)
void
png_handle_tIME(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_byte buf[7];
	png_time mod_time;
	png_debug(1, "in png_handle_tIME\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Out of place tIME chunk");
	else if(info_ptr != NULL && (info_ptr->valid & PNG_INFO_tIME))
	{
		png_warning(png_ptr, "Duplicate tIME chunk");
		png_crc_finish(png_ptr, length);
		return;
	}
	if(png_ptr->mode & PNG_HAVE_IDAT)
		png_ptr->mode |= PNG_AFTER_IDAT;
	if(length != 7)
	{
		png_warning(png_ptr, "Incorrect tIME chunk length");
		png_crc_finish(png_ptr, length);
		return;
	}
	png_crc_read(png_ptr, buf, 7);
	if(png_crc_finish(png_ptr, 0))
		return;
	mod_time.second = buf[6];
	mod_time.minute = buf[5];
	mod_time.hour = buf[4];
	mod_time.day = buf[3];
	mod_time.month = buf[2];
	mod_time.year = png_get_uint_16(buf);
	png_set_tIME(png_ptr, info_ptr, &mod_time);
}
#endif
#if defined(PNG_READ_tEXt_SUPPORTED)
void
png_handle_tEXt(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_textp text_ptr;
	png_charp key;
	png_charp text;
	png_uint_32 skip = 0;
	png_size_t slength;
	int ret;
	png_debug(1, "in png_handle_tEXt\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before tEXt");
	if(png_ptr->mode & PNG_HAVE_IDAT)
		png_ptr->mode |= PNG_AFTER_IDAT;
#ifdef PNG_MAX_MALLOC_64K
	if(length > (png_uint_32)65535L)
	{
		png_warning(png_ptr, "tEXt chunk too large to fit in memory");
		skip = length - (png_uint_32)65535L;
		length = (png_uint_32)65535L;
	}
#endif
	key = (png_charp)png_malloc_warn(png_ptr, length + 1);
	if(key == NULL)
	{
		png_warning(png_ptr, "No memory to process text chunk.");
		return;
	}
	slength = (png_size_t)length;
	png_crc_read(png_ptr, (png_bytep)key, slength);
	if(png_crc_finish(png_ptr, skip))
	{
		png_free(png_ptr, key);
		return;
	}
	key[slength] = 0x00;
	for(text = key; *text; text++)
		;
	if(text != key + slength)
		text++;
	text_ptr = (png_textp)png_malloc_warn(png_ptr,
		(png_uint_32)png_sizeof(png_text));
	if(text_ptr == NULL)
	{
		png_warning(png_ptr, "Not enough memory to process text chunk.");
		png_free(png_ptr, key);
		return;
	}
	text_ptr->compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr->key = key;
#ifdef PNG_iTXt_SUPPORTED
	text_ptr->lang = NULL;
	text_ptr->lang_key = NULL;
	text_ptr->itxt_length = 0;
#endif
	text_ptr->text = text;
	text_ptr->text_length = png_strlen(text);
	ret = png_set_text_2(png_ptr, info_ptr, text_ptr, 1);
	png_free(png_ptr, key);
	png_free(png_ptr, text_ptr);
	if(ret)
		png_warning(png_ptr, "Insufficient memory to process text chunk.");
}
#endif
#if defined(PNG_READ_zTXt_SUPPORTED)
void
png_handle_zTXt(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_textp text_ptr;
	png_charp chunkdata;
	png_charp text;
	int comp_type;
	int ret;
	png_size_t slength, prefix_len, data_len;
	png_debug(1, "in png_handle_zTXt\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before zTXt");
	if(png_ptr->mode & PNG_HAVE_IDAT)
		png_ptr->mode |= PNG_AFTER_IDAT;
#ifdef PNG_MAX_MALLOC_64K
	if(length > (png_uint_32)65535L)
	{
		png_warning(png_ptr,"zTXt chunk too large to fit in memory");
		png_crc_finish(png_ptr, length);
		return;
	}
#endif
	chunkdata = (png_charp)png_malloc_warn(png_ptr, length + 1);
	if(chunkdata == NULL)
	{
		png_warning(png_ptr,"Out of memory processing zTXt chunk.");
		return;
	}
	slength = (png_size_t)length;
	png_crc_read(png_ptr, (png_bytep)chunkdata, slength);
	if(png_crc_finish(png_ptr, 0))
	{
		png_free(png_ptr, chunkdata);
		return;
	}
	chunkdata[slength] = 0x00;
	for(text = chunkdata; *text; text++)
		;
	if(text == chunkdata + slength)
	{
		comp_type = PNG_TEXT_COMPRESSION_NONE;
		png_warning(png_ptr, "Zero length zTXt chunk");
	}
	else
	{
		comp_type = *(++text);
		if(comp_type != PNG_TEXT_COMPRESSION_zTXt)
		{
			png_warning(png_ptr, "Unknown compression type in zTXt chunk");
			comp_type = PNG_TEXT_COMPRESSION_zTXt;
		}
		text++;
	}
	prefix_len = text - chunkdata;
	chunkdata = (png_charp)png_decompress_chunk(png_ptr, comp_type, chunkdata,
		(png_size_t)length, prefix_len, &data_len);
	text_ptr = (png_textp)png_malloc_warn(png_ptr,
		(png_uint_32)png_sizeof(png_text));
	if(text_ptr == NULL)
	{
		png_warning(png_ptr,"Not enough memory to process zTXt chunk.");
		png_free(png_ptr, chunkdata);
		return;
	}
	text_ptr->compression = comp_type;
	text_ptr->key = chunkdata;
#ifdef PNG_iTXt_SUPPORTED
	text_ptr->lang = NULL;
	text_ptr->lang_key = NULL;
	text_ptr->itxt_length = 0;
#endif
	text_ptr->text = chunkdata + prefix_len;
	text_ptr->text_length = data_len;
	ret = png_set_text_2(png_ptr, info_ptr, text_ptr, 1);
	png_free(png_ptr, text_ptr);
	png_free(png_ptr, chunkdata);
	if(ret)
		png_error(png_ptr, "Insufficient memory to store zTXt chunk.");
}
#endif
#if defined(PNG_READ_iTXt_SUPPORTED)
void
png_handle_iTXt(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_textp text_ptr;
	png_charp chunkdata;
	png_charp key, lang, text, lang_key;
	int comp_flag;
	int comp_type = 0;
	int ret;
	png_size_t slength, prefix_len, data_len;
	png_debug(1, "in png_handle_iTXt\n");
	if(!(png_ptr->mode & PNG_HAVE_IHDR))
		png_error(png_ptr, "Missing IHDR before iTXt");
	if(png_ptr->mode & PNG_HAVE_IDAT)
		png_ptr->mode |= PNG_AFTER_IDAT;
#ifdef PNG_MAX_MALLOC_64K
	if(length > (png_uint_32)65535L)
	{
		png_warning(png_ptr,"iTXt chunk too large to fit in memory");
		png_crc_finish(png_ptr, length);
		return;
	}
#endif
	chunkdata = (png_charp)png_malloc_warn(png_ptr, length + 1);
	if(chunkdata == NULL)
	{
		png_warning(png_ptr, "No memory to process iTXt chunk.");
		return;
	}
	slength = (png_size_t)length;
	png_crc_read(png_ptr, (png_bytep)chunkdata, slength);
	if(png_crc_finish(png_ptr, 0))
	{
		png_free(png_ptr, chunkdata);
		return;
	}
	chunkdata[slength] = 0x00;
	for(lang = chunkdata; *lang; lang++)
		;
	lang++;
	if(lang >= chunkdata + slength)
	{
		comp_flag = PNG_TEXT_COMPRESSION_NONE;
		png_warning(png_ptr, "Zero length iTXt chunk");
	}
	else
	{
		comp_flag = *lang++;
		comp_type = *lang++;
	}
	for(lang_key = lang; *lang_key; lang_key++)
		;
	lang_key++;
	for(text = lang_key; *text; text++)
		;
	text++;
	prefix_len = text - chunkdata;
	key = chunkdata;
	if(comp_flag)
		chunkdata = png_decompress_chunk(png_ptr, comp_type, chunkdata,
		(size_t)length, prefix_len, &data_len);
	else
		data_len = png_strlen(chunkdata + prefix_len);
	text_ptr = (png_textp)png_malloc_warn(png_ptr,
		(png_uint_32)png_sizeof(png_text));
	if(text_ptr == NULL)
	{
		png_warning(png_ptr,"Not enough memory to process iTXt chunk.");
		png_free(png_ptr, chunkdata);
		return;
	}
	text_ptr->compression = (int)comp_flag + 1;
	text_ptr->lang_key = chunkdata+(lang_key-key);
	text_ptr->lang = chunkdata+(lang-key);
	text_ptr->itxt_length = data_len;
	text_ptr->text_length = 0;
	text_ptr->key = chunkdata;
	text_ptr->text = chunkdata + prefix_len;
	ret = png_set_text_2(png_ptr, info_ptr, text_ptr, 1);
	png_free(png_ptr, text_ptr);
	png_free(png_ptr, chunkdata);
	if(ret)
		png_error(png_ptr, "Insufficient memory to store iTXt chunk.");
}
#endif
void
png_handle_unknown(png_structp png_ptr, png_infop info_ptr, png_uint_32 length)
{
	png_uint_32 skip = 0;
	png_debug(1, "in png_handle_unknown\n");
	if(png_ptr->mode & PNG_HAVE_IDAT)
	{
#ifdef PNG_USE_LOCAL_ARRAYS
		PNG_IDAT;
#endif
		if(png_memcmp(png_ptr->chunk_name, png_IDAT, 4))
			png_ptr->mode |= PNG_AFTER_IDAT;
	}
	png_check_chunk_name(png_ptr, png_ptr->chunk_name);
	if(!(png_ptr->chunk_name[0] & 0x20))
	{
#if defined(PNG_READ_UNKNOWN_CHUNKS_SUPPORTED)
		if(png_handle_as_unknown(png_ptr, png_ptr->chunk_name) !=
			PNG_HANDLE_CHUNK_ALWAYS
#if defined(PNG_READ_USER_CHUNKS_SUPPORTED)
			&& png_ptr->read_user_chunk_fn == NULL
#endif
			)
#endif
			png_chunk_error(png_ptr, "unknown critical chunk");
	}
#if defined(PNG_READ_UNKNOWN_CHUNKS_SUPPORTED)
	if(png_ptr->flags & PNG_FLAG_KEEP_UNKNOWN_CHUNKS)
	{
		png_unknown_chunk chunk;
#ifdef PNG_MAX_MALLOC_64K
		if(length > (png_uint_32)65535L)
		{
			png_warning(png_ptr, "unknown chunk too large to fit in memory");
			skip = length - (png_uint_32)65535L;
			length = (png_uint_32)65535L;
		}
#endif
		png_strcpy((png_charp)chunk.name, (png_charp)png_ptr->chunk_name);
		chunk.data = (png_bytep)png_malloc(png_ptr, length);
		chunk.size = (png_size_t)length;
		png_crc_read(png_ptr, (png_bytep)chunk.data, length);
#if defined(PNG_READ_USER_CHUNKS_SUPPORTED)
		if(png_ptr->read_user_chunk_fn != NULL)
		{
			if((*(png_ptr->read_user_chunk_fn)) (png_ptr, &chunk) <= 0)
			{
				if(!(png_ptr->chunk_name[0] & 0x20))
					if(png_handle_as_unknown(png_ptr, png_ptr->chunk_name) !=
						PNG_HANDLE_CHUNK_ALWAYS)
					{
						png_free(png_ptr, chunk.data);
						png_chunk_error(png_ptr, "unknown critical chunk");
					}
					png_set_unknown_chunks(png_ptr, info_ptr, &chunk, 1);
			}
		}
		else
#endif
			png_set_unknown_chunks(png_ptr, info_ptr, &chunk, 1);
		png_free(png_ptr, chunk.data);
	}
	else
#endif
		skip = length;
	png_crc_finish(png_ptr, skip);
#if !defined(PNG_READ_USER_CHUNKS_SUPPORTED)
	if(&info_ptr == NULL)
		return;
#endif
}
#define isnonalpha(c) ((c) < 65 || (c) > 122 || ((c) > 90 && (c) < 97))
void
png_check_chunk_name(png_structp png_ptr, png_bytep chunk_name)
{
	png_debug(1, "in png_check_chunk_name\n");
	if(isnonalpha(chunk_name[0]) || isnonalpha(chunk_name[1]) ||
		isnonalpha(chunk_name[2]) || isnonalpha(chunk_name[3]))
	{
		png_chunk_error(png_ptr, "invalid chunk type");
	}
}
#ifndef PNG_HAVE_ASSEMBLER_COMBINE_ROW
void
png_combine_row(png_structp png_ptr, png_bytep row, int mask)
{
	png_debug(1,"in png_combine_row\n");
	if(mask == 0xff)
	{
		png_memcpy(row, png_ptr->row_buf + 1,
			PNG_ROWBYTES(png_ptr->row_info.pixel_depth, png_ptr->width));
	}
	else
	{
		switch (png_ptr->row_info.pixel_depth)
		{
		case 1:
			{
				png_bytep sp = png_ptr->row_buf + 1;
				png_bytep dp = row;
				int s_inc, s_start, s_end;
				int m = 0x80;
				int shift;
				png_uint_32 i;
				png_uint_32 row_width = png_ptr->width;
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
				if(png_ptr->transformations & PNG_PACKSWAP)
				{
					s_start = 0;
					s_end = 7;
					s_inc = 1;
				}
				else
#endif
				{
					s_start = 7;
					s_end = 0;
					s_inc = -1;
				}
				shift = s_start;
				for(i = 0; i < row_width; i++)
				{
					if(m & mask)
					{
						int value;
						value = (*sp >> shift) & 0x01;
						*dp &= (png_byte)((0x7f7f >> (7 - shift)) & 0xff);
						*dp |= (png_byte)(value << shift);
					}
					if(shift == s_end)
					{
						shift = s_start;
						sp++;
						dp++;
					}
					else
						shift += s_inc;
					if(m == 1)
						m = 0x80;
					else
						m >>= 1;
				}
				break;
			}
		case 2:
			{
				png_bytep sp = png_ptr->row_buf + 1;
				png_bytep dp = row;
				int s_start, s_end, s_inc;
				int m = 0x80;
				int shift;
				png_uint_32 i;
				png_uint_32 row_width = png_ptr->width;
				int value;
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
				if(png_ptr->transformations & PNG_PACKSWAP)
				{
					s_start = 0;
					s_end = 6;
					s_inc = 2;
				}
				else
#endif
				{
					s_start = 6;
					s_end = 0;
					s_inc = -2;
				}
				shift = s_start;
				for(i = 0; i < row_width; i++)
				{
					if(m & mask)
					{
						value = (*sp >> shift) & 0x03;
						*dp &= (png_byte)((0x3f3f >> (6 - shift)) & 0xff);
						*dp |= (png_byte)(value << shift);
					}
					if(shift == s_end)
					{
						shift = s_start;
						sp++;
						dp++;
					}
					else
						shift += s_inc;
					if(m == 1)
						m = 0x80;
					else
						m >>= 1;
				}
				break;
			}
		case 4:
			{
				png_bytep sp = png_ptr->row_buf + 1;
				png_bytep dp = row;
				int s_start, s_end, s_inc;
				int m = 0x80;
				int shift;
				png_uint_32 i;
				png_uint_32 row_width = png_ptr->width;
				int value;
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
				if(png_ptr->transformations & PNG_PACKSWAP)
				{
					s_start = 0;
					s_end = 4;
					s_inc = 4;
				}
				else
#endif
				{
					s_start = 4;
					s_end = 0;
					s_inc = -4;
				}
				shift = s_start;
				for(i = 0; i < row_width; i++)
				{
					if(m & mask)
					{
						value = (*sp >> shift) & 0xf;
						*dp &= (png_byte)((0xf0f >> (4 - shift)) & 0xff);
						*dp |= (png_byte)(value << shift);
					}
					if(shift == s_end)
					{
						shift = s_start;
						sp++;
						dp++;
					}
					else
						shift += s_inc;
					if(m == 1)
						m = 0x80;
					else
						m >>= 1;
				}
				break;
			}
		default:
			{
				png_bytep sp = png_ptr->row_buf + 1;
				png_bytep dp = row;
				png_size_t pixel_bytes = (png_ptr->row_info.pixel_depth >> 3);
				png_uint_32 i;
				png_uint_32 row_width = png_ptr->width;
				png_byte m = 0x80;
				for(i = 0; i < row_width; i++)
				{
					if(m & mask)
					{
						png_memcpy(dp, sp, pixel_bytes);
					}
					sp += pixel_bytes;
					dp += pixel_bytes;
					if(m == 1)
						m = 0x80;
					else
						m >>= 1;
				}
				break;
			}
		}
	}
}
#endif
#ifdef PNG_READ_INTERLACING_SUPPORTED
#ifndef PNG_HAVE_ASSEMBLER_READ_INTERLACE
void
png_do_read_interlace(png_structp png_ptr)
{
	png_row_infop row_info = &(png_ptr->row_info);
	png_bytep row = png_ptr->row_buf + 1;
	int pass = png_ptr->pass;
	png_uint_32 transformations = png_ptr->transformations;
#ifdef PNG_USE_LOCAL_ARRAYS
	const int png_pass_inc[7] = {8, 8, 4, 4, 2, 2, 1};
#endif
	png_debug(1,"in png_do_read_interlace (stock C version)\n");
	if(row != NULL && row_info != NULL)
	{
		png_uint_32 final_width;
		final_width = row_info->width * png_pass_inc[pass];
		switch (row_info->pixel_depth)
		{
		case 1:
			{
				png_bytep sp = row + (png_size_t)((row_info->width - 1) >> 3);
				png_bytep dp = row + (png_size_t)((final_width - 1) >> 3);
				int sshift, dshift;
				int s_start, s_end, s_inc;
				int jstop = png_pass_inc[pass];
				png_byte v;
				png_uint_32 i;
				int j;
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
				if(transformations & PNG_PACKSWAP)
				{
					sshift = (int)((row_info->width + 7) & 0x07);
					dshift = (int)((final_width + 7) & 0x07);
					s_start = 7;
					s_end = 0;
					s_inc = -1;
				}
				else
#endif
				{
					sshift = 7 - (int)((row_info->width + 7) & 0x07);
					dshift = 7 - (int)((final_width + 7) & 0x07);
					s_start = 0;
					s_end = 7;
					s_inc = 1;
				}
				for(i = 0; i < row_info->width; i++)
				{
					v = (png_byte)((*sp >> sshift) & 0x01);
					for(j = 0; j < jstop; j++)
					{
						*dp &= (png_byte)((0x7f7f >> (7 - dshift)) & 0xff);
						*dp |= (png_byte)(v << dshift);
						if(dshift == s_end)
						{
							dshift = s_start;
							dp--;
						}
						else
							dshift += s_inc;
					}
					if(sshift == s_end)
					{
						sshift = s_start;
						sp--;
					}
					else
						sshift += s_inc;
				}
				break;
			}
		case 2:
			{
				png_bytep sp = row + (png_uint_32)((row_info->width - 1) >> 2);
				png_bytep dp = row + (png_uint_32)((final_width - 1) >> 2);
				int sshift, dshift;
				int s_start, s_end, s_inc;
				int jstop = png_pass_inc[pass];
				png_uint_32 i;
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
				if(transformations & PNG_PACKSWAP)
				{
					sshift = (int)(((row_info->width + 3) & 0x03) << 1);
					dshift = (int)(((final_width + 3) & 0x03) << 1);
					s_start = 6;
					s_end = 0;
					s_inc = -2;
				}
				else
#endif
				{
					sshift = (int)((3 - ((row_info->width + 3) & 0x03)) << 1);
					dshift = (int)((3 - ((final_width + 3) & 0x03)) << 1);
					s_start = 0;
					s_end = 6;
					s_inc = 2;
				}
				for(i = 0; i < row_info->width; i++)
				{
					png_byte v;
					int j;
					v = (png_byte)((*sp >> sshift) & 0x03);
					for(j = 0; j < jstop; j++)
					{
						*dp &= (png_byte)((0x3f3f >> (6 - dshift)) & 0xff);
						*dp |= (png_byte)(v << dshift);
						if(dshift == s_end)
						{
							dshift = s_start;
							dp--;
						}
						else
							dshift += s_inc;
					}
					if(sshift == s_end)
					{
						sshift = s_start;
						sp--;
					}
					else
						sshift += s_inc;
				}
				break;
			}
		case 4:
			{
				png_bytep sp = row + (png_size_t)((row_info->width - 1) >> 1);
				png_bytep dp = row + (png_size_t)((final_width - 1) >> 1);
				int sshift, dshift;
				int s_start, s_end, s_inc;
				png_uint_32 i;
				int jstop = png_pass_inc[pass];
#if defined(PNG_READ_PACKSWAP_SUPPORTED)
				if(transformations & PNG_PACKSWAP)
				{
					sshift = (int)(((row_info->width + 1) & 0x01) << 2);
					dshift = (int)(((final_width + 1) & 0x01) << 2);
					s_start = 4;
					s_end = 0;
					s_inc = -4;
				}
				else
#endif
				{
					sshift = (int)((1 - ((row_info->width + 1) & 0x01)) << 2);
					dshift = (int)((1 - ((final_width + 1) & 0x01)) << 2);
					s_start = 0;
					s_end = 4;
					s_inc = 4;
				}
				for(i = 0; i < row_info->width; i++)
				{
					png_byte v = (png_byte)((*sp >> sshift) & 0xf);
					int j;
					for(j = 0; j < jstop; j++)
					{
						*dp &= (png_byte)((0xf0f >> (4 - dshift)) & 0xff);
						*dp |= (png_byte)(v << dshift);
						if(dshift == s_end)
						{
							dshift = s_start;
							dp--;
						}
						else
							dshift += s_inc;
					}
					if(sshift == s_end)
					{
						sshift = s_start;
						sp--;
					}
					else
						sshift += s_inc;
				}
				break;
			}
		default:
			{
				png_size_t pixel_bytes = (row_info->pixel_depth >> 3);
				png_bytep sp = row + (png_size_t)(row_info->width - 1) * pixel_bytes;
				png_bytep dp = row + (png_size_t)(final_width - 1) * pixel_bytes;
				int jstop = png_pass_inc[pass];
				png_uint_32 i;
				for(i = 0; i < row_info->width; i++)
				{
					png_byte v[8];
					int j;
					png_memcpy(v, sp, pixel_bytes);
					for(j = 0; j < jstop; j++)
					{
						png_memcpy(dp, v, pixel_bytes);
						dp -= pixel_bytes;
					}
					sp -= pixel_bytes;
				}
				break;
			}
		}
		row_info->width = final_width;
		row_info->rowbytes = PNG_ROWBYTES(row_info->pixel_depth,final_width);
	}
#if !defined(PNG_READ_PACKSWAP_SUPPORTED)
	if(&transformations == NULL)
		return;
#endif
}
#endif
#endif
#ifndef PNG_HAVE_ASSEMBLER_READ_FILTER_ROW
void
png_read_filter_row(png_structp png_ptr, png_row_infop row_info, png_bytep row,
					png_bytep prev_row, int filter)
{
	png_debug(1, "in png_read_filter_row\n");
	png_debug2(2,"row = %lu, filter = %d\n", png_ptr->row_number, filter);
	switch (filter)
	{
	case PNG_FILTER_VALUE_NONE:
		break;
	case PNG_FILTER_VALUE_SUB:
		{
			png_uint_32 i;
			png_uint_32 istop = row_info->rowbytes;
			png_uint_32 bpp = (row_info->pixel_depth + 7) >> 3;
			png_bytep rp = row + bpp;
			png_bytep lp = row;
			for(i = bpp; i < istop; i++)
			{
				*rp = (png_byte)(((int)(*rp) + (int)(*lp++)) & 0xff);
				rp++;
			}
			break;
		}
	case PNG_FILTER_VALUE_UP:
		{
			png_uint_32 i;
			png_uint_32 istop = row_info->rowbytes;
			png_bytep rp = row;
			png_bytep pp = prev_row;
			for(i = 0; i < istop; i++)
			{
				*rp = (png_byte)(((int)(*rp) + (int)(*pp++)) & 0xff);
				rp++;
			}
			break;
		}
	case PNG_FILTER_VALUE_AVG:
		{
			png_uint_32 i;
			png_bytep rp = row;
			png_bytep pp = prev_row;
			png_bytep lp = row;
			png_uint_32 bpp = (row_info->pixel_depth + 7) >> 3;
			png_uint_32 istop = row_info->rowbytes - bpp;
			for(i = 0; i < bpp; i++)
			{
				*rp = (png_byte)(((int)(*rp) +
					((int)(*pp++) / 2 )) & 0xff);
				rp++;
			}
			for(i = 0; i < istop; i++)
			{
				*rp = (png_byte)(((int)(*rp) +
					(int)(*pp++ + *lp++) / 2 ) & 0xff);
				rp++;
			}
			break;
		}
	case PNG_FILTER_VALUE_PAETH:
		{
			png_uint_32 i;
			png_bytep rp = row;
			png_bytep pp = prev_row;
			png_bytep lp = row;
			png_bytep cp = prev_row;
			png_uint_32 bpp = (row_info->pixel_depth + 7) >> 3;
			png_uint_32 istop = row_info->rowbytes - bpp;
			for(i = 0; i < bpp; i++)
			{
				*rp = (png_byte)(((int)(*rp) + (int)(*pp++)) & 0xff);
				rp++;
			}
			for(i = 0; i < istop; i++)
			{
				int a, b, c, pa, pb, pc, p;
				a = *lp++;
				b = *pp++;
				c = *cp++;
				p = b - c;
				pc = a - c;
#ifdef PNG_USE_ABS
				pa = abs(p);
				pb = abs(pc);
				pc = abs(p + pc);
#else
				pa = p < 0 ? -p : p;
				pb = pc < 0 ? -pc : pc;
				pc = (p + pc) < 0 ? -(p + pc) : p + pc;
#endif
				p = (pa <= pb && pa <=pc) ? a : (pb <= pc) ? b : c;
				*rp = (png_byte)(((int)(*rp) + p) & 0xff);
				rp++;
			}
			break;
		}
	default:
		png_warning(png_ptr, "Ignoring bad adaptive filter type");
		*row = 0;
		break;
	}
}
#endif
void
png_read_finish_row(png_structp png_ptr)
{
#ifdef PNG_USE_LOCAL_ARRAYS
	const int png_pass_start[7] = {0, 4, 0, 2, 0, 1, 0};
	const int png_pass_inc[7] = {8, 8, 4, 4, 2, 2, 1};
	const int png_pass_ystart[7] = {0, 0, 4, 0, 2, 0, 1};
	const int png_pass_yinc[7] = {8, 8, 8, 4, 4, 2, 2};
#endif
	png_debug(1, "in png_read_finish_row\n");
	png_ptr->row_number++;
	if(png_ptr->row_number < png_ptr->num_rows)
		return;
	if(png_ptr->interlaced)
	{
		png_ptr->row_number = 0;
		png_memset_check(png_ptr, png_ptr->prev_row, 0, png_ptr->rowbytes + 1);
		do
		{
			png_ptr->pass++;
			if(png_ptr->pass >= 7)
				break;
			png_ptr->iwidth = (png_ptr->width +
				png_pass_inc[png_ptr->pass] - 1 -
				png_pass_start[png_ptr->pass]) /
				png_pass_inc[png_ptr->pass];
			png_ptr->irowbytes = PNG_ROWBYTES(png_ptr->pixel_depth,
				png_ptr->iwidth) + 1;
			if(!(png_ptr->transformations & PNG_INTERLACE))
			{
				png_ptr->num_rows = (png_ptr->height +
					png_pass_yinc[png_ptr->pass] - 1 -
					png_pass_ystart[png_ptr->pass]) /
					png_pass_yinc[png_ptr->pass];
				if(!(png_ptr->num_rows))
					continue;
			}
			else
				break;
		} while(png_ptr->iwidth == 0);
		if(png_ptr->pass < 7)
			return;
	}
	if(!(png_ptr->flags & PNG_FLAG_ZLIB_FINISHED))
	{
#ifdef PNG_USE_LOCAL_ARRAYS
		PNG_IDAT;
#endif
		char extra;
		int ret;
		png_ptr->zstream.next_out = (Byte *)&extra;
		png_ptr->zstream.avail_out = (uInt)1;
		for(;;)
		{
			if(!(png_ptr->zstream.avail_in))
			{
				while(!png_ptr->idat_size)
				{
					png_byte chunk_length[4];
					png_crc_finish(png_ptr, 0);
					png_read_data(png_ptr, chunk_length, 4);
					png_ptr->idat_size = png_get_uint_31(png_ptr, chunk_length);
					png_reset_crc(png_ptr);
					png_crc_read(png_ptr, png_ptr->chunk_name, 4);
					if(png_memcmp(png_ptr->chunk_name, (png_bytep)png_IDAT, 4))
						png_error(png_ptr, "Not enough image data");
				}
				png_ptr->zstream.avail_in = (uInt)png_ptr->zbuf_size;
				png_ptr->zstream.next_in = png_ptr->zbuf;
				if(png_ptr->zbuf_size > png_ptr->idat_size)
					png_ptr->zstream.avail_in = (uInt)png_ptr->idat_size;
				png_crc_read(png_ptr, png_ptr->zbuf, png_ptr->zstream.avail_in);
				png_ptr->idat_size -= png_ptr->zstream.avail_in;
			}
			ret = inflate(&png_ptr->zstream, Z_PARTIAL_FLUSH);
			if(ret == Z_STREAM_END)
			{
				if(!(png_ptr->zstream.avail_out) || png_ptr->zstream.avail_in ||
					png_ptr->idat_size)
					png_warning(png_ptr, "Extra compressed data");
				png_ptr->mode |= PNG_AFTER_IDAT;
				png_ptr->flags |= PNG_FLAG_ZLIB_FINISHED;
				break;
			}
			if(ret != Z_OK)
				png_error(png_ptr, png_ptr->zstream.msg ? png_ptr->zstream.msg :
			"Decompression Error");
			if(!(png_ptr->zstream.avail_out))
			{
				png_warning(png_ptr, "Extra compressed data.");
				png_ptr->mode |= PNG_AFTER_IDAT;
				png_ptr->flags |= PNG_FLAG_ZLIB_FINISHED;
				break;
			}
		}
		png_ptr->zstream.avail_out = 0;
	}
	if(png_ptr->idat_size || png_ptr->zstream.avail_in)
		png_warning(png_ptr, "Extra compression data");
	inflateReset(&png_ptr->zstream);
	png_ptr->mode |= PNG_AFTER_IDAT;
}
void
png_read_start_row(png_structp png_ptr)
{
#ifdef PNG_USE_LOCAL_ARRAYS
	const int png_pass_start[7] = {0, 4, 0, 2, 0, 1, 0};
	const int png_pass_inc[7] = {8, 8, 4, 4, 2, 2, 1};
	const int png_pass_ystart[7] = {0, 0, 4, 0, 2, 0, 1};
	const int png_pass_yinc[7] = {8, 8, 8, 4, 4, 2, 2};
#endif
	int max_pixel_depth;
	png_uint_32 row_bytes;
	png_debug(1, "in png_read_start_row\n");
	png_ptr->zstream.avail_in = 0;
	png_init_read_transformations(png_ptr);
	if(png_ptr->interlaced)
	{
		if(!(png_ptr->transformations & PNG_INTERLACE))
			png_ptr->num_rows = (png_ptr->height + png_pass_yinc[0] - 1 -
			png_pass_ystart[0]) / png_pass_yinc[0];
		else
			png_ptr->num_rows = png_ptr->height;
		png_ptr->iwidth = (png_ptr->width +
			png_pass_inc[png_ptr->pass] - 1 -
			png_pass_start[png_ptr->pass]) /
			png_pass_inc[png_ptr->pass];
		row_bytes = PNG_ROWBYTES(png_ptr->pixel_depth,png_ptr->iwidth) + 1;
		png_ptr->irowbytes = (png_uint_32)row_bytes;
		if((png_uint_32)png_ptr->irowbytes != row_bytes)
			png_error(png_ptr, "Rowbytes overflow in png_read_start_row");
	}
	else
	{
		png_ptr->num_rows = png_ptr->height;
		png_ptr->iwidth = png_ptr->width;
		png_ptr->irowbytes = png_ptr->rowbytes + 1;
	}
	max_pixel_depth = png_ptr->pixel_depth;
#if defined(PNG_READ_PACK_SUPPORTED)
	if((png_ptr->transformations & PNG_PACK) && png_ptr->bit_depth < 8)
		max_pixel_depth = 8;
#endif
#if defined(PNG_READ_EXPAND_SUPPORTED)
	if(png_ptr->transformations & PNG_EXPAND)
	{
		if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		{
			if(png_ptr->num_trans)
				max_pixel_depth = 32;
			else
				max_pixel_depth = 24;
		}
		else if(png_ptr->color_type == PNG_COLOR_TYPE_GRAY)
		{
			if(max_pixel_depth < 8)
				max_pixel_depth = 8;
			if(png_ptr->num_trans)
				max_pixel_depth *= 2;
		}
		else if(png_ptr->color_type == PNG_COLOR_TYPE_RGB)
		{
			if(png_ptr->num_trans)
			{
				max_pixel_depth *= 4;
				max_pixel_depth /= 3;
			}
		}
	}
#endif
#if defined(PNG_READ_FILLER_SUPPORTED)
	if(png_ptr->transformations & (PNG_FILLER))
	{
		if(png_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
			max_pixel_depth = 32;
		else if(png_ptr->color_type == PNG_COLOR_TYPE_GRAY)
		{
			if(max_pixel_depth <= 8)
				max_pixel_depth = 16;
			else
				max_pixel_depth = 32;
		}
		else if(png_ptr->color_type == PNG_COLOR_TYPE_RGB)
		{
			if(max_pixel_depth <= 32)
				max_pixel_depth = 32;
			else
				max_pixel_depth = 64;
		}
	}
#endif
#if defined(PNG_READ_GRAY_TO_RGB_SUPPORTED)
	if(png_ptr->transformations & PNG_GRAY_TO_RGB)
	{
		if(
#if defined(PNG_READ_EXPAND_SUPPORTED)
			(png_ptr->num_trans && (png_ptr->transformations & PNG_EXPAND)) ||
#endif
#if defined(PNG_READ_FILLER_SUPPORTED)
			(png_ptr->transformations & (PNG_FILLER)) ||
#endif
			png_ptr->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		{
			if(max_pixel_depth <= 16)
				max_pixel_depth = 32;
			else
				max_pixel_depth = 64;
		}
		else
		{
			if(max_pixel_depth <= 8)
			{
				if(png_ptr->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
					max_pixel_depth = 32;
				else
					max_pixel_depth = 24;
			}
			else if(png_ptr->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
				max_pixel_depth = 64;
			else
				max_pixel_depth = 48;
		}
	}
#endif
#if defined(PNG_READ_USER_TRANSFORM_SUPPORTED) && \
	defined(PNG_USER_TRANSFORM_PTR_SUPPORTED)
	if(png_ptr->transformations & PNG_USER_TRANSFORM)
	{
		int user_pixel_depth = png_ptr->user_transform_depth*
			png_ptr->user_transform_channels;
		if(user_pixel_depth > max_pixel_depth)
			max_pixel_depth = user_pixel_depth;
	}
#endif
	row_bytes = ((png_ptr->width + 7) & ~((png_uint_32)7));
	row_bytes = PNG_ROWBYTES(max_pixel_depth,row_bytes) +
		1 + ((max_pixel_depth + 7) >> 3);
#ifdef PNG_MAX_MALLOC_64K
	if(row_bytes > (png_uint_32)65536L)
		png_error(png_ptr, "This image requires a row greater than 64KB");
#endif
	png_ptr->big_row_buf = (png_bytep)png_malloc(png_ptr, row_bytes+64);
	png_ptr->row_buf = png_ptr->big_row_buf+32;
#if defined(PNG_DEBUG) && defined(PNG_USE_PNGGCCRD)
	png_ptr->row_buf_size = row_bytes;
#endif
#ifdef PNG_MAX_MALLOC_64K
	if((png_uint_32)png_ptr->rowbytes + 1 > (png_uint_32)65536L)
		png_error(png_ptr, "This image requires a row greater than 64KB");
#endif
	if((png_uint_32)png_ptr->rowbytes > PNG_SIZE_MAX - 1)
		png_error(png_ptr, "Row has too many bytes to allocate in memory.");
	png_ptr->prev_row = (png_bytep)png_malloc(png_ptr, (png_uint_32)(
		png_ptr->rowbytes + 1));
	png_memset_check(png_ptr, png_ptr->prev_row, 0, png_ptr->rowbytes + 1);
	png_debug1(3, "width = %lu,\n", png_ptr->width);
	png_debug1(3, "height = %lu,\n", png_ptr->height);
	png_debug1(3, "iwidth = %lu,\n", png_ptr->iwidth);
	png_debug1(3, "num_rows = %lu\n", png_ptr->num_rows);
	png_debug1(3, "rowbytes = %lu,\n", png_ptr->rowbytes);
	png_debug1(3, "irowbytes = %lu,\n", png_ptr->irowbytes);
	png_ptr->flags |= PNG_FLAG_ROW_INIT;
}
#define PNG_INTERNAL
#include "png.h"
#if defined(PNG_bKGD_SUPPORTED)
void PNGAPI
png_set_bKGD(png_structp png_ptr, png_infop info_ptr, png_color_16p background)
{
	png_debug1(1, "in %s storage function\n", "bKGD");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	png_memcpy(&(info_ptr->background), background, png_sizeof(png_color_16));
	info_ptr->valid |= PNG_INFO_bKGD;
}
#endif
#if defined(PNG_cHRM_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
void PNGAPI
png_set_cHRM(png_structp png_ptr, png_infop info_ptr,
			 double white_x, double white_y, double red_x, double red_y,
			 double green_x, double green_y, double blue_x, double blue_y)
{
	png_debug1(1, "in %s storage function\n", "cHRM");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(white_x < 0.0 || white_y < 0.0 ||
		red_x < 0.0 ||   red_y < 0.0 ||
		green_x < 0.0 || green_y < 0.0 ||
		blue_x < 0.0 ||  blue_y < 0.0)
	{
		png_warning(png_ptr,
			"Ignoring attempt to set negative chromaticity value");
		return;
	}
	if(white_x > 21474.83 || white_y > 21474.83 ||
		red_x > 21474.83 ||   red_y > 21474.83 ||
		green_x > 21474.83 || green_y > 21474.83 ||
		blue_x > 21474.83 ||  blue_y > 21474.83)
	{
		png_warning(png_ptr,
			"Ignoring attempt to set chromaticity value exceeding 21474.83");
		return;
	}
	info_ptr->x_white = (float)white_x;
	info_ptr->y_white = (float)white_y;
	info_ptr->x_red   = (float)red_x;
	info_ptr->y_red   = (float)red_y;
	info_ptr->x_green = (float)green_x;
	info_ptr->y_green = (float)green_y;
	info_ptr->x_blue  = (float)blue_x;
	info_ptr->y_blue  = (float)blue_y;
#ifdef PNG_FIXED_POINT_SUPPORTED
	info_ptr->int_x_white = (png_fixed_point)(white_x*100000.+0.5);
	info_ptr->int_y_white = (png_fixed_point)(white_y*100000.+0.5);
	info_ptr->int_x_red   = (png_fixed_point)(  red_x*100000.+0.5);
	info_ptr->int_y_red   = (png_fixed_point)(  red_y*100000.+0.5);
	info_ptr->int_x_green = (png_fixed_point)(green_x*100000.+0.5);
	info_ptr->int_y_green = (png_fixed_point)(green_y*100000.+0.5);
	info_ptr->int_x_blue  = (png_fixed_point)( blue_x*100000.+0.5);
	info_ptr->int_y_blue  = (png_fixed_point)( blue_y*100000.+0.5);
#endif
	info_ptr->valid |= PNG_INFO_cHRM;
}
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
void PNGAPI
png_set_cHRM_fixed(png_structp png_ptr, png_infop info_ptr,
				   png_fixed_point white_x, png_fixed_point white_y, png_fixed_point red_x,
				   png_fixed_point red_y, png_fixed_point green_x, png_fixed_point green_y,
				   png_fixed_point blue_x, png_fixed_point blue_y)
{
	png_debug1(1, "in %s storage function\n", "cHRM");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(white_x < 0 || white_y < 0 ||
		red_x < 0 ||   red_y < 0 ||
		green_x < 0 || green_y < 0 ||
		blue_x < 0 ||  blue_y < 0)
	{
		png_warning(png_ptr,
			"Ignoring attempt to set negative chromaticity value");
		return;
	}
	if(white_x > (double) PNG_UINT_31_MAX ||
		white_y > (double) PNG_UINT_31_MAX ||
		red_x > (double) PNG_UINT_31_MAX ||
		red_y > (double) PNG_UINT_31_MAX ||
		green_x > (double) PNG_UINT_31_MAX ||
		green_y > (double) PNG_UINT_31_MAX ||
		blue_x > (double) PNG_UINT_31_MAX ||
		blue_y > (double) PNG_UINT_31_MAX)
	{
		png_warning(png_ptr,
			"Ignoring attempt to set chromaticity value exceeding 21474.83");
		return;
	}
	info_ptr->int_x_white = white_x;
	info_ptr->int_y_white = white_y;
	info_ptr->int_x_red   = red_x;
	info_ptr->int_y_red   = red_y;
	info_ptr->int_x_green = green_x;
	info_ptr->int_y_green = green_y;
	info_ptr->int_x_blue  = blue_x;
	info_ptr->int_y_blue  = blue_y;
#ifdef PNG_FLOATING_POINT_SUPPORTED
	info_ptr->x_white = (float)(white_x/100000.);
	info_ptr->y_white = (float)(white_y/100000.);
	info_ptr->x_red   = (float)(  red_x/100000.);
	info_ptr->y_red   = (float)(  red_y/100000.);
	info_ptr->x_green = (float)(green_x/100000.);
	info_ptr->y_green = (float)(green_y/100000.);
	info_ptr->x_blue  = (float)( blue_x/100000.);
	info_ptr->y_blue  = (float)( blue_y/100000.);
#endif
	info_ptr->valid |= PNG_INFO_cHRM;
}
#endif
#endif
#if defined(PNG_gAMA_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
void PNGAPI
png_set_gAMA(png_structp png_ptr, png_infop info_ptr, double file_gamma)
{
	double gamma;
	png_debug1(1, "in %s storage function\n", "gAMA");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(file_gamma > 21474.83)
	{
		png_warning(png_ptr, "Limiting gamma to 21474.83");
		gamma = 21474.83;
	}
	else
		gamma = file_gamma;
	info_ptr->gamma = (float)gamma;
#ifdef PNG_FIXED_POINT_SUPPORTED
	info_ptr->int_gamma = (int)(gamma*100000.+.5);
#endif
	info_ptr->valid |= PNG_INFO_gAMA;
	if(gamma == 0.0)
		png_warning(png_ptr, "Setting gamma = 0");
}
#endif
void PNGAPI
png_set_gAMA_fixed(png_structp png_ptr, png_infop info_ptr, png_fixed_point
				   int_gamma)
{
	png_fixed_point gamma;
	png_debug1(1, "in %s storage function\n", "gAMA");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(int_gamma > (png_fixed_point) PNG_UINT_31_MAX)
	{
		png_warning(png_ptr, "Limiting gamma to 21474.83");
		gamma = PNG_UINT_31_MAX;
	}
	else
	{
		if(int_gamma < 0)
		{
			png_warning(png_ptr, "Setting negative gamma to zero");
			gamma = 0;
		}
		else
			gamma = int_gamma;
	}
#ifdef PNG_FLOATING_POINT_SUPPORTED
	info_ptr->gamma = (float)(gamma/100000.);
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
	info_ptr->int_gamma = gamma;
#endif
	info_ptr->valid |= PNG_INFO_gAMA;
	if(gamma == 0)
		png_warning(png_ptr, "Setting gamma = 0");
}
#endif
#if defined(PNG_hIST_SUPPORTED)
void PNGAPI
png_set_hIST(png_structp png_ptr, png_infop info_ptr, png_uint_16p hist)
{
	int i;
	png_debug1(1, "in %s storage function\n", "hIST");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(info_ptr->num_palette == 0)
	{
		png_warning(png_ptr,
			"Palette size 0, hIST allocation skipped.");
		return;
	}
#ifdef PNG_FREE_ME_SUPPORTED
	png_free_data(png_ptr, info_ptr, PNG_FREE_HIST, 0);
#endif
	png_ptr->hist = (png_uint_16p)png_malloc_warn(png_ptr,
		(png_uint_32)(256 * png_sizeof (png_uint_16)));
	if(png_ptr->hist == NULL)
	{
		png_warning(png_ptr, "Insufficient memory for hIST chunk data.");
		return;
	}
	for(i = 0; i < info_ptr->num_palette; i++)
		png_ptr->hist[i] = hist[i];
	info_ptr->hist = png_ptr->hist;
	info_ptr->valid |= PNG_INFO_hIST;
#ifdef PNG_FREE_ME_SUPPORTED
	info_ptr->free_me |= PNG_FREE_HIST;
#else
	png_ptr->flags |= PNG_FLAG_FREE_HIST;
#endif
}
#endif
void PNGAPI
png_set_IHDR(png_structp png_ptr, png_infop info_ptr,
			 png_uint_32 width, png_uint_32 height, int bit_depth,
			 int color_type, int interlace_type, int compression_type,
			 int filter_type)
{
	png_debug1(1, "in %s storage function\n", "IHDR");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(width == 0 || height == 0)
		png_error(png_ptr, "Image width or height is zero in IHDR");
#ifdef PNG_SET_USER_LIMITS_SUPPORTED
	if(width > png_ptr->user_width_max || height > png_ptr->user_height_max)
		png_error(png_ptr, "image size exceeds user limits in IHDR");
#else
	if(width > PNG_USER_WIDTH_MAX || height > PNG_USER_HEIGHT_MAX)
		png_error(png_ptr, "image size exceeds user limits in IHDR");
#endif
	if(width > PNG_UINT_31_MAX || height > PNG_UINT_31_MAX)
		png_error(png_ptr, "Invalid image size in IHDR");
	if( width > (PNG_UINT_32_MAX
		>> 3)
		- 64
		- 1
		- 7*8
		- 8)
		png_warning(png_ptr, "Width is too large for libpng to process pixels");
	if(bit_depth != 1 && bit_depth != 2 && bit_depth != 4 &&
		bit_depth != 8 && bit_depth != 16)
		png_error(png_ptr, "Invalid bit depth in IHDR");
	if(color_type < 0 || color_type == 1 ||
		color_type == 5 || color_type > 6)
		png_error(png_ptr, "Invalid color type in IHDR");
	if(((color_type == PNG_COLOR_TYPE_PALETTE) && bit_depth > 8) ||
		((color_type == PNG_COLOR_TYPE_RGB ||
		color_type == PNG_COLOR_TYPE_GRAY_ALPHA ||
		color_type == PNG_COLOR_TYPE_RGB_ALPHA) && bit_depth < 8))
		png_error(png_ptr, "Invalid color type/bit depth combination in IHDR");
	if(interlace_type >= PNG_INTERLACE_LAST)
		png_error(png_ptr, "Unknown interlace method in IHDR");
	if(compression_type != PNG_COMPRESSION_TYPE_BASE)
		png_error(png_ptr, "Unknown compression method in IHDR");
#if defined(PNG_MNG_FEATURES_SUPPORTED)
	if((png_ptr->mode&PNG_HAVE_PNG_SIGNATURE)&&png_ptr->mng_features_permitted)
		png_warning(png_ptr,"MNG features are not allowed in a PNG datastream\n");
	if(filter_type != PNG_FILTER_TYPE_BASE)
	{
		if(!((png_ptr->mng_features_permitted & PNG_FLAG_MNG_FILTER_64) &&
			(filter_type == PNG_INTRAPIXEL_DIFFERENCING) &&
			((png_ptr->mode&PNG_HAVE_PNG_SIGNATURE) == 0) &&
			(color_type == PNG_COLOR_TYPE_RGB ||
			color_type == PNG_COLOR_TYPE_RGB_ALPHA)))
			png_error(png_ptr, "Unknown filter method in IHDR");
		if(png_ptr->mode&PNG_HAVE_PNG_SIGNATURE)
			png_warning(png_ptr, "Invalid filter method in IHDR");
	}
#else
	if(filter_type != PNG_FILTER_TYPE_BASE)
		png_error(png_ptr, "Unknown filter method in IHDR");
#endif
	info_ptr->width = width;
	info_ptr->height = height;
	info_ptr->bit_depth = (png_byte)bit_depth;
	info_ptr->color_type =(png_byte) color_type;
	info_ptr->compression_type = (png_byte)compression_type;
	info_ptr->filter_type = (png_byte)filter_type;
	info_ptr->interlace_type = (png_byte)interlace_type;
	if(info_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
		info_ptr->channels = 1;
	else if(info_ptr->color_type & PNG_COLOR_MASK_COLOR)
		info_ptr->channels = 3;
	else
		info_ptr->channels = 1;
	if(info_ptr->color_type & PNG_COLOR_MASK_ALPHA)
		info_ptr->channels++;
	info_ptr->pixel_depth = (png_byte)(info_ptr->channels * info_ptr->bit_depth);
	if( width > (PNG_UINT_32_MAX
		>> 3)
		- 64
		- 1
		- 7*8
		- 8)
		info_ptr->rowbytes = (png_size_t)0;
	else
		info_ptr->rowbytes = PNG_ROWBYTES(info_ptr->pixel_depth,width);
}
#if defined(PNG_oFFs_SUPPORTED)
void PNGAPI
png_set_oFFs(png_structp png_ptr, png_infop info_ptr,
			 png_int_32 offset_x, png_int_32 offset_y, int unit_type)
{
	png_debug1(1, "in %s storage function\n", "oFFs");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	info_ptr->x_offset = offset_x;
	info_ptr->y_offset = offset_y;
	info_ptr->offset_unit_type = (png_byte)unit_type;
	info_ptr->valid |= PNG_INFO_oFFs;
}
#endif
#if defined(PNG_pCAL_SUPPORTED)
void PNGAPI
png_set_pCAL(png_structp png_ptr, png_infop info_ptr,
			 png_charp purpose, png_int_32 X0, png_int_32 X1, int type, int nparams,
			 png_charp units, png_charpp params)
{
	png_uint_32 length;
	int i;
	png_debug1(1, "in %s storage function\n", "pCAL");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	length = (png_uint_32)(png_strlen(purpose) + 1);
	png_debug1(3, "allocating purpose for info (%lu bytes)\n", length);
	info_ptr->pcal_purpose = (png_charp)png_malloc_warn(png_ptr, length);
	if(info_ptr->pcal_purpose == NULL)
	{
		png_warning(png_ptr, "Insufficient memory for pCAL purpose.");
		return;
	}
	png_memcpy(info_ptr->pcal_purpose, purpose, (png_size_t)length);
	png_debug(3, "storing X0, X1, type, and nparams in info\n");
	info_ptr->pcal_X0 = X0;
	info_ptr->pcal_X1 = X1;
	info_ptr->pcal_type = (png_byte)type;
	info_ptr->pcal_nparams = (png_byte)nparams;
	length = (png_uint_32)(png_strlen(units) + 1);
	png_debug1(3, "allocating units for info (%lu bytes)\n", length);
	info_ptr->pcal_units = (png_charp)png_malloc_warn(png_ptr, length);
	if(info_ptr->pcal_units == NULL)
	{
		png_warning(png_ptr, "Insufficient memory for pCAL units.");
		return;
	}
	png_memcpy(info_ptr->pcal_units, units, (png_size_t)length);
	info_ptr->pcal_params = (png_charpp)png_malloc_warn(png_ptr,
		(png_uint_32)((nparams + 1) * png_sizeof(png_charp)));
	if(info_ptr->pcal_params == NULL)
	{
		png_warning(png_ptr, "Insufficient memory for pCAL params.");
		return;
	}
	info_ptr->pcal_params[nparams] = NULL;
	for(i = 0; i < nparams; i++)
	{
		length = (png_uint_32)(png_strlen(params[i]) + 1);
		png_debug2(3, "allocating parameter %d for info (%lu bytes)\n", i, length);
		info_ptr->pcal_params[i] = (png_charp)png_malloc_warn(png_ptr, length);
		if(info_ptr->pcal_params[i] == NULL)
		{
			png_warning(png_ptr, "Insufficient memory for pCAL parameter.");
			return;
		}
		png_memcpy(info_ptr->pcal_params[i], params[i], (png_size_t)length);
	}
	info_ptr->valid |= PNG_INFO_pCAL;
#ifdef PNG_FREE_ME_SUPPORTED
	info_ptr->free_me |= PNG_FREE_PCAL;
#endif
}
#endif
#if defined(PNG_READ_sCAL_SUPPORTED) || defined(PNG_WRITE_sCAL_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
void PNGAPI
png_set_sCAL(png_structp png_ptr, png_infop info_ptr,
			 int unit, double width, double height)
{
	png_debug1(1, "in %s storage function\n", "sCAL");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	info_ptr->scal_unit = (png_byte)unit;
	info_ptr->scal_pixel_width = width;
	info_ptr->scal_pixel_height = height;
	info_ptr->valid |= PNG_INFO_sCAL;
}
#else
#ifdef PNG_FIXED_POINT_SUPPORTED
void PNGAPI
png_set_sCAL_s(png_structp png_ptr, png_infop info_ptr,
			   int unit, png_charp swidth, png_charp sheight)
{
	png_uint_32 length;
	png_debug1(1, "in %s storage function\n", "sCAL");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	info_ptr->scal_unit = (png_byte)unit;
	length = png_strlen(swidth) + 1;
	png_debug1(3, "allocating unit for info (%d bytes)\n", length);
	info_ptr->scal_s_width = (png_charp)png_malloc_warn(png_ptr, length);
	if(info_ptr->scal_s_width == NULL)
	{
		png_warning(png_ptr, "Memory allocation failed while processing sCAL.");
	}
	png_memcpy(info_ptr->scal_s_width, swidth, (png_size_t)length);
	length = png_strlen(sheight) + 1;
	png_debug1(3, "allocating unit for info (%d bytes)\n", length);
	info_ptr->scal_s_height = (png_charp)png_malloc_warn(png_ptr, length);
	if(info_ptr->scal_s_height == NULL)
	{
		png_free (png_ptr, info_ptr->scal_s_width);
		png_warning(png_ptr, "Memory allocation failed while processing sCAL.");
	}
	png_memcpy(info_ptr->scal_s_height, sheight, (png_size_t)length);
	info_ptr->valid |= PNG_INFO_sCAL;
#ifdef PNG_FREE_ME_SUPPORTED
	info_ptr->free_me |= PNG_FREE_SCAL;
#endif
}
#endif
#endif
#endif
#if defined(PNG_pHYs_SUPPORTED)
void PNGAPI
png_set_pHYs(png_structp png_ptr, png_infop info_ptr,
			 png_uint_32 res_x, png_uint_32 res_y, int unit_type)
{
	png_debug1(1, "in %s storage function\n", "pHYs");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	info_ptr->x_pixels_per_unit = res_x;
	info_ptr->y_pixels_per_unit = res_y;
	info_ptr->phys_unit_type = (png_byte)unit_type;
	info_ptr->valid |= PNG_INFO_pHYs;
}
#endif
void PNGAPI
png_set_PLTE(png_structp png_ptr, png_infop info_ptr,
			 png_colorp palette, int num_palette)
{
	png_debug1(1, "in %s storage function\n", "PLTE");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
#ifdef PNG_FREE_ME_SUPPORTED
	png_free_data(png_ptr, info_ptr, PNG_FREE_PLTE, 0);
#endif
	png_ptr->palette = (png_colorp)png_malloc(png_ptr,
		256 * png_sizeof(png_color));
	png_memset(png_ptr->palette, 0, 256 * png_sizeof(png_color));
	png_memcpy(png_ptr->palette, palette, num_palette * png_sizeof (png_color));
	info_ptr->palette = png_ptr->palette;
	info_ptr->num_palette = png_ptr->num_palette = (png_uint_16)num_palette;
#ifdef PNG_FREE_ME_SUPPORTED
	info_ptr->free_me |= PNG_FREE_PLTE;
#else
	png_ptr->flags |= PNG_FLAG_FREE_PLTE;
#endif
	info_ptr->valid |= PNG_INFO_PLTE;
}
#if defined(PNG_sBIT_SUPPORTED)
void PNGAPI
png_set_sBIT(png_structp png_ptr, png_infop info_ptr,
			 png_color_8p sig_bit)
{
	png_debug1(1, "in %s storage function\n", "sBIT");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	png_memcpy(&(info_ptr->sig_bit), sig_bit, png_sizeof (png_color_8));
	info_ptr->valid |= PNG_INFO_sBIT;
}
#endif
#if defined(PNG_sRGB_SUPPORTED)
void PNGAPI
png_set_sRGB(png_structp png_ptr, png_infop info_ptr, int intent)
{
	png_debug1(1, "in %s storage function\n", "sRGB");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	info_ptr->srgb_intent = (png_byte)intent;
	info_ptr->valid |= PNG_INFO_sRGB;
}
void PNGAPI
png_set_sRGB_gAMA_and_cHRM(png_structp png_ptr, png_infop info_ptr,
						   int intent)
{
#if defined(PNG_gAMA_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
	float file_gamma;
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
	png_fixed_point int_file_gamma;
#endif
#endif
#if defined(PNG_cHRM_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
	float white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y;
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
	png_fixed_point int_white_x, int_white_y, int_red_x, int_red_y, int_green_x,
		int_green_y, int_blue_x, int_blue_y;
#endif
#endif
	png_debug1(1, "in %s storage function\n", "sRGB_gAMA_and_cHRM");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	png_set_sRGB(png_ptr, info_ptr, intent);
#if defined(PNG_gAMA_SUPPORTED)
#ifdef PNG_FLOATING_POINT_SUPPORTED
	file_gamma = (float).45455;
	png_set_gAMA(png_ptr, info_ptr, file_gamma);
#endif
#ifdef PNG_FIXED_POINT_SUPPORTED
	int_file_gamma = 45455L;
	png_set_gAMA_fixed(png_ptr, info_ptr, int_file_gamma);
#endif
#endif
#if defined(PNG_cHRM_SUPPORTED)
#ifdef PNG_FIXED_POINT_SUPPORTED
	int_white_x = 31270L;
	int_white_y = 32900L;
	int_red_x   = 64000L;
	int_red_y   = 33000L;
	int_green_x = 30000L;
	int_green_y = 60000L;
	int_blue_x  = 15000L;
	int_blue_y  =  6000L;
	png_set_cHRM_fixed(png_ptr, info_ptr,
		int_white_x, int_white_y, int_red_x, int_red_y, int_green_x, int_green_y,
		int_blue_x, int_blue_y);
#endif
#ifdef PNG_FLOATING_POINT_SUPPORTED
	white_x = (float).3127;
	white_y = (float).3290;
	red_x   = (float).64;
	red_y   = (float).33;
	green_x = (float).30;
	green_y = (float).60;
	blue_x  = (float).15;
	blue_y  = (float).06;
	png_set_cHRM(png_ptr, info_ptr,
		white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y);
#endif
#endif
}
#endif
#if defined(PNG_iCCP_SUPPORTED)
void PNGAPI
png_set_iCCP(png_structp png_ptr, png_infop info_ptr,
			 png_charp name, int compression_type,
			 png_charp profile, png_uint_32 proflen)
{
	png_charp new_iccp_name;
	png_charp new_iccp_profile;
	png_debug1(1, "in %s storage function\n", "iCCP");
	if(png_ptr == NULL || info_ptr == NULL || name == NULL || profile == NULL)
		return;
	new_iccp_name = (png_charp)png_malloc_warn(png_ptr, (png_uint_32)(png_strlen(name)+1));
	if(new_iccp_name == NULL)
	{
		png_warning(png_ptr, "Insufficient memory to process iCCP chunk.");
		return;
	}
	png_strcpy(new_iccp_name, name);
	new_iccp_profile = (png_charp)png_malloc_warn(png_ptr, proflen);
	if(new_iccp_profile == NULL)
	{
		png_free (png_ptr, new_iccp_name);
		png_warning(png_ptr, "Insufficient memory to process iCCP profile.");
		return;
	}
	png_memcpy(new_iccp_profile, profile, (png_size_t)proflen);
	png_free_data(png_ptr, info_ptr, PNG_FREE_ICCP, 0);
	info_ptr->iccp_proflen = proflen;
	info_ptr->iccp_name = new_iccp_name;
	info_ptr->iccp_profile = new_iccp_profile;
	info_ptr->iccp_compression = (png_byte)compression_type;
#ifdef PNG_FREE_ME_SUPPORTED
	info_ptr->free_me |= PNG_FREE_ICCP;
#endif
	info_ptr->valid |= PNG_INFO_iCCP;
}
#endif
#if defined(PNG_TEXT_SUPPORTED)
void PNGAPI
png_set_text(png_structp png_ptr, png_infop info_ptr, png_textp text_ptr,
			 int num_text)
{
	int ret;
	ret = png_set_text_2(png_ptr, info_ptr, text_ptr, num_text);
	if(ret)
		png_error(png_ptr, "Insufficient memory to store text");
}
int
png_set_text_2(png_structp png_ptr, png_infop info_ptr, png_textp text_ptr,
			   int num_text)
{
	int i;
	png_debug1(1, "in %s storage function\n", (png_ptr->chunk_name[0] == '\0' ?
		"text" : (png_const_charp)png_ptr->chunk_name));
	if(png_ptr == NULL || info_ptr == NULL || num_text == 0)
		return(0);
	if(info_ptr->num_text + num_text > info_ptr->max_text)
	{
		if(info_ptr->text != NULL)
		{
			png_textp old_text;
			int old_max;
			old_max = info_ptr->max_text;
			info_ptr->max_text = info_ptr->num_text + num_text + 8;
			old_text = info_ptr->text;
			info_ptr->text = (png_textp)png_malloc_warn(png_ptr,
				(png_uint_32)(info_ptr->max_text * png_sizeof (png_text)));
			if(info_ptr->text == NULL)
			{
				png_free(png_ptr, old_text);
				return(1);
			}
			png_memcpy(info_ptr->text, old_text, (png_size_t)(old_max *
				png_sizeof(png_text)));
			png_free(png_ptr, old_text);
		}
		else
		{
			info_ptr->max_text = num_text + 8;
			info_ptr->num_text = 0;
			info_ptr->text = (png_textp)png_malloc_warn(png_ptr,
				(png_uint_32)(info_ptr->max_text * png_sizeof (png_text)));
			if(info_ptr->text == NULL)
				return(1);
#ifdef PNG_FREE_ME_SUPPORTED
			info_ptr->free_me |= PNG_FREE_TEXT;
#endif
		}
		png_debug1(3, "allocated %d entries for info_ptr->text\n",
			info_ptr->max_text);
	}
	for(i = 0; i < num_text; i++)
	{
		png_size_t text_length,key_len;
		png_size_t lang_len,lang_key_len;
		png_textp textp = &(info_ptr->text[info_ptr->num_text]);
		if(text_ptr[i].key == NULL)
			continue;
		key_len = png_strlen(text_ptr[i].key);
		if(text_ptr[i].compression <= 0)
		{
			lang_len = 0;
			lang_key_len = 0;
		}
		else
#ifdef PNG_iTXt_SUPPORTED
		{
			if(text_ptr[i].lang != NULL)
				lang_len = png_strlen(text_ptr[i].lang);
			else
				lang_len = 0;
			if(text_ptr[i].lang_key != NULL)
				lang_key_len = png_strlen(text_ptr[i].lang_key);
			else
				lang_key_len = 0;
		}
#else
		{
			png_warning(png_ptr, "iTXt chunk not supported.");
			continue;
		}
#endif
		if(text_ptr[i].text == NULL || text_ptr[i].text[0] == '\0')
		{
			text_length = 0;
#ifdef PNG_iTXt_SUPPORTED
			if(text_ptr[i].compression > 0)
				textp->compression = PNG_ITXT_COMPRESSION_NONE;
			else
#endif
				textp->compression = PNG_TEXT_COMPRESSION_NONE;
		}
		else
		{
			text_length = png_strlen(text_ptr[i].text);
			textp->compression = text_ptr[i].compression;
		}
		textp->key = (png_charp)png_malloc_warn(png_ptr,
			(png_uint_32)(key_len + text_length + lang_len + lang_key_len + 4));
		if(textp->key == NULL)
			return(1);
		png_debug2(2, "Allocated %lu bytes at %x in png_set_text\n",
			(png_uint_32)(key_len + lang_len + lang_key_len + text_length + 4),
			(int)textp->key);
		png_memcpy(textp->key, text_ptr[i].key,
			(png_size_t)(key_len));
		*(textp->key+key_len) = '\0';
#ifdef PNG_iTXt_SUPPORTED
		if(text_ptr[i].compression > 0)
		{
			textp->lang = textp->key + key_len + 1;
			png_memcpy(textp->lang, text_ptr[i].lang, lang_len);
			*(textp->lang+lang_len) = '\0';
			textp->lang_key = textp->lang + lang_len + 1;
			png_memcpy(textp->lang_key, text_ptr[i].lang_key, lang_key_len);
			*(textp->lang_key+lang_key_len) = '\0';
			textp->text = textp->lang_key + lang_key_len + 1;
		}
		else
#endif
		{
#ifdef PNG_iTXt_SUPPORTED
			textp->lang = NULL;
			textp->lang_key = NULL;
#endif
			textp->text = textp->key + key_len + 1;
		}
		if(text_length)
			png_memcpy(textp->text, text_ptr[i].text,
			(png_size_t)(text_length));
		*(textp->text+text_length) = '\0';
#ifdef PNG_iTXt_SUPPORTED
		if(textp->compression > 0)
		{
			textp->text_length = 0;
			textp->itxt_length = text_length;
		}
		else
#endif
		{
			textp->text_length = text_length;
#ifdef PNG_iTXt_SUPPORTED
			textp->itxt_length = 0;
#endif
		}
		info_ptr->text[info_ptr->num_text]= *textp;
		info_ptr->num_text++;
		png_debug1(3, "transferred text chunk %d\n", info_ptr->num_text);
	}
	return(0);
}
#endif
#if defined(PNG_tIME_SUPPORTED)
void PNGAPI
png_set_tIME(png_structp png_ptr, png_infop info_ptr, png_timep mod_time)
{
	png_debug1(1, "in %s storage function\n", "tIME");
	if(png_ptr == NULL || info_ptr == NULL ||
		(png_ptr->mode & PNG_WROTE_tIME))
		return;
	png_memcpy(&(info_ptr->mod_time), mod_time, png_sizeof (png_time));
	info_ptr->valid |= PNG_INFO_tIME;
}
#endif
#if defined(PNG_tRNS_SUPPORTED)
void PNGAPI
png_set_tRNS(png_structp png_ptr, png_infop info_ptr,
			 png_bytep trans, int num_trans, png_color_16p trans_values)
{
	png_debug1(1, "in %s storage function\n", "tRNS");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(trans != NULL)
	{
#ifdef PNG_FREE_ME_SUPPORTED
		png_free_data(png_ptr, info_ptr, PNG_FREE_TRNS, 0);
#endif
		png_ptr->trans = info_ptr->trans = (png_bytep)png_malloc(png_ptr,
			(png_uint_32)256);
		png_memcpy(info_ptr->trans, trans, (png_size_t)num_trans);
#ifdef PNG_FREE_ME_SUPPORTED
		info_ptr->free_me |= PNG_FREE_TRNS;
#else
		png_ptr->flags |= PNG_FLAG_FREE_TRNS;
#endif
	}
	if(trans_values != NULL)
	{
		png_memcpy(&(info_ptr->trans_values), trans_values,
			png_sizeof(png_color_16));
		if(num_trans == 0)
			num_trans = 1;
	}
	info_ptr->num_trans = (png_uint_16)num_trans;
	info_ptr->valid |= PNG_INFO_tRNS;
}
#endif
#if defined(PNG_sPLT_SUPPORTED)
void PNGAPI
png_set_sPLT(png_structp png_ptr,
			 png_infop info_ptr, png_sPLT_tp entries, int nentries)
{
	png_sPLT_tp np;
	int i;
	np = (png_sPLT_tp)png_malloc_warn(png_ptr,
		(info_ptr->splt_palettes_num + nentries) * png_sizeof(png_sPLT_t));
	if(np == NULL)
	{
		png_warning(png_ptr, "No memory for sPLT palettes.");
		return;
	}
	png_memcpy(np, info_ptr->splt_palettes,
		info_ptr->splt_palettes_num * png_sizeof(png_sPLT_t));
	png_free(png_ptr, info_ptr->splt_palettes);
	info_ptr->splt_palettes = NULL;
	for(i = 0; i < nentries; i++)
	{
		png_sPLT_tp to = np + info_ptr->splt_palettes_num + i;
		png_sPLT_tp from = entries + i;
		to->name = (png_charp)png_malloc(png_ptr,
			(png_uint_32)(png_strlen(from->name) + 1));
		png_strcpy(to->name, from->name);
		to->entries = (png_sPLT_entryp)png_malloc(png_ptr,
			from->nentries * png_sizeof(png_sPLT_t));
		png_memcpy(to->entries, from->entries,
			from->nentries * png_sizeof(png_sPLT_t));
		to->nentries = from->nentries;
		to->depth = from->depth;
	}
	info_ptr->splt_palettes = np;
	info_ptr->splt_palettes_num += nentries;
	info_ptr->valid |= PNG_INFO_sPLT;
#ifdef PNG_FREE_ME_SUPPORTED
	info_ptr->free_me |= PNG_FREE_SPLT;
#endif
}
#endif
#if defined(PNG_UNKNOWN_CHUNKS_SUPPORTED)
void PNGAPI
png_set_unknown_chunks(png_structp png_ptr,
					   png_infop info_ptr, png_unknown_chunkp unknowns, int num_unknowns)
{
	png_unknown_chunkp np;
	int i;
	if(png_ptr == NULL || info_ptr == NULL || num_unknowns == 0)
		return;
	np = (png_unknown_chunkp)png_malloc_warn(png_ptr,
		(png_uint_32)(info_ptr->unknown_chunks_num + num_unknowns) *
		png_sizeof(png_unknown_chunk));
	if(np == NULL)
	{
		png_warning(png_ptr, "Out of memory while processing unknown chunk.");
		return;
	}
	png_memcpy(np, info_ptr->unknown_chunks,
		info_ptr->unknown_chunks_num * png_sizeof(png_unknown_chunk));
	png_free(png_ptr, info_ptr->unknown_chunks);
	info_ptr->unknown_chunks = NULL;
	for(i = 0; i < num_unknowns; i++)
	{
		png_unknown_chunkp to = np + info_ptr->unknown_chunks_num + i;
		png_unknown_chunkp from = unknowns + i;
		png_strncpy((png_charp)to->name, (png_charp)from->name, 5);
		to->data = (png_bytep)png_malloc_warn(png_ptr, (png_uint_32)(from->size));
		if(to->data == NULL)
		{
			png_warning(png_ptr, "Out of memory processing unknown chunk.");
		}
		else
		{
			png_memcpy(to->data, from->data, from->size);
			to->size = from->size;
			to->location = (png_byte)(png_ptr->mode & 0xff);
		}
	}
	info_ptr->unknown_chunks = np;
	info_ptr->unknown_chunks_num += num_unknowns;
#ifdef PNG_FREE_ME_SUPPORTED
	info_ptr->free_me |= PNG_FREE_UNKN;
#endif
}
void PNGAPI
png_set_unknown_chunk_location(png_structp png_ptr, png_infop info_ptr,
							   int chunk, int location)
{
	if(png_ptr != NULL && info_ptr != NULL && chunk >= 0 && chunk <
		(int)info_ptr->unknown_chunks_num)
		info_ptr->unknown_chunks[chunk].location = (png_byte)location;
}
#endif
#if defined(PNG_READ_EMPTY_PLTE_SUPPORTED) || \
	defined(PNG_WRITE_EMPTY_PLTE_SUPPORTED)
void PNGAPI
png_permit_empty_plte (png_structp png_ptr, int empty_plte_permitted)
{
	png_debug(1, "in png_permit_empty_plte, DEPRECATED.\n");
	if(png_ptr == NULL)
		return;
	png_ptr->mng_features_permitted = (png_byte)
		((png_ptr->mng_features_permitted & (~(PNG_FLAG_MNG_EMPTY_PLTE))) |
		((empty_plte_permitted & PNG_FLAG_MNG_EMPTY_PLTE)));
}
#endif
#if defined(PNG_MNG_FEATURES_SUPPORTED)
png_uint_32 PNGAPI
png_permit_mng_features (png_structp png_ptr, png_uint_32 mng_features)
{
	png_debug(1, "in png_permit_mng_features\n");
	if(png_ptr == NULL)
		return (png_uint_32)0;
	png_ptr->mng_features_permitted =
		(png_byte)(mng_features & PNG_ALL_MNG_FEATURES);
	return (png_uint_32)png_ptr->mng_features_permitted;
}
#endif
#if defined(PNG_UNKNOWN_CHUNKS_SUPPORTED)
void PNGAPI
png_set_keep_unknown_chunks(png_structp png_ptr, int keep, png_bytep
							chunk_list, int num_chunks)
{
	png_bytep new_list, p;
	int i, old_num_chunks;
	if(num_chunks == 0)
	{
		if(keep == PNG_HANDLE_CHUNK_ALWAYS || keep == PNG_HANDLE_CHUNK_IF_SAFE)
			png_ptr->flags |= PNG_FLAG_KEEP_UNKNOWN_CHUNKS;
		else
			png_ptr->flags &= ~PNG_FLAG_KEEP_UNKNOWN_CHUNKS;
		if(keep == PNG_HANDLE_CHUNK_ALWAYS)
			png_ptr->flags |= PNG_FLAG_KEEP_UNSAFE_CHUNKS;
		else
			png_ptr->flags &= ~PNG_FLAG_KEEP_UNSAFE_CHUNKS;
		return;
	}
	if(chunk_list == NULL)
		return;
	old_num_chunks = png_ptr->num_chunk_list;
	new_list = (png_bytep)png_malloc(png_ptr,
		(png_uint_32)(5*(num_chunks+old_num_chunks)));
	if(png_ptr->chunk_list != NULL)
	{
		png_memcpy(new_list, png_ptr->chunk_list,
			(png_size_t)(5*old_num_chunks));
		png_free(png_ptr, png_ptr->chunk_list);
		png_ptr->chunk_list = NULL;
	}
	png_memcpy(new_list+5*old_num_chunks, chunk_list,
		(png_size_t)(5*num_chunks));
	for(p = new_list+5*old_num_chunks+4, i = 0; i<num_chunks; i++, p += 5)
		*p = (png_byte)keep;
	png_ptr->num_chunk_list = old_num_chunks+num_chunks;
	png_ptr->chunk_list = new_list;
#ifdef PNG_FREE_ME_SUPPORTED
	png_ptr->free_me |= PNG_FREE_LIST;
#endif
}
#endif
#if defined(PNG_READ_USER_CHUNKS_SUPPORTED)
void PNGAPI
png_set_read_user_chunk_fn(png_structp png_ptr, png_voidp user_chunk_ptr,
						   png_user_chunk_ptr read_user_chunk_fn)
{
	png_debug(1, "in png_set_read_user_chunk_fn\n");
	png_ptr->read_user_chunk_fn = read_user_chunk_fn;
	png_ptr->user_chunk_ptr = user_chunk_ptr;
}
#endif
#if defined(PNG_INFO_IMAGE_SUPPORTED)
void PNGAPI
png_set_rows(png_structp png_ptr, png_infop info_ptr, png_bytepp row_pointers)
{
	png_debug1(1, "in %s storage function\n", "rows");
	if(png_ptr == NULL || info_ptr == NULL)
		return;
	if(info_ptr->row_pointers && (info_ptr->row_pointers != row_pointers))
		png_free_data(png_ptr, info_ptr, PNG_FREE_ROWS, 0);
	info_ptr->row_pointers = row_pointers;
	if(row_pointers)
		info_ptr->valid |= PNG_INFO_IDAT;
}
#endif
#ifdef PNG_WRITE_SUPPORTED
void PNGAPI
png_set_compression_buffer_size(png_structp png_ptr, png_uint_32 size)
{
	if(png_ptr->zbuf)
		png_free(png_ptr, png_ptr->zbuf);
	png_ptr->zbuf_size = (png_size_t)size;
	png_ptr->zbuf = (png_bytep)png_malloc(png_ptr, size);
	png_ptr->zstream.next_out = png_ptr->zbuf;
	png_ptr->zstream.avail_out = (uInt)png_ptr->zbuf_size;
}
#endif
void PNGAPI
png_set_invalid(png_structp png_ptr, png_infop info_ptr, int mask)
{
	if(png_ptr && info_ptr)
		info_ptr->valid &= ~(mask);
}
#ifndef PNG_1_0_X
#ifdef PNG_ASSEMBLER_CODE_SUPPORTED
void PNGAPI
png_set_asm_flags (png_structp png_ptr, png_uint_32 asm_flags)
{
	png_uint_32 settable_asm_flags;
	png_uint_32 settable_mmx_flags;
	settable_mmx_flags =
#ifdef PNG_HAVE_ASSEMBLER_COMBINE_ROW
		PNG_ASM_FLAG_MMX_READ_COMBINE_ROW  |
#endif
#ifdef PNG_HAVE_ASSEMBLER_READ_INTERLACE
		PNG_ASM_FLAG_MMX_READ_INTERLACE    |
#endif
#ifdef PNG_HAVE_ASSEMBLER_READ_FILTER_ROW
		PNG_ASM_FLAG_MMX_READ_FILTER_SUB   |
		PNG_ASM_FLAG_MMX_READ_FILTER_UP    |
		PNG_ASM_FLAG_MMX_READ_FILTER_AVG   |
		PNG_ASM_FLAG_MMX_READ_FILTER_PAETH |
#endif
		0;
	settable_asm_flags = settable_mmx_flags;
	if(!(png_ptr->asm_flags & PNG_ASM_FLAG_MMX_SUPPORT_COMPILED) ||
		!(png_ptr->asm_flags & PNG_ASM_FLAG_MMX_SUPPORT_IN_CPU))
	{
		settable_asm_flags &= ~settable_mmx_flags;
		png_ptr->asm_flags &= ~settable_mmx_flags;
	}
	png_ptr->asm_flags &= ~settable_asm_flags;
	png_ptr->asm_flags |= (asm_flags & settable_asm_flags);
}
#endif
#ifdef PNG_ASSEMBLER_CODE_SUPPORTED
void PNGAPI
png_set_mmx_thresholds (png_structp png_ptr,
						png_byte mmx_bitdepth_threshold,
						png_uint_32 mmx_rowbytes_threshold)
{
	png_ptr->mmx_bitdepth_threshold = mmx_bitdepth_threshold;
	png_ptr->mmx_rowbytes_threshold = mmx_rowbytes_threshold;
}
#endif
#ifdef PNG_SET_USER_LIMITS_SUPPORTED
void PNGAPI
png_set_user_limits (png_structp png_ptr, png_uint_32 user_width_max,
					 png_uint_32 user_height_max)
{
	png_ptr->user_width_max = user_width_max;
	png_ptr->user_height_max = user_height_max;
}
#endif
#endif
#define PNG_INTERNAL
#include "png.h"
#if defined(PNG_READ_BGR_SUPPORTED) || defined(PNG_WRITE_BGR_SUPPORTED)
void PNGAPI
png_set_bgr(png_structp png_ptr)
{
	png_debug(1, "in png_set_bgr\n");
	png_ptr->transformations |= PNG_BGR;
}
#endif
#if defined(PNG_READ_SWAP_SUPPORTED) || defined(PNG_WRITE_SWAP_SUPPORTED)
void PNGAPI
png_set_swap(png_structp png_ptr)
{
	png_debug(1, "in png_set_swap\n");
	if(png_ptr->bit_depth == 16)
		png_ptr->transformations |= PNG_SWAP_BYTES;
}
#endif
#if defined(PNG_READ_PACK_SUPPORTED) || defined(PNG_WRITE_PACK_SUPPORTED)
void PNGAPI
png_set_packing(png_structp png_ptr)
{
	png_debug(1, "in png_set_packing\n");
	if(png_ptr->bit_depth < 8)
	{
		png_ptr->transformations |= PNG_PACK;
		png_ptr->usr_bit_depth = 8;
	}
}
#endif
#if defined(PNG_READ_PACKSWAP_SUPPORTED)||defined(PNG_WRITE_PACKSWAP_SUPPORTED)
void PNGAPI
png_set_packswap(png_structp png_ptr)
{
	png_debug(1, "in png_set_packswap\n");
	if(png_ptr->bit_depth < 8)
		png_ptr->transformations |= PNG_PACKSWAP;
}
#endif
#if defined(PNG_READ_SHIFT_SUPPORTED) || defined(PNG_WRITE_SHIFT_SUPPORTED)
void PNGAPI
png_set_shift(png_structp png_ptr, png_color_8p true_bits)
{
	png_debug(1, "in png_set_shift\n");
	png_ptr->transformations |= PNG_SHIFT;
	png_ptr->shift = *true_bits;
}
#endif
#if defined(PNG_READ_INTERLACING_SUPPORTED) || \
	defined(PNG_WRITE_INTERLACING_SUPPORTED)
int PNGAPI
png_set_interlace_handling(png_structp png_ptr)
{
	png_debug(1, "in png_set_interlace handling\n");
	if(png_ptr->interlaced)
	{
		png_ptr->transformations |= PNG_INTERLACE;
		return (7);
	}
	return (1);
}
#endif
#if defined(PNG_READ_FILLER_SUPPORTED) || defined(PNG_WRITE_FILLER_SUPPORTED)
void PNGAPI
png_set_filler(png_structp png_ptr, png_uint_32 filler, int filler_loc)
{
	png_debug(1, "in png_set_filler\n");
	png_ptr->transformations |= PNG_FILLER;
	png_ptr->filler = (png_byte)filler;
	if(filler_loc == PNG_FILLER_AFTER)
		png_ptr->flags |= PNG_FLAG_FILLER_AFTER;
	else
		png_ptr->flags &= ~PNG_FLAG_FILLER_AFTER;
	if(png_ptr->color_type == PNG_COLOR_TYPE_RGB)
	{
		png_ptr->usr_channels = 4;
	}
	if(png_ptr->color_type == PNG_COLOR_TYPE_GRAY && png_ptr->bit_depth >= 8)
	{
		png_ptr->usr_channels = 2;
	}
}
#if !defined(PNG_1_0_X)
void PNGAPI
png_set_add_alpha(png_structp png_ptr, png_uint_32 filler, int filler_loc)
{
	png_debug(1, "in png_set_add_alpha\n");
	png_set_filler(png_ptr, filler, filler_loc);
	png_ptr->transformations |= PNG_ADD_ALPHA;
}
#endif
#endif
#if defined(PNG_READ_SWAP_ALPHA_SUPPORTED) || \
	defined(PNG_WRITE_SWAP_ALPHA_SUPPORTED)
void PNGAPI
png_set_swap_alpha(png_structp png_ptr)
{
	png_debug(1, "in png_set_swap_alpha\n");
	png_ptr->transformations |= PNG_SWAP_ALPHA;
}
#endif
#if defined(PNG_READ_INVERT_ALPHA_SUPPORTED) || \
	defined(PNG_WRITE_INVERT_ALPHA_SUPPORTED)
void PNGAPI
png_set_invert_alpha(png_structp png_ptr)
{
	png_debug(1, "in png_set_invert_alpha\n");
	png_ptr->transformations |= PNG_INVERT_ALPHA;
}
#endif
#if defined(PNG_READ_INVERT_SUPPORTED) || defined(PNG_WRITE_INVERT_SUPPORTED)
void PNGAPI
png_set_invert_mono(png_structp png_ptr)
{
	png_debug(1, "in png_set_invert_mono\n");
	png_ptr->transformations |= PNG_INVERT_MONO;
}
void
png_do_invert(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_invert\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row == NULL || row_info == NULL)
		return;
#endif
	if(row_info->color_type == PNG_COLOR_TYPE_GRAY)
	{
		png_bytep rp = row;
		png_uint_32 i;
		png_uint_32 istop = row_info->rowbytes;
		for(i = 0; i < istop; i++)
		{
			*rp = (png_byte)(~(*rp));
			rp++;
		}
	}
	else if(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA &&
		row_info->bit_depth == 8)
	{
		png_bytep rp = row;
		png_uint_32 i;
		png_uint_32 istop = row_info->rowbytes;
		for(i = 0; i < istop; i += 2)
		{
			*rp = (png_byte)(~(*rp));
			rp += 2;
		}
	}
	else if(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA &&
		row_info->bit_depth == 16)
	{
		png_bytep rp = row;
		png_uint_32 i;
		png_uint_32 istop = row_info->rowbytes;
		for(i = 0; i < istop; i += 4)
		{
			*rp = (png_byte)(~(*rp));
			*(rp+1) = (png_byte)(~(*(rp+1)));
			rp += 4;
		}
	}
}
#endif
#if defined(PNG_READ_SWAP_SUPPORTED) || defined(PNG_WRITE_SWAP_SUPPORTED)
void
png_do_swap(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_swap\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		row_info->bit_depth == 16)
	{
		png_bytep rp = row;
		png_uint_32 i;
		png_uint_32 istop= row_info->width * row_info->channels;
		for(i = 0; i < istop; i++, rp += 2)
		{
			png_byte t = *rp;
			*rp = *(rp + 1);
			*(rp + 1) = t;
		}
	}
}
#endif
#if defined(PNG_READ_PACKSWAP_SUPPORTED)||defined(PNG_WRITE_PACKSWAP_SUPPORTED)
static png_byte onebppswaptable[256] =
{
	0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0,
		0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
		0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8,
		0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
		0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4,
		0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
		0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC,
		0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
		0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2,
		0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
		0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA,
		0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
		0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6,
		0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
		0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE,
		0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
		0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1,
		0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
		0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9,
		0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
		0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5,
		0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
		0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED,
		0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
		0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3,
		0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
		0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB,
		0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
		0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7,
		0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
		0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF,
		0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
};
static png_byte twobppswaptable[256] =
{
	0x00, 0x40, 0x80, 0xC0, 0x10, 0x50, 0x90, 0xD0,
		0x20, 0x60, 0xA0, 0xE0, 0x30, 0x70, 0xB0, 0xF0,
		0x04, 0x44, 0x84, 0xC4, 0x14, 0x54, 0x94, 0xD4,
		0x24, 0x64, 0xA4, 0xE4, 0x34, 0x74, 0xB4, 0xF4,
		0x08, 0x48, 0x88, 0xC8, 0x18, 0x58, 0x98, 0xD8,
		0x28, 0x68, 0xA8, 0xE8, 0x38, 0x78, 0xB8, 0xF8,
		0x0C, 0x4C, 0x8C, 0xCC, 0x1C, 0x5C, 0x9C, 0xDC,
		0x2C, 0x6C, 0xAC, 0xEC, 0x3C, 0x7C, 0xBC, 0xFC,
		0x01, 0x41, 0x81, 0xC1, 0x11, 0x51, 0x91, 0xD1,
		0x21, 0x61, 0xA1, 0xE1, 0x31, 0x71, 0xB1, 0xF1,
		0x05, 0x45, 0x85, 0xC5, 0x15, 0x55, 0x95, 0xD5,
		0x25, 0x65, 0xA5, 0xE5, 0x35, 0x75, 0xB5, 0xF5,
		0x09, 0x49, 0x89, 0xC9, 0x19, 0x59, 0x99, 0xD9,
		0x29, 0x69, 0xA9, 0xE9, 0x39, 0x79, 0xB9, 0xF9,
		0x0D, 0x4D, 0x8D, 0xCD, 0x1D, 0x5D, 0x9D, 0xDD,
		0x2D, 0x6D, 0xAD, 0xED, 0x3D, 0x7D, 0xBD, 0xFD,
		0x02, 0x42, 0x82, 0xC2, 0x12, 0x52, 0x92, 0xD2,
		0x22, 0x62, 0xA2, 0xE2, 0x32, 0x72, 0xB2, 0xF2,
		0x06, 0x46, 0x86, 0xC6, 0x16, 0x56, 0x96, 0xD6,
		0x26, 0x66, 0xA6, 0xE6, 0x36, 0x76, 0xB6, 0xF6,
		0x0A, 0x4A, 0x8A, 0xCA, 0x1A, 0x5A, 0x9A, 0xDA,
		0x2A, 0x6A, 0xAA, 0xEA, 0x3A, 0x7A, 0xBA, 0xFA,
		0x0E, 0x4E, 0x8E, 0xCE, 0x1E, 0x5E, 0x9E, 0xDE,
		0x2E, 0x6E, 0xAE, 0xEE, 0x3E, 0x7E, 0xBE, 0xFE,
		0x03, 0x43, 0x83, 0xC3, 0x13, 0x53, 0x93, 0xD3,
		0x23, 0x63, 0xA3, 0xE3, 0x33, 0x73, 0xB3, 0xF3,
		0x07, 0x47, 0x87, 0xC7, 0x17, 0x57, 0x97, 0xD7,
		0x27, 0x67, 0xA7, 0xE7, 0x37, 0x77, 0xB7, 0xF7,
		0x0B, 0x4B, 0x8B, 0xCB, 0x1B, 0x5B, 0x9B, 0xDB,
		0x2B, 0x6B, 0xAB, 0xEB, 0x3B, 0x7B, 0xBB, 0xFB,
		0x0F, 0x4F, 0x8F, 0xCF, 0x1F, 0x5F, 0x9F, 0xDF,
		0x2F, 0x6F, 0xAF, 0xEF, 0x3F, 0x7F, 0xBF, 0xFF
};
static png_byte fourbppswaptable[256] =
{
	0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70,
		0x80, 0x90, 0xA0, 0xB0, 0xC0, 0xD0, 0xE0, 0xF0,
		0x01, 0x11, 0x21, 0x31, 0x41, 0x51, 0x61, 0x71,
		0x81, 0x91, 0xA1, 0xB1, 0xC1, 0xD1, 0xE1, 0xF1,
		0x02, 0x12, 0x22, 0x32, 0x42, 0x52, 0x62, 0x72,
		0x82, 0x92, 0xA2, 0xB2, 0xC2, 0xD2, 0xE2, 0xF2,
		0x03, 0x13, 0x23, 0x33, 0x43, 0x53, 0x63, 0x73,
		0x83, 0x93, 0xA3, 0xB3, 0xC3, 0xD3, 0xE3, 0xF3,
		0x04, 0x14, 0x24, 0x34, 0x44, 0x54, 0x64, 0x74,
		0x84, 0x94, 0xA4, 0xB4, 0xC4, 0xD4, 0xE4, 0xF4,
		0x05, 0x15, 0x25, 0x35, 0x45, 0x55, 0x65, 0x75,
		0x85, 0x95, 0xA5, 0xB5, 0xC5, 0xD5, 0xE5, 0xF5,
		0x06, 0x16, 0x26, 0x36, 0x46, 0x56, 0x66, 0x76,
		0x86, 0x96, 0xA6, 0xB6, 0xC6, 0xD6, 0xE6, 0xF6,
		0x07, 0x17, 0x27, 0x37, 0x47, 0x57, 0x67, 0x77,
		0x87, 0x97, 0xA7, 0xB7, 0xC7, 0xD7, 0xE7, 0xF7,
		0x08, 0x18, 0x28, 0x38, 0x48, 0x58, 0x68, 0x78,
		0x88, 0x98, 0xA8, 0xB8, 0xC8, 0xD8, 0xE8, 0xF8,
		0x09, 0x19, 0x29, 0x39, 0x49, 0x59, 0x69, 0x79,
		0x89, 0x99, 0xA9, 0xB9, 0xC9, 0xD9, 0xE9, 0xF9,
		0x0A, 0x1A, 0x2A, 0x3A, 0x4A, 0x5A, 0x6A, 0x7A,
		0x8A, 0x9A, 0xAA, 0xBA, 0xCA, 0xDA, 0xEA, 0xFA,
		0x0B, 0x1B, 0x2B, 0x3B, 0x4B, 0x5B, 0x6B, 0x7B,
		0x8B, 0x9B, 0xAB, 0xBB, 0xCB, 0xDB, 0xEB, 0xFB,
		0x0C, 0x1C, 0x2C, 0x3C, 0x4C, 0x5C, 0x6C, 0x7C,
		0x8C, 0x9C, 0xAC, 0xBC, 0xCC, 0xDC, 0xEC, 0xFC,
		0x0D, 0x1D, 0x2D, 0x3D, 0x4D, 0x5D, 0x6D, 0x7D,
		0x8D, 0x9D, 0xAD, 0xBD, 0xCD, 0xDD, 0xED, 0xFD,
		0x0E, 0x1E, 0x2E, 0x3E, 0x4E, 0x5E, 0x6E, 0x7E,
		0x8E, 0x9E, 0xAE, 0xBE, 0xCE, 0xDE, 0xEE, 0xFE,
		0x0F, 0x1F, 0x2F, 0x3F, 0x4F, 0x5F, 0x6F, 0x7F,
		0x8F, 0x9F, 0xAF, 0xBF, 0xCF, 0xDF, 0xEF, 0xFF
};
void
png_do_packswap(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_packswap\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		row_info->bit_depth < 8)
	{
		png_bytep rp, end, table;
		end = row + row_info->rowbytes;
		if(row_info->bit_depth == 1)
			table = onebppswaptable;
		else if(row_info->bit_depth == 2)
			table = twobppswaptable;
		else if(row_info->bit_depth == 4)
			table = fourbppswaptable;
		else
			return;
		for(rp = row; rp < end; rp++)
			*rp = table[*rp];
	}
}
#endif
#if defined(PNG_WRITE_FILLER_SUPPORTED) || \
	defined(PNG_READ_STRIP_ALPHA_SUPPORTED)
void
png_do_strip_filler(png_row_infop row_info, png_bytep row, png_uint_32 flags)
{
	png_debug(1, "in png_do_strip_filler\n");
#if defined(PNG_USELESS_TESTS_SUPPORTED)
	if(row != NULL && row_info != NULL)
#endif
	{
		png_bytep sp = row;
		png_bytep dp = row;
		png_uint_32 row_width = row_info->width;
		png_uint_32 i;
		if((row_info->color_type == PNG_COLOR_TYPE_RGB ||
			(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA &&
			(flags & PNG_FLAG_STRIP_ALPHA))) &&
			row_info->channels == 4)
		{
			if(row_info->bit_depth == 8)
			{
				if(flags & PNG_FLAG_FILLER_AFTER)
				{
					dp += 3; sp += 4;
					for(i = 1; i < row_width; i++)
					{
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						sp++;
					}
				}
				else
				{
					for(i = 0; i < row_width; i++)
					{
						sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
					}
				}
				row_info->pixel_depth = 24;
				row_info->rowbytes = row_width * 3;
			}
			else
			{
				if(flags & PNG_FLAG_FILLER_AFTER)
				{
					sp += 8; dp += 6;
					for(i = 1; i < row_width; i++)
					{
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						sp += 2;
					}
				}
				else
				{
					for(i = 0; i < row_width; i++)
					{
						sp += 2;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
						*dp++ = *sp++;
					}
				}
				row_info->pixel_depth = 48;
				row_info->rowbytes = row_width * 6;
			}
			row_info->channels = 3;
		}
		else if((row_info->color_type == PNG_COLOR_TYPE_GRAY ||
			(row_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA &&
			(flags & PNG_FLAG_STRIP_ALPHA))) &&
			row_info->channels == 2)
		{
			if(row_info->bit_depth == 8)
			{
				if(flags & PNG_FLAG_FILLER_AFTER)
				{
					for(i = 0; i < row_width; i++)
					{
						*dp++ = *sp++;
						sp++;
					}
				}
				else
				{
					for(i = 0; i < row_width; i++)
					{
						sp++;
						*dp++ = *sp++;
					}
				}
				row_info->pixel_depth = 8;
				row_info->rowbytes = row_width;
			}
			else
			{
				if(flags & PNG_FLAG_FILLER_AFTER)
				{
					sp += 4; dp += 2;
					for(i = 1; i < row_width; i++)
					{
						*dp++ = *sp++;
						*dp++ = *sp++;
						sp += 2;
					}
				}
				else
				{
					for(i = 0; i < row_width; i++)
					{
						sp += 2;
						*dp++ = *sp++;
						*dp++ = *sp++;
					}
				}
				row_info->pixel_depth = 16;
				row_info->rowbytes = row_width * 2;
			}
			row_info->channels = 1;
		}
		if(flags & PNG_FLAG_STRIP_ALPHA)
			row_info->color_type &= ~PNG_COLOR_MASK_ALPHA;
	}
}
#endif
#if defined(PNG_READ_BGR_SUPPORTED) || defined(PNG_WRITE_BGR_SUPPORTED)
void
png_do_bgr(png_row_infop row_info, png_bytep row)
{
	png_debug(1, "in png_do_bgr\n");
	if(
#if defined(PNG_USELESS_TESTS_SUPPORTED)
		row != NULL && row_info != NULL &&
#endif
		(row_info->color_type & PNG_COLOR_MASK_COLOR))
	{
		png_uint_32 row_width = row_info->width;
		if(row_info->bit_depth == 8)
		{
			if(row_info->color_type == PNG_COLOR_TYPE_RGB)
			{
				png_bytep rp;
				png_uint_32 i;
				for(i = 0, rp = row; i < row_width; i++, rp += 3)
				{
					png_byte save = *rp;
					*rp = *(rp + 2);
					*(rp + 2) = save;
				}
			}
			else if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
			{
				png_bytep rp;
				png_uint_32 i;
				for(i = 0, rp = row; i < row_width; i++, rp += 4)
				{
					png_byte save = *rp;
					*rp = *(rp + 2);
					*(rp + 2) = save;
				}
			}
		}
		else if(row_info->bit_depth == 16)
		{
			if(row_info->color_type == PNG_COLOR_TYPE_RGB)
			{
				png_bytep rp;
				png_uint_32 i;
				for(i = 0, rp = row; i < row_width; i++, rp += 6)
				{
					png_byte save = *rp;
					*rp = *(rp + 4);
					*(rp + 4) = save;
					save = *(rp + 1);
					*(rp + 1) = *(rp + 5);
					*(rp + 5) = save;
				}
			}
			else if(row_info->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
			{
				png_bytep rp;
				png_uint_32 i;
				for(i = 0, rp = row; i < row_width; i++, rp += 8)
				{
					png_byte save = *rp;
					*rp = *(rp + 4);
					*(rp + 4) = save;
					save = *(rp + 1);
					*(rp + 1) = *(rp + 5);
					*(rp + 5) = save;
				}
			}
		}
	}
}
#endif
#if defined(PNG_READ_USER_TRANSFORM_SUPPORTED) || \
	defined(PNG_WRITE_USER_TRANSFORM_SUPPORTED) || \
	defined(PNG_LEGACY_SUPPORTED)
void PNGAPI
png_set_user_transform_info(png_structp png_ptr, png_voidp
							user_transform_ptr, int user_transform_depth, int user_transform_channels)
{
	png_debug(1, "in png_set_user_transform_info\n");
#if defined(PNG_USER_TRANSFORM_PTR_SUPPORTED)
	png_ptr->user_transform_ptr = user_transform_ptr;
	png_ptr->user_transform_depth = (png_byte)user_transform_depth;
	png_ptr->user_transform_channels = (png_byte)user_transform_channels;
#else
	if(user_transform_ptr || user_transform_depth || user_transform_channels)
		png_warning(png_ptr,
		"This version of libpng does not support user transform info");
#endif
}
#endif
png_voidp PNGAPI
png_get_user_transform_ptr(png_structp png_ptr)
{
#if defined(PNG_USER_TRANSFORM_PTR_SUPPORTED)
	return ((png_voidp)png_ptr->user_transform_ptr);
#else
	if(png_ptr)
		return (NULL);
	return (NULL);
#endif
}
